--
-- ocp_monitor_b128r64.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
-- OCP Monitor
--
-- Module for capturing the last 2 OCP transactions

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.ocpad.all;
use work.ocpad_components.all;
library unisim;
use unisim.vcomponents.all;

entity ocp_monitor_b128r64 is
  port (
    -- Interface to OCP Burst Profile
    burst_clk : in  std_logic;
    burst_m2s : in  ocpad_burst128_m2sT;
    burst_s2m : in  ocpad_burst128_s2mT;
    -- Interface to OCP Register Profile
    reg_clk  : in  std_logic;    
    reg_m2s  : in  ocpad_register64_m2sT;
    reg_s2m  : out ocpad_register64_s2mT;
    -- Debug Data Inputs
    debug    : in  std_logic_vector(127 downto 0)
    );

end ocp_monitor_b128r64;

architecture mixed of ocp_monitor_b128r64 is
  
  constant number_of_registers  : integer := 3;  

  signal registers_out,registers_back : register64_array(0 to number_of_registers-1);
  signal registers_written : std_logic_vector(number_of_registers-1 downto 0);
  
  signal reg_m2s_array : ocpad_register64_m2sT_array(0 to 10);
  signal reg_s2m_array : ocpad_register64_s2mT_array(0 to 10);

  type bram_addr_type is array (0 to 9) of std_logic_vector(15 downto 0);
  type bram_data_type is array (0 to 9) of std_logic_vector(63 downto 0);
  
  signal bram_addr  : bram_addr_type;
  signal bram_write : std_logic_vector(9 downto 0);
  signal bram_read  : std_logic_vector(9 downto 0);
  signal bram_wdata : bram_data_type;
  signal bram_rdata : bram_data_type;
 
  
  component ocp_bram_interface64
    generic (
      read_latency : integer := 1);
    port (
      -- OCP Interface
      ocp_clk    : in  std_logic;
      ocp_rst    : in  std_logic;
      ocp_m2s    : in  ocpad_register64_m2sT;
      ocp_s2m    : out ocpad_register64_s2mT;
      -- simple bus interface
      bram_addr  : out std_logic_vector(15 downto 0);
      bram_write : out std_logic;
      bram_read  : out std_logic;
      bram_be    : out std_logic_vector(7 downto 0);
      bram_wdata : out std_logic_vector(63 downto 0);
      bram_rdata : in  std_logic_vector(63 downto 0)
      );
  end component;
  
  signal counter_msw,counter_lsw : std_logic_vector(31 downto 0) := (others => '0');
    
  signal capturing : std_logic := '0';
  signal capture_buffer : std_logic := '0';
  
  signal capture_data0 : std_logic_vector(64*5-1 downto 0);
  signal capture_data : std_logic_vector(64*10-1 downto 0);
  signal capture_we : std_logic_vector(9 downto 0) := (others => '0');
  signal capture_addr: std_logic_vector(8 downto 0);
  
  signal capture_count0,capture_count1 : std_logic_vector(63 downto 0);
  
  signal capture_data_r4 : std_logic_vector(64*10-1 downto 0);
  signal capture_data_r3 : std_logic_vector(64*10-1 downto 0);
  signal capture_data_r2 : std_logic_vector(64*10-1 downto 0);
  signal capture_data_r1 : std_logic_vector(64*10-1 downto 0);
  
  
  component infer_sdpbram 
    generic (
      width : integer := 64);
    port (
      wclk   : in  std_logic;
      rclk   : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      waddr  : in  std_logic_vector(8 downto 0) := (others => '0');
      raddr  : in  std_logic_vector(8 downto 0) := (others => '0');
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  signal capture_only_reads,capture_only_writes : std_logic;
  signal debug_trigger : std_logic;

  
begin  -- mixed


  
  capture_bus_data: process (burst_clk)
  begin  -- process retime_regs_bclk
    if burst_clk'event and burst_clk = '1' then  -- rising clock edge
      counter_lsw <= counter_lsw+1;
      if AND_reduce(counter_lsw) = '1'  then
        counter_msw <= counter_msw+1;
      end if;

      capture_only_reads <= registers_out(0)(0);
      capture_only_writes <= registers_out(0)(1);
      debug_trigger <= registers_out(0)(2);

  
      if (debug_trigger = '1' and capturing = '0' and debug(0) = '1' ) or (debug_trigger = '0' and ((burst_m2s.Cmd = ocp_Cmd_WRITE and capture_only_reads = '0') or (burst_m2s.Cmd = ocp_Cmd_READ and capture_only_writes = '0'))  and burst_s2m.CmdAccept = '1') then
        capturing <= '1';
        if capture_buffer = '0' then
          capture_we <= "0000011111";
          capture_count0 <= counter_msw & counter_lsw;
        else
          capture_we <= "1111100000";
          capture_count1 <= counter_msw & counter_lsw;
        end if;
        capture_buffer <= not capture_buffer;
        capture_addr <= (others => '0');
      else
        if capturing = '1' then
          capture_addr <= capture_addr+1;
          if AND_reduce(capture_addr) = '1' then
            capturing <= '0';
            capture_we <= (others => '0');
          end if;
        end if;
      end if;
      if capture_only_writes = '0' then
        capture_data0(64*5-1 downto 64*3) <= burst_s2m.Data;
      else
        capture_data0(64*5-1 downto 64*3) <= debug;
      end if;
      if capture_only_reads = '0' then
        capture_data0(64*3-1 downto 64*1) <= burst_m2s.Data;
      else
        capture_data0(64*3-1 downto 64*1) <= debug;
      end if;
      capture_data0(60 downto 36) <= burst_m2s.Addr(28 downto 4);
      capture_data0(35 downto 32) <= counter_lsw(3 downto 0);
      capture_data0(2 downto 0) <= burst_m2s.Cmd;
      capture_data0(3) <= burst_m2s.DataValid;
      capture_data0(4) <= burst_m2s.RespAccept;
      capture_data0(5) <= burst_s2m.CmdAccept;
      capture_data0(6) <= burst_s2m.DataAccept;
      capture_data0(7) <= burst_m2s.DataLast;
      capture_data0(15 downto 8) <= burst_m2s.BurstLength;
      capture_data0(31 downto 16) <= burst_m2s.DataByteEn;
      capture_data0(63 downto 62) <= burst_s2m.Resp;
      capture_data0(61) <= burst_s2m.RespLast;
      capture_data <= capture_data_r1;
      capture_data_r1 <= capture_data_r2;
      capture_data_r2 <= capture_data_r3;
      capture_data_r3 <= capture_data_r4;
      capture_data_r4 <= capture_data0 & capture_data0;
    end if;
  end process capture_bus_data;

  
  
   dm0 : ocp_demux_r64
    generic map (
      number_of_slaves  => 11,
      demux_address_lsb => 12)
    port map (
      clk        => reg_clk,
      master_m2s => reg_m2s,
      master_s2m => reg_s2m,
      slaves_m2s => reg_m2s_array,
      slaves_s2m => reg_s2m_array);
  
  -- Register Interface
  ocp_ra0: ocp_register_interface64
      generic map (
        number_of_registers => number_of_registers)
      port map (
        ocp_clk           => reg_clk,
        ocp_m2s           => reg_m2s_array(0),
        ocp_s2m           => reg_s2m_array(0),
        registers_out     => registers_out,
        registers_back    => registers_back,
        register_written => registers_written);
  
   registers_back(0) <= EXT( capture_addr & capture_buffer & capturing,64);
   registers_back(1) <= capture_count0;
   registers_back(2) <= capture_count1;
        
        
  gen_debug_rams: for i in 0 to 9 generate        
    ori1 : ocp_bram_interface64
      generic map (
        read_latency => 2)
      port map (
        ocp_clk    => reg_clk,
        ocp_rst    => '0',
        ocp_m2s    => reg_m2s_array(i+1),
        ocp_s2m    => reg_s2m_array(i+1),
        bram_addr  => bram_addr(i),
        bram_write => open,
        bram_read  => bram_read(i),
        bram_be    => open,
        bram_wdata => open,
        bram_rdata => bram_rdata(i));

    sdpbram: infer_sdpbram
      generic map (
        width => 64)
      port map (
        wclk => burst_clk,
        rclk => reg_clk,
        wadv => capture_we(i),
        wdata => capture_data(64*i+63 downto 64*i),
        waddr => capture_addr(8 downto 0), 
        raddr => bram_addr(i)(8 downto 0),
        rdata => bram_rdata(i));  
 
     end generate;
      
   
      
  
  
end mixed;
