--
-- ocpad_pkg.vhd
-- Package Header of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--  
-- OCP AD Type definitions
--
-- Defines a number of Alpha Data OCP Compliant Bus record Types
--

library IEEE;
  use IEEE.std_logic_1164.all;

package ocpad is

  constant ocpad_burst_width : integer := 8;
  subtype ocp_BurstLengthT is std_logic_vector(ocpad_burst_width-1 downto 0);

  subtype ocp_Data128T is std_logic_vector(127 downto 0);
  subtype ocp_Data64T is std_logic_vector(63 downto 0);
  subtype ocp_Data32T is std_logic_vector(31 downto 0);
  subtype ocp_Data16T is std_logic_vector(15 downto 0);
  subtype ocp_Data8T  is std_logic_vector(7  downto 0);
  subtype ocp_Data4T  is std_logic_vector(3  downto 0);

  subtype  ocp_CmdT is std_logic_vector(2 downto 0);
  constant ocp_Cmd_IDLE         : ocp_CmdT  := "000";
  constant ocp_Cmd_WRITE        : ocp_CmdT  := "001";
  constant ocp_Cmd_READ         : ocp_CmdT  := "010";
  constant ocp_Cmd_WRITENONPOST : ocp_CmdT  := "101";

  subtype  ocp_RespT is std_logic_vector(1 downto 0);
  constant ocp_Resp_NONE    : ocp_RespT := "00";
  constant ocp_Resp_VALID   : ocp_RespT := "01";
  constant ocp_Resp_FAILED  : ocp_RespT := "10";
  constant ocp_Resp_ERROR   : ocp_RespT := "11";

  function to_std_logic(bool : boolean) return std_logic;

  -- Register Bus for simple non-bursting register control interfaces
  -- 32 bits wide
  type ocpad_register32_m2sT is record   
    -- Request group...
    Cmd          : ocp_CmdT;     -- Command b'000=Idle b'010=Read b'001=Write
    Addr         : ocp_Data32T;  -- Starting Byte Address 
    -- DataHandshake group...
    Data         : ocp_Data32T;  -- 32-bits of write data 
    DataByteEn   : ocp_Data4T;   -- 4-bits of byte-lane
    -- Sideband group...
    Reset_n      : std_logic;
  end record ocpad_register32_m2sT;       
  type ocpad_register32_s2mT is record
    -- Request group...
    CmdAccept    : std_logic;    -- Slave indication command/request accepted
    -- Response group...
    Data         : ocp_Data32T;  -- 32-bits of read data 
    Resp         : ocp_RespT;    -- Slave read response b'00=None b'01=Valid
  end record ocpad_register32_s2mT;

  --
  -- Define default tieoffs...
  constant ocpad_register32_default_m2s : ocpad_register32_m2sT :=
  ( Cmd          => ocp_Cmd_IDLE,
    Addr         => (others=>'0'),
    Data         => (others=>'0'),
    DataByteEn   => (others=>'0'),
    Reset_n      => '1'
  );
  constant ocpad_register32_default_s2m : ocpad_register32_s2mT :=
  ( CmdAccept    => '0',
    Data         => (others=>'0'),
    Resp         => ocp_Resp_NONE
  );  
  
  -- Register Bus for simple non-bursting register control interfaces
  -- 64 bits wide
  type ocpad_register64_m2sT is record   
    -- Request group...
    Cmd          : ocp_CmdT;     -- Command b'000=Idle b'010=Read b'001=Write
    Addr         : ocp_Data32T;  -- Starting Byte Address 
    -- DataHandshake group...
    Data         : ocp_Data64T;  -- 32-bits of write data 
    DataByteEn   : ocp_Data8T;   -- 4-bits of byte-lane
    -- Sideband group...
    Reset_n      : std_logic;
  end record ocpad_register64_m2sT;       
  type ocpad_register64_s2mT is record
    -- Request group...
    CmdAccept    : std_logic;    -- Slave indication command/request accepted
    -- Response group...
    Data         : ocp_Data64T;  -- 32-bits of read data 
    Resp         : ocp_RespT;    -- Slave read response b'00=None b'01=Valid
  end record ocpad_register64_s2mT;
  
  --
  -- Define default tieoffs...
  constant ocpad_register64_default_m2s : ocpad_register64_m2sT :=
  ( Cmd          => ocp_Cmd_IDLE,
    Addr         => (others=>'0'),
    Data         => (others=>'0'),
    DataByteEn   => (others=>'0'),
    Reset_n      => '1'
  );
  constant ocpad_register64_default_s2m : ocpad_register64_s2mT :=
  ( CmdAccept    => '0',
    Data         => (others=>'0'),
    Resp         => ocp_Resp_NONE
  );  

  
  -- Burst Capable bus for high performance data flows
  -- 64 bit wide data path
  type ocpad_burst64_m2sT is record    -- Block Data Flow from Master to Slave...
    -- Request group...
    Cmd          : ocp_CmdT;     -- Command b'000=Idle b'010=Read b'001=Write
    Addr         : ocp_Data32T;  -- Starting Byte Address 
    BurstLength  : ocp_BurstLengthT;   -- BurstLength in number of transfers
    BurstSingleReq : std_logic; -- 1=BurstLength is exact, known at start; 0=Use DataLast
    -- DataHandshake group...
    Data         : ocp_Data64T;  -- 64-bits of write data 
    DataByteEn   : ocp_Data8T;   -- 8-bits of byte-lane enables (when supported)
    DataLast     : std_logic;    -- Indicates the last write Data beat of this burst
    DataValid    : std_logic;    -- Qualifies Data and the other signals in bundle
    -- Response group...
    RespAccept   : std_logic;    -- Master indication read data was accepted
    -- Sideband group...
    Reset_n      : std_logic;
  end record ocpad_burst64_m2sT;       
  --
  --
  type ocpad_burst64_s2mT is record    -- Block Data Flow from Slave to Master...
    -- Request group...
    CmdAccept    : std_logic;    -- Slave indication command/request accepted
    -- DataHandshake group...
    DataAccept   : std_logic;    -- Slave indication write data accepted
    -- Response group...
    Data         : ocp_Data64T;  -- 64-bits of read data 
    Resp         : ocp_RespT;    -- Slave read response b'00=None b'01=Valid
    RespLast     : std_logic;    -- Indicates the last  (read) response of this burst
    -- Sideband group...
    Error        : std_logic;
  end record ocpad_burst64_s2mT;

  --
  -- Define default tieoffs...
  constant ocpad_burst64_default_m2s : ocpad_burst64_m2sT :=
  ( Cmd          => ocp_Cmd_IDLE,
    Addr         => (others=>'0'),
    BurstLength  => (others=>'0'),
    BurstSingleReq => '0',
    Data         => (others=>'0'),
    DataByteEn   => (others=>'0'),
    DataLast     => '0',
    DataValid    => '0',
    RespAccept   => '0',
    Reset_n      => '1'
  );
  constant ocpad_burst64_default_s2m : ocpad_burst64_s2mT :=
  ( CmdAccept    => '0',
    DataAccept   => '0',
    Data         => (others=>'0'),
    Resp         => ocp_Resp_NONE,
    RespLast     => '0',
    Error        => '0'
  );
  constant ocpad_burst64_error_s2m : ocpad_burst64_s2mT :=
  ( CmdAccept    => '0',
    DataAccept   => '0',
    Data         => (others=>'0'),
    Resp         => ocp_Resp_NONE,
    RespLast     => '0',
    Error        => '0'
  );  

  
  -- Burst Capable bus for high performance data flows
  -- 128 bit wide data path
  type ocpad_burst128_m2sT is record    -- Block Data Flow from Master to Slave...
    -- Request group...
    Cmd          : ocp_CmdT;     -- Command b'000=Idle b'010=Read b'001=Write
    Addr         : ocp_Data32T;  -- Starting Byte Address 
    BurstLength  : ocp_BurstLengthT;   -- BurstLength in number of transfers
    BurstSingleReq : std_logic; -- 1=BurstLength is exact, known at start; 0=Use DataLast
    -- DataHandshake group...
    Data         : ocp_Data128T;  -- 128-bits of write data 
    DataByteEn   : ocp_Data16T;   -- 16-bits of byte-lane enables (when supported)
    DataLast     : std_logic;    -- Indicates the last write Data beat of this burst
    DataValid    : std_logic;    -- Qualifies Data and the other signals in bundle
    -- Response group...
    RespAccept   : std_logic;    -- Master indication read data was accepted
    -- Sideband group...
    Reset_n      : std_logic;
  end record ocpad_burst128_m2sT;       
  type ocpad_burst128_s2mT is record    -- Block Data Flow from Slave to Master...
    -- Request group...
    CmdAccept    : std_logic;    -- Slave indication command/request accepted
    -- DataHandshake group...
    DataAccept   : std_logic;    -- Slave indication write data accepted
    -- Response group...
    Data         : ocp_Data128T;  -- 64-bits of read data 
    Resp         : ocp_RespT;    -- Slave read response b'00=None b'01=Valid
    RespLast     : std_logic;    -- Indicates the last  (read) response of this burst
    -- Sideband group...
    Error        : std_logic;
  end record ocpad_burst128_s2mT;       

  --
  -- Define default tieoffs...
  constant ocpad_burst128_default_m2s : ocpad_burst128_m2sT :=
  ( Cmd          => ocp_Cmd_IDLE,
    Addr         => (others=>'0'),
    BurstLength  => (others=>'0'),
    BurstSingleReq => '0',
    Data         => (others=>'0'),
    DataByteEn   => (others=>'0'),
    DataLast     => '0',
    DataValid    => '0',
    RespAccept   => '0',
    Reset_n      => '1'
  );
  constant ocpad_burst128_default_s2m : ocpad_burst128_s2mT :=
  ( CmdAccept    => '0',
    DataAccept   => '0',
    Data         => (others=>'0'),
    Resp         => ocp_Resp_NONE,
    RespLast     => '0',
    Error        => '0'
  );

  constant ocpad_burst128_error_s2m : ocpad_burst128_s2mT :=
  ( CmdAccept    => '0',
    DataAccept   => '0',
    Data         => (others=>'0'),
    Resp         => ocp_Resp_NONE,
    RespLast     => '0',
    Error        => '0'
  );


  -- Burst Capable bus for high performance data flows
  -- 32 bit wide data path
  type ocpad_burst32_m2sT is record    -- Block Data Flow from Master to Slave...
    -- Request group...
    Cmd          : ocp_CmdT;     -- Command b'000=Idle b'010=Read b'001=Write
    Addr         : ocp_Data32T;  -- Starting Byte Address 
    BurstLength  : ocp_BurstLengthT;   -- BurstLength in number of transfers
    BurstSingleReq : std_logic; -- 1=BurstLength is exact, known at start; 0=Use DataLast
    -- DataHandshake group...
    Data         : ocp_Data32T;  -- 128-bits of write data 
    DataByteEn   : ocp_Data4T;   -- 16-bits of byte-lane enables (when supported)
    DataLast     : std_logic;    -- Indicates the last write Data beat of this burst
    DataValid    : std_logic;    -- Qualifies Data and the other signals in bundle
    -- Response group...
    RespAccept   : std_logic;    -- Master indication read data was accepted
    -- Sideband group...
    Reset_n      : std_logic;
  end record ocpad_burst32_m2sT;       
  type ocpad_burst32_s2mT is record    -- Block Data Flow from Slave to Master...
    -- Request group...
    CmdAccept    : std_logic;    -- Slave indication command/request accepted
    -- DataHandshake group...
    DataAccept   : std_logic;    -- Slave indication write data accepted
    -- Response group...
    Data         : ocp_Data32T;  -- 64-bits of read data 
    Resp         : ocp_RespT;    -- Slave read response b'00=None b'01=Valid
    RespLast     : std_logic;    -- Indicates the last  (read) response of this burst
    -- Sideband group...
    Error        : std_logic;
  end record ocpad_burst32_s2mT;       

  --
  -- Define default tieoffs...
  constant ocpad_burst32_default_m2s : ocpad_burst32_m2sT :=
  ( Cmd          => ocp_Cmd_IDLE,
    Addr         => (others=>'0'),
    BurstLength  => (others=>'0'),
    BurstSingleReq => '0',
    Data         => (others=>'0'),
    DataByteEn   => (others=>'0'),
    DataLast     => '0',
    DataValid    => '0',
    RespAccept   => '0',
    Reset_n      => '1'
  );
  constant ocpad_burst32_default_s2m : ocpad_burst32_s2mT :=
  ( CmdAccept    => '0',
    DataAccept   => '0',
    Data         => (others=>'0'),
    Resp         => ocp_Resp_NONE,
    RespLast     => '0',
    Error        => '0'
  );

   constant ocpad_burst32_error_s2m : ocpad_burst32_s2mT :=
  ( CmdAccept    => '0',
    DataAccept   => '0',
    Data         => (others=>'0'),
    Resp         => ocp_Resp_NONE,
    RespLast     => '0',
    Error        => '1'
  );  

 
  type ocpad_register32_s2mT_array is array (natural range <>) of ocpad_register32_s2mT;
  type ocpad_register32_m2sT_array is array (natural range <>) of ocpad_register32_m2sT;
  type ocpad_register64_s2mT_array is array (natural range <>) of ocpad_register64_s2mT;
  type ocpad_register64_m2sT_array is array (natural range <>) of ocpad_register64_m2sT;
  type ocpad_burst64_s2mT_array is array (natural range <>) of ocpad_burst64_s2mT;
  type ocpad_burst64_m2sT_array is array (natural range <>) of ocpad_burst64_m2sT;
  type ocpad_burst128_s2mT_array is array (natural range <>) of ocpad_burst128_s2mT;
  type ocpad_burst128_m2sT_array is array (natural range <>) of ocpad_burst128_m2sT;
  type ocpad_burst32_s2mT_array is array (natural range <>) of ocpad_burst32_s2mT;
  type ocpad_burst32_m2sT_array is array (natural range <>) of ocpad_burst32_m2sT;
 
  

  type register64_array is array (natural range <>) of std_logic_vector(63 downto 0);
  type register32_array is array (natural range <>) of std_logic_vector(31 downto 0);


  component ocp_register_interface64
    generic (
      number_of_registers : natural := 1);
    port (
      -- OCP Interface
      ocp_clk : in std_logic;
      ocp_m2s : in ocpad_register64_m2sT;
      ocp_s2m : out ocpad_register64_s2mT;
      -- simple bus interface
      registers_out : out register64_array(0 to number_of_registers-1);
      registers_back : in register64_array(0 to number_of_registers-1);
      register_written : out std_logic_vector(number_of_registers-1 downto 0) 
      );
  end component;

  component ocp_register_interface32
    generic (
      number_of_registers : natural := 1);
    port (
      -- OCP Interface
      ocp_clk : in std_logic;
      ocp_m2s : in ocpad_register32_m2sT;
      ocp_s2m : out ocpad_register32_s2mT;
      -- simple bus interface
      registers_out : out register32_array(0 to number_of_registers-1);
      registers_back : in register32_array(0 to number_of_registers-1);
      register_written : out std_logic_vector(number_of_registers-1 downto 0) 
      );
  end component;
  
  
end package ocpad;

package body ocpad is

  function to_std_logic (bool : boolean) return std_logic is begin
   if (bool) then return '1'; else return '0'; end if;
  end function to_std_logic;

end package body ocpad;
