--
-- data_sink_b128r32.vhd
-- Dataflow Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
-- Data Sink
--
-- Module for recording a sequence of valid data words, writing it from a
-- memory bank located on a burst 128 profile


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.ocpad.all;

entity data_sink_b128r32 is
  
  generic (
    data_width : integer := 16;
    data_shift : integer := 16;
    shift_count_width : integer := 4;
    family     : family_t := family_virtex4;
    fifo_order : integer := 9);
  port (
    -- Interface to OCP Burst Profile
    burst_clk : in  std_logic;
    burst_m2s : out ocpad_burst128_m2sT;
    burst_s2m : in  ocpad_burst128_s2mT;
    -- Interface to OCP Register Profile
    reg_clk  : in  std_logic;    
    reg_m2s  : in  ocpad_register32_m2sT;
    reg_s2m  : out ocpad_register32_s2mT;
    -- Interface to Data Flow
    data_clk : in  std_logic;
    data : in std_logic_vector(data_width-1 downto 0);
    data_valid : in std_logic;
    rfd : out std_logic;
    flush_packet : in std_logic;
    -- Interrupt out
    irq : out std_logic);

end data_sink_b128r32;

architecture mixed of data_sink_b128r32 is

  constant number_of_registers  : integer := 7;
  constant burst_width : integer := 128;
  constant reg_width : integer := 32;
  constant shift_count : integer := burst_width/data_shift;
  constant burst_length : integer := 64;
  
  
  signal registers_out,registers_back : register32_array(0 to number_of_registers-1);
  signal registers_written : std_logic_vector(number_of_registers-1 downto 0);
  
  signal start_address,word_limit : std_logic_vector(31 downto 0);
  signal continuous : std_logic;
  signal running : std_logic;
  signal ctrl_written : std_logic;
  signal ctrl_start_bit : std_logic;
  signal ctrl_clear_bit : std_logic;
  signal start : std_logic;
  signal clear : std_logic;
  signal rst : std_logic;
  
  signal interrupt_count : std_logic_vector(31 downto 0) := (others => '0');
  signal interrupt_limit : std_logic_vector(31 downto 0) := (others => '0');
  signal interrupt_limit_m1 : std_logic_vector(31 downto 0):= (others => '0');
  signal icount_active : std_logic := '0';

  signal ctrl_written_d : std_logic;
  signal ctrl_clear_bit_d : std_logic;
  signal clear_d : std_logic;
  
  type state_type is (idle,gen_write,wait_write);
  signal state : state_type := idle;
  signal next_burst : std_Logic_vector(7 downto 0):= (others => '0');
  signal addr : std_logic_vector(31 downto 0):= (others => '0');
  signal word_count : std_logic_vector(31 downto 0):= (others => '0');
  signal last_data : std_logic;
  signal burst_count : std_logic_vector(7 downto 0):= (others => '0');
  
  signal burst_s2m_dv : std_logic;

  signal fifo_nfull : std_logic;
  signal fifo_empty : std_logic;
  signal fifo_data_out : std_logic_vector(burst_width-1 downto 0);
  signal fifo_radv : std_logic;
  signal fifo_rlevel : std_logic_vector(15 downto 0);
  signal fifo_wadv : std_logic;
  
  signal data_ce : std_logic;
  signal rate_divide_pulse : std_logic;
  
  signal shift_counter : std_logic_vector(shift_count_width-1 downto 0):= (others => '0');
  signal shift_reg : std_logic_vector(burst_width-1 downto 0):= (others => '0');

  signal flush_packet_bclk : std_logic := '0';
  signal flush_packet_delayed : std_logic_vector(7 downto 0):= (others => '0');
  signal flushing_shift : std_logic := '0';
  signal flushing  : std_logic := '0';
  signal last_burst  : std_logic := '0';

  signal running_dclk : std_logic;
  
  component async_pulse
    port (
      clk : in std_logic;
      I   : in std_Logic;
      O   : out std_Logic);
  end component;
  
  signal clearing : std_Logic;
  signal clear_fifo : std_Logic;

begin  -- mixed

  -- Register Interface
  ocp_ra0: ocp_register_interface32
      generic map (
        number_of_registers => number_of_registers)
      port map (
        ocp_clk           => reg_clk,
        ocp_m2s           => reg_m2s,
        ocp_s2m           => reg_s2m,
        registers_out     => registers_out,
        registers_back    => registers_back,
        register_written => registers_written);

  -- Resample Registers into Burst Clock Domain
  retime_regs_bclk: process (burst_clk)
  begin  -- process retime_regs_bclk
    if burst_clk'event and burst_clk = '1' then  -- rising clock edge
      start_address <= registers_out(0)(31 downto 0);
      word_limit <= registers_out(1)(31 downto 0);
      continuous <= registers_out(3)(1);
      ctrl_start_bit <= registers_out(3)(0);
      ctrl_clear_bit <= registers_out(3)(2); 
      start <= ctrl_written and ctrl_start_bit;
      clear <= ctrl_written and ctrl_clear_bit;
    end if;
  end process retime_regs_bclk;

  -- Test when CTRL (reg 3) is written
  ap0: async_pulse
    port map (
      clk => burst_clk,
      I   => registers_written(3),
      O   => ctrl_written);

  -- Test when CTRL (reg 3) is written
  ap1: async_pulse
    port map (
      clk => data_clk,
      I   => registers_written(3),
      O   => ctrl_written_d);
  
  -- Resample Registers into Data Clock Domain
  retime_regs_dclk: process (data_clk)
  begin  -- process retime_regs_bclk
    if data_clk'event and data_clk = '1' then  -- rising clock edge
      interrupt_limit <= registers_out(2)(31 downto 0);
      ctrl_clear_bit_d <= registers_out(3)(2);
      clear_d <= ctrl_written_d and ctrl_clear_bit_d;
    end if;
  end process retime_regs_dclk;

  -- Make registers readable
  registers_back(0) <= registers_out(0);
  registers_back(1) <= registers_out(1);
  registers_back(2) <= registers_out(2);
  registers_back(3) <= EXT(registers_out(3)(1) & running,reg_width); 
  registers_back(4) <= EXT(addr,reg_width);
  registers_back(5) <= EXT(word_count,reg_width);
  registers_back(6) <= EXT(interrupt_count,reg_width);


    
  -- Burst Read State Machine
  rst <= not reg_m2s.Reset_n;

  -- Flush OCP burst state machine
  ap2: async_pulse
    port map (
      clk => burst_clk,
      I   => flush_packet_delayed(7),
      O   => flush_packet_bclk);
  
  burst_reader: process (burst_clk, rst)
  begin  -- process burst_reader
    if rst = '1' then             -- asynchronous reset
      addr <= (others => '0');
      word_count <= (others => '0');
      state <= idle;
      next_burst <= CONV_STD_LOGIC_VECTOR(burst_length,8);
      running <= '0';
      burst_count <= (others => '0');
      flushing <= '0';
      last_burst <= '0';
      clearing <= '0';
      clear_fifo <= '0';
    elsif burst_clk'event and burst_clk = '1' then  -- rising clock edge      
      if running = '0' then
        if start = '1' then
          addr <= start_address;
          word_count <= word_limit; 
          running <= '1';         
        end if;
        if clearing = '1' then
          clearing <= '0';
          clear_fifo <= '1';
        else
          clear_fifo <= clear;
        end if;
        state <= idle;
        flushing <= '0';
        last_burst <= '0';
      else
        if flush_packet_bclk = '1' then
          flushing <= '1';
        end if;
        case state is
          when idle =>   
            if clearing = '1' then
              running <= '0';
            elsif OR_reduce(word_count) = '1' then
              if fifo_rlevel >= EXT(next_burst,16) and fifo_empty = '0' then
                state <= gen_write;
              elsif flushing = '1' and fifo_empty = '1' then
                word_count <= (others => '0');
              end if;            
            else
              if continuous = '1' then
                addr <= start_address;
                word_count <= word_limit;
                flushing <= '0';
              else
                running <= '0';
              end if;
            end if;
          when gen_write =>
            if burst_s2m.CmdAccept = '1' then
              addr <= addr + burst_length*(burst_width/8);  -- Increment
                                                            -- address by burst data width in bytes
              state <= wait_write;
              if word_count < burst_length or last_burst = '1' or clearing = '1' then
                word_count <= (others => '0');
              else
                word_count <= word_count - burst_length; 

              end if;
              burst_count <= next_burst-1;
            end if;
          when wait_write =>
            if burst_s2m.DataAccept = '1' and last_data = '1' then
              state <= idle;
            end if;
            if burst_s2m.DataAccept = '1' then
              burst_count <= burst_count-1;
            end if;
          when others =>
            state <= idle;
        end case;
        if clear = '1' or burst_s2m.error = '1' then
          clearing <= '1';
        end if;
      end if;
      if word_count < burst_length  then
        next_burst <= word_count(7 downto 0);
      elsif flushing = '1' and fifo_rlevel < CONV_STD_LOGIC_VECTOR(burst_length,16) then
        next_burst <= fifo_rlevel(7 downto 0);
        last_burst <= '1';
      else
        next_burst <= CONV_STD_LOGIC_VECTOR(burst_length,8);
        last_burst <= '0';
      end if;
      
    end if;
  end process burst_reader;
  
  last_data <= not OR_reduce(burst_count);
  
  burst_m2s.Cmd <= ocp_Cmd_WRITE when state = gen_write else ocp_Cmd_IDLE;
  burst_m2s.Addr <= EXT(addr,32);
  burst_m2s.BurstLength <= EXT(next_burst,ocpad_burst_width);
  burst_m2s.BurstSingleReq <= '1';
  burst_m2s.Data <= fifo_data_out;
  burst_m2s.DataByteEn <= (others => '1');
  burst_m2s.DataLast <= last_data;
  burst_m2s.DataValid <= '1' when state = wait_write else '0';
  burst_m2s.RespAccept <= '1';
  burst_m2s.Reset_n <= not rst;

  fifo_radv <= burst_s2m.DataAccept;
  
  data_fifo : afifo
    generic map(
      width  => burst_width,
      order  => fifo_order,
      wpfl   => burst_length,
      wpfl2  => 0,
      rpel   => 2,
      family => family,
      safety => false,
      fflush => true)
    port map(
      rst    => rst,
      wk     => data_clk,
      wd     => shift_reg,
      wadv   => fifo_wadv,
      we     => open,
      wpf    => fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => burst_clk,
      rq     => fifo_data_out,
      radv   => fifo_radv,
      rflush => clear_fifo,
      re     => fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => fifo_rlevel);


  
  -- Data Shifter
  shifter: process (data_clk, rst)
  begin  -- process shifter
    if rst = '1' then               -- asynchronous reset (active low)
      shift_counter <= (others => '0');
      shift_reg <= (others => '0');
      fifo_wadv <= '0';
      flushing_shift <= '0';
    elsif data_clk'event and data_clk = '1' then  -- rising clock edge
      if (data_valid = '1' or flushing_shift = '1') and fifo_nfull = '0' then
        shift_reg(burst_width-data_shift-1 downto 0) <= shift_reg(burst_width-1 downto data_shift);
        if flushing_shift = '0' then
          shift_reg(burst_width-data_shift+data_width-1 downto burst_width-data_shift) <= data;
        else
          shift_reg(burst_width-data_shift+data_width-1 downto burst_width-data_shift) <= (others => '0');
        end if;
        
        if shift_counter = CONV_STD_LOGIC_VECTOR(shift_count-1,shift_count_width) then
          shift_counter <= (others => '0');
           fifo_wadv <= '1';
        else
          shift_counter <= shift_counter+1;
          fifo_wadv <= '0';
        end if;
      else
        fifo_wadv <= '0';      
      end if;
      if clear_d = '1' then
        shift_counter <= (others => '0');
      end if;
      if flush_packet = '1' then
        if OR_reduce(shift_counter) = '1' then
          flushing_shift <= '1';
          flush_packet_delayed(0) <= '0';
        else
          flush_packet_delayed(0) <= '1';
        end if;
      elsif flushing_shift = '1' and shift_counter = CONV_STD_LOGIC_VECTOR(shift_count-1,shift_count_width) then
        flushing_shift <= '0';
        flush_packet_delayed(0) <= '1';
      else
        flush_packet_delayed(0) <= '0';
      end if;
      flush_packet_delayed(7 downto 1) <= flush_packet_delayed(6 downto 0);
    end if;
  end process shifter;

  rfd <= running_dclk and not fifo_nfull;
  
  -- Interrupt counter
  -- Generates an interrupt for every "interrupt_limit" valid samples output

  icount0: process (data_clk)
  begin  -- process icount0
    if data_clk'event and data_clk = '1' then  -- rising clock edge
      interrupt_limit_m1 <= interrupt_limit-1;
      running_dclk <= running;
      icount_active <= OR_reduce(interrupt_limit) and running_dclk;
      if icount_active = '1' and data_valid = '1' then
        if interrupt_count = interrupt_limit_m1 then
          irq <= '1';
          interrupt_count <= (others => '0');
        else
          interrupt_count <= interrupt_count+1;
          irq <= '0';
        end if;
      else
        irq <= '0';
      end if;
      if clear_d = '1' then
        interrupt_count <= (others => '0');
      end if;
    end if;
  end process icount0;
  
  
end mixed;
