--
-- data_source_2D_b128r64.vhd
-- Dataflow Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
-- Data Source 2D
--
-- Module for generating a sequence of valid data words, by reading it from a
-- memory bank located on a burst 128 profile
--
-- Allows 2D sectioning of data
-- Extra parameters:
-- Section Width      :
-- Section Height     :
-- Line Skip          : Address increment at end of each line
-- Section Skip       : Address increment (-ve) at end of each section
-- No Sections        : number of sections


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.ocpad.all;

entity data_source_2D_b128r64 is
  
  generic (
    data_width : integer := 16;
    data_shift : integer := 16;
    shift_count_width : integer := 4;
    family     : family_t := family_virtex4;
    fifo_order : integer := 9;
    use_rate_control : boolean := true;
    use_rfd : boolean := true);
  port (
    -- Interface to OCP Burst Profile
    burst_clk : in  std_logic;
    burst_m2s : out ocpad_burst128_m2sT;
    burst_s2m : in  ocpad_burst128_s2mT;
    -- Interface to OCP Register Profile
    reg_clk  : in  std_logic;    
    reg_m2s  : in  ocpad_register64_m2sT;
    reg_s2m  : out ocpad_register64_s2mT;
    -- Interface to Data Flow
    data_clk : in  std_logic;
    data : out std_logic_vector(data_width-1 downto 0);
    data_valid : out std_logic;
    rfd : in std_logic;
    -- Interrupt out
    irq : out std_logic);

end data_source_2D_b128r64;

architecture mixed of data_source_2D_b128r64 is

  constant number_of_registers  : integer := 12;
  constant burst_width : integer := 128;
  constant reg_width : integer := 64;
  constant shift_count : integer := burst_width/data_shift;
  constant burst_length : integer := 64;
  
  
  signal registers_out,registers_back : register64_array(0 to number_of_registers-1);
  signal registers_written : std_logic_vector(number_of_registers-1 downto 0);
  
  signal start_address,word_limit : std_logic_vector(31 downto 0);
  signal continuous : std_logic;
  signal running : std_logic;
  signal ctrl_written : std_logic;
  signal ctrl_start_bit : std_logic;
  signal start : std_logic;
  signal start_d : std_logic;
  signal rst : std_logic;
  
  signal interrupt_count : std_logic_vector(31 downto 0) := (others => '0');
  signal rate_count : std_logic_vector(15 downto 0) := (others => '0');
  signal interrupt_limit : std_logic_vector(31 downto 0) := (others => '0');
  signal rate_limit : std_logic_vector(15 downto 0) := (others => '0');  
  signal interrupt_limit_m1 : std_logic_vector(31 downto 0):= (others => '0');
  signal icount_active : std_logic := '0';
  signal rcount_active : std_logic := '0';
  signal rate_limit_m1 : std_logic_vector(15 downto 0) := (others => '0');  
  
  type state_type is (idle,gen_read,wait_read);
  signal state : state_type := idle;
  signal next_burst : std_Logic_vector(7 downto 0) := (others => '0');
  signal addr : std_logic_vector(31 downto 0) := (others => '0');
  signal word_count : std_logic_vector(31 downto 0) := (others => '0');

  signal burst_s2m_dv : std_logic;

  signal fifo_nfull : std_logic;
  signal fifo_empty : std_logic;
  signal fifo_data_out : std_logic_vector(burst_width-1 downto 0);
  signal fifo_radv : std_logic;
  
  signal data_ce : std_logic;
  signal rate_divide_pulse : std_logic;
  
  signal shift_counter : std_logic_vector(shift_count_width-1 downto 0)  := (others => '0');
  signal shift_reg : std_logic_vector(burst_width-1 downto 0) := (others => '0');
  
  signal dv_reg : std_logic := '0';

  component async_pulse
    port (
      clk : in std_logic;
      I   : in std_Logic;
      O   : out std_Logic);
  end component;

  signal pixel_count : std_logic_vector(15 downto 0);
  signal pixel_limit : std_Logic_vector(15 downto 0);
  signal row_count : std_logic_vector(15 downto 0);
  signal row_limit : std_Logic_vector(15 downto 0);
  signal line_skip : std_logic_vector(31 downto 0);
  signal section_skip : std_logic_vector(31 downto 0);
  signal end_skip : std_logic_vector(31 downto 0);
  signal no_sections : std_logic_vector(7 downto 0);
  signal section_count : std_logic_vector(7 downto 0);
  signal pc_active : std_logic;

  signal pixel_count_d : std_logic_vector(15 downto 0);
  signal pixel_limit_d : std_Logic_vector(15 downto 0);
  signal pixel_limit_d_m1 : std_Logic_vector(15 downto 0);
  signal row_count_d : std_logic_vector(15 downto 0);
  signal row_limit_d : std_Logic_vector(15 downto 0);
  signal row_limit_d_m1 : std_Logic_vector(15 downto 0);
  signal no_sections_d : std_logic_vector(7 downto 0);
  signal no_sections_d_m1 : std_logic_vector(7 downto 0);
  signal section_count_d : std_logic_vector(7 downto 0);

  signal skip_first_count0 : std_logic_vector(7 downto 0);
  signal skip_first_count1 : std_logic_vector(7 downto 0);
  signal skip_first_count2 : std_logic_vector(7 downto 0);
  signal skip_first_count3 : std_logic_vector(7 downto 0);
  signal skip_last_count0 : std_logic_vector(7 downto 0);
  signal skip_last_count1 : std_logic_vector(7 downto 0);
  signal skip_last_count2 : std_logic_vector(7 downto 0);
  signal skip_last_count3 : std_logic_vector(7 downto 0);
  
  signal mask_valid_start0 : std_Logic_vector(15 downto 0);
  signal mask_valid_end0 : std_Logic_vector(15 downto 0);
  signal mask_valid_start1 : std_Logic_vector(15 downto 0);
  signal mask_valid_end1 : std_Logic_vector(15 downto 0);
  signal mask_valid_start2 : std_Logic_vector(15 downto 0);
  signal mask_valid_end2 : std_Logic_vector(15 downto 0);
  signal mask_valid_start3 : std_Logic_vector(15 downto 0);
  signal mask_valid_end3 : std_Logic_vector(15 downto 0);
  signal skip_first0_is_0 : std_logic;
  signal skip_first1_is_0 : std_logic;
  signal skip_first2_is_0 : std_logic;
  signal skip_first3_is_0 : std_logic;

  signal row_valid : std_logic;
  
begin  -- mixed

  -- Register Interface
  ocp_ra0: ocp_register_interface64
    generic map (
      number_of_registers => number_of_registers)
    port map (
      ocp_clk           => reg_clk,
      ocp_m2s           => reg_m2s,
      ocp_s2m           => reg_s2m,
      registers_out     => registers_out,
      registers_back    => registers_back,
      register_written => registers_written);

  -- Resample Registers into Burst Clock Domain
  retime_regs_bclk: process (burst_clk)
  begin  -- process retime_regs_bclk
    if burst_clk'event and burst_clk = '1' then  -- rising clock edge
      start_address <= registers_out(0)(31 downto 0);
      word_limit <= registers_out(1)(31 downto 0);
      continuous <= registers_out(3)(1);
      ctrl_start_bit <= registers_out(3)(0); 
      start <= ctrl_written and ctrl_start_bit;
      pixel_limit <= registers_out(5)(15 downto 0);
      row_limit <= registers_out(6)(15 downto 0);
      no_sections <= registers_out(7)(7 downto 0);
      pc_active <= OR_reduce(pixel_limit);
      line_skip <= registers_out(8)(31 downto 0);
      section_skip <= registers_out(9)(31 downto 0);
      end_skip <= registers_out(10)(31 downto 0);
    end if;
  end process retime_regs_bclk;

  -- Test when CTRL (reg 3) is written
  ap0: async_pulse
    port map (
      clk => burst_clk,
      I   => registers_written(3),
      O   => ctrl_written);

  ap1: async_pulse
    port map (
      clk => data_clk,
      I   => start,
      O   => start_d);
  
  -- Resample Registers into Data Clock Domain
  retime_regs_dclk: process (data_clk)
  begin  -- process retime_regs_bclk
    if data_clk'event and data_clk = '1' then  -- rising clock edge
      interrupt_limit <= registers_out(2)(31 downto 0);
      rate_limit <= registers_out(4)(15 downto 0);

      pixel_limit_d <= registers_out(5)(15 downto 0);
      if shift_count = 2 then
        pixel_limit_d_m1 <= (pixel_limit_d(14 downto 0) & "0") -1;
      elsif shift_count = 4 then
        pixel_limit_d_m1 <= (pixel_limit_d(13 downto 0) & "00") -1;
      elsif shift_count = 8 then
        pixel_limit_d_m1 <= (pixel_limit_d(12 downto 0) & "000") -1;
      elsif shift_count = 16 then
        pixel_limit_d_m1 <= (pixel_limit_d(11 downto 0) & "0000") -1;               
      end if;
      row_limit_d <= registers_out(6)(15 downto 0);
      row_limit_d_m1 <= row_limit_d-1;
      no_sections_d <= registers_out(7)(7 downto 0);
      no_sections_d_m1 <= no_sections_d-1;

      skip_first_count0 <= registers_out(11)(7 downto 0);
      skip_last_count0 <= registers_out(11)(15 downto 8);
      skip_first_count1 <= registers_out(11)(23 downto 16);
      skip_last_count1 <= registers_out(11)(31 downto 24);
      skip_first_count2 <= registers_out(11)(39 downto 32);
      skip_last_count2 <= registers_out(11)(47 downto 40);
      skip_first_count3 <= registers_out(11)(55 downto 48);
      skip_last_count3 <= registers_out(11)(63 downto 56);
      skip_first0_is_0 <= not OR_reduce(skip_first_count0);
      skip_first1_is_0 <= not OR_reduce(skip_first_count1);
      skip_first2_is_0 <= not OR_reduce(skip_first_count2);
      skip_first3_is_0 <= not OR_reduce(skip_first_count3);
      mask_valid_start0 <= EXT(skip_first_count0,16)-1;
      mask_valid_start1 <= EXT(skip_first_count1,16)-1;
      mask_valid_start2 <= EXT(skip_first_count2,16)-1;
      mask_valid_start3 <= EXT(skip_first_count3,16)-1;
      mask_valid_end0 <= pixel_limit_d_m1 - EXT(skip_last_count0,16);
      mask_valid_end1 <= pixel_limit_d_m1 - EXT(skip_last_count1,16);
      mask_valid_end2 <= pixel_limit_d_m1 - EXT(skip_last_count2,16);
      mask_valid_end3 <= pixel_limit_d_m1 - EXT(skip_last_count3,16);
      
    end if;
  end process retime_regs_dclk;

  -- Make registers readable
  registers_back(0)(31 downto 0) <= registers_out(0)(31 downto 0);
  registers_back(1)(31 downto 0) <= registers_out(1)(31 downto 0);
  registers_back(2)(31 downto 0) <= registers_out(2)(31 downto 0);
  registers_back(3)(1 downto 0) <= registers_out(3)(1) & running;
  registers_back(1)(63 downto 32) <= word_count;
  registers_back(4)(15 downto 0) <= registers_out(4)(15 downto 0);
  registers_back(5)(15 downto 0) <= registers_out(5)(15 downto 0);
  registers_back(5)(31 downto 16) <= pixel_count;
  registers_back(6)(15 downto 0) <= registers_out(6)(15 downto 0);
  registers_back(6)(31 downto 16) <= row_count;
  registers_back(7)(7 downto 0) <= registers_out(7)(7 downto 0);
  registers_back(8)(31 downto 0) <= registers_out(8)(31 downto 0);
  registers_back(9)(31 downto 0) <= registers_out(9)(31 downto 0);
  registers_back(10)(31 downto 0) <= registers_out(10)(31 downto 0);
  registers_back(11) <= registers_out(11);
  
  

  
  -- Burst Read State Machine
  rst <= not reg_m2s.Reset_n;
  
  burst_reader: process (burst_clk, rst)
  begin  -- process burst_reader
    if rst = '1' then             -- asynchronous reset
      addr <= (others => '0');
      word_count <= (others => '0');
      state <= idle;
      next_burst <= CONV_STD_LOGIC_VECTOR(burst_length,8);
      running <= '0';
    elsif burst_clk'event and burst_clk = '1' then  -- rising clock edge
      if running = '0' then
        if start = '1' then
          addr <= start_address;
          word_count <= word_limit; 
          running <= '1';
          row_count <= row_limit-1;
          pixel_count <= pixel_limit;
          section_count <= no_sections-1;
        end if;
        state <= idle;
      else
        case state is
          when idle =>
            if word_count < burst_length  then
              next_burst <= word_count(7 downto 0);
            elsif pixel_count < burst_length and pc_active = '1' then
              next_burst <= pixel_count(7 downto 0);
            else
              next_burst <= CONV_STD_LOGIC_VECTOR(burst_length,8);
            end if;
            if OR_reduce(word_count) = '1' then
              if OR_reduce(pixel_count) = '1' or pc_active = '0' then
                if fifo_nfull = '0' then
                  state <= gen_read;          
                end if;
              else
                pixel_count <= pixel_limit;
                if OR_reduce(row_count) = '1' then                 
                  row_count <= row_count-1;
                  addr <= addr+line_skip;                 
                else
                  row_count <= row_limit-1;
                  if OR_reduce(section_count) = '1' then
                    section_count <= section_count-1;
                    addr <= addr+section_skip;
                  else
                    section_count <= no_sections-1;
                    addr <= addr+end_skip;
                  end if;                  
                end if;                
              end if;
            else
              if continuous = '1' then
                addr <= start_address;
                word_count <= word_limit;
              else
                running <= '0';
              end if;
            end if;
            
          when gen_read =>
            if burst_s2m.CmdAccept = '1' then
              addr <= addr + burst_length*(burst_width/8);  -- Increment
                                                            -- address by burst data width in bytes
              state <= wait_read;
              if word_count < burst_length then
                word_count <= (others => '0');
              else
                word_count <= word_count - burst_length;
              end if;

              if pixel_count < burst_length then
                pixel_count <= (others => '0');
              else
                pixel_count <= pixel_count - burst_length;
              end if;
              
            end if;
          when wait_read =>
            if burst_s2m.RespLast = '1' and burst_s2m.Resp = ocp_Resp_VALID then
              state <= idle;              
            end if;
          when others =>
            state <= idle;
        end case;
        if burst_s2m.error = '1' then
          running <= '0';
        end if;
      end if;
    end if;
  end process burst_reader;
  

  burst_m2s.Cmd <= ocp_Cmd_READ when state = gen_read else ocp_Cmd_IDLE;
  burst_m2s.Addr <= EXT(addr,32);
  burst_m2s.BurstLength <= EXT(next_burst,ocpad_burst_width);
  burst_m2s.BurstSingleReq <= '1';
  burst_m2s.Data <= (others => '0');
  burst_m2s.DataByteEn <= (others => '0');
  burst_m2s.DataLast <= '0';
  burst_m2s.DataValid <= '0';
  burst_m2s.RespAccept <= '1';
  burst_m2s.Reset_n <= not rst;


  burst_s2m_dv <= '1' when burst_s2m.Resp =  ocp_Resp_VALID else '0';
  
  data_fifo : afifo
    generic map(
      width  => burst_width,
      order  => fifo_order,
      wpfl   => burst_length,
      wpfl2  => 0,
      rpel   => 2,
      family => family,
      safety => false,
      fflush => false)
    port map(
      rst    => rst,
      wk     => burst_clk,
      wd     => burst_s2m.Data,
      wadv   => burst_s2m_dv,
      we     => open,
      wpf    => fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => data_clk,
      rq     => fifo_data_out,
      radv   => fifo_radv,
      rflush => '0',
      re     => fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  -- Generate Data Side Clock Enable
  rate_counter: process (data_clk)
  begin  -- process rate_counter
    if data_clk'event and data_clk = '1' then  -- rising clock edge
      -- Interpret Rate Limit of 0, as 1
      rcount_active <= OR_reduce(rate_limit);
      rate_limit_m1 <= rate_limit-1;
      
      if rcount_active = '1' then
        if rate_count = rate_limit_m1 then
          rate_count <= (others => '0');
          rate_divide_pulse <= '1';
        else
          rate_count <= rate_count+1;
          rate_divide_pulse <= '0';
        end if;
      else
        rate_divide_pulse <= '1';
      end if;

    end if;
  end process rate_counter;
  
  gen_always_clock: if not use_rfd and not use_rate_control generate
    data_ce <= '1';
  end generate gen_always_clock;

  gen_rfd: if use_rfd and not use_rate_control generate
    data_ce <= rfd;
  end generate gen_rfd;

  gen_rate_control: if not use_rfd and use_rate_control generate
    data_ce <= rate_divide_pulse;
  end generate gen_rate_control;

  gen_rfd_rate_control: if use_rfd and use_rate_control generate
    data_ce <= rate_divide_pulse and rfd;
  end generate gen_rfd_rate_control;

  -- Data Shifter
  shifter: process (data_clk, rst)
  begin  -- process shifter
    if rst = '1' then               -- asynchronous reset (active low)
      shift_counter <= (others => '0');
      shift_reg <= (others => '0');
      dv_reg <= '0';
      row_valid <= '1';
      pixel_count_d <= (others => '0');
      row_count_d <= (others => '0');
      section_count_d <= (others => '0');
    elsif data_clk'event and data_clk = '1' then  -- rising clock edge
      if data_ce = '1' and fifo_empty = '0' then
        if OR_reduce(shift_counter) = '0' then
          shift_reg <= fifo_data_out;
        else
          shift_reg(burst_width-data_shift-1 downto 0) <= shift_reg(burst_width-1 downto data_shift);
        end if;
        if shift_counter = CONV_STD_LOGIC_VECTOR(shift_count-1,shift_count_width) then
          shift_counter <= (others => '0');
        else
          shift_counter <= shift_counter+1;
        end if;
      end if;
      dv_reg <= data_ce and not fifo_empty;
      data <= shift_reg(data_width-1 downto 0);
      data_valid <= dv_reg and row_valid;

  
      
      if start_d = '1' then
        row_count_d <= (others => '0');
        pixel_count_d <= (others => '0');
        section_count_d <= (others => '0');
        row_valid <= skip_first0_is_0;
      else
        if dv_reg = '1' then     
          if pixel_count_d = pixel_limit_d_m1 then
            pixel_count_d <= (others => '0');
            if row_count_d = row_limit_d_m1 then
              if section_count_d = no_sections_d_m1 then
                section_count_d <= (others => '0');
                row_valid <= skip_first0_is_0;
              else
                section_count_d <= section_count_d+1;
                if section_count_d(1 downto 0) = "00" then
                  row_valid <= skip_first1_is_0;
                elsif section_count_d(1 downto 0) = "01" then
                  row_valid <= skip_first2_is_0;
                elsif section_count_d(1 downto 0) = "10" then
                  row_valid <= skip_first3_is_0;
                else
                  --section_count_d(1 downto 0) = "11" then
                  row_valid <= skip_first0_is_0;
                end if;
              end if;
              row_count_d <= (others => '0');
            else
              row_count_d <= row_count_d+1;
              if section_count_d(1 downto 0) = "00" then
                row_valid <= skip_first0_is_0;
              elsif section_count_d(1 downto 0) = "01" then
                row_valid <= skip_first1_is_0;
              elsif section_count_d(1 downto 0) = "10" then
                row_valid <= skip_first2_is_0;
              else
                --section_count_d(1 downto 0) = "11" then
                row_valid <= skip_first3_is_0;
              end if;
            end if;
          else
            pixel_count_d <= pixel_count_d+1;
            if section_count_d(1 downto 0) = "00" then
              if pixel_count_d = mask_valid_start0 then
                row_valid <= '1';
              elsif pixel_count_d = mask_valid_end0  then
                row_valid <= '0';              
              end if;            
            elsif section_count_d(1 downto 0) = "01" then
             if pixel_count_d = mask_valid_start1 then
                row_valid <= '1';
              elsif pixel_count_d = mask_valid_end1  then
                row_valid <= '0';
              end if;             
            elsif section_count_d(1 downto 0) = "10" then
             if pixel_count_d = mask_valid_start2 then
                row_valid <= '1';
              elsif pixel_count_d = mask_valid_end2  then
                row_valid <= '0';              
              end if;              
            else
              -- section_count_d(1 downto 0) = "11"
             if pixel_count_d = mask_valid_start3 then
                row_valid <= '1';
              elsif pixel_count_d = mask_valid_end3  then
                row_valid <= '0';              
              end if;                                                        
            end if;
          end if;
        end if;
      end if;

      
    end if;
  end process shifter;

  fifo_radv <= data_ce and not fifo_empty when shift_counter = CONV_STD_LOGIC_VECTOR(shift_count-1,shift_count_width) else '0';

  
  -- Interrupt counter
  -- Generates an interrupt for every "interrupt_limit" valid samples output

  icount0: process (data_clk)
  begin  -- process icount0
    if data_clk'event and data_clk = '1' then  -- rising clock edge
      interrupt_limit_m1 <= interrupt_limit-1;
      icount_active <= OR_reduce(interrupt_limit);
      if icount_active = '1' and dv_reg = '1' then
        if interrupt_count = interrupt_limit_m1 then
          irq <= '1';
          interrupt_count <= (others => '0');
        else
          interrupt_count <= interrupt_count+1;
          irq <= '0';
        end if;
      else
        irq <= '0';
      end if;
    end if;
  end process icount0;
  
  
end mixed;
