--
-- data_source_b64r32.vhd
-- Dataflow Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
-- Data Source
--
-- Module for generating a sequence of valid data words, by reading it from a
-- memory bank located on a burst 64 profile


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.ocpad.all;

entity data_source_b64r32 is
  
  generic (
    data_width : integer := 16;
    data_shift : integer := 16;
    shift_count_width : integer := 4;
    family     : family_t := family_virtex4;
    fifo_order : integer := 9;
    use_rate_control : boolean := true;
    use_rfd : boolean := true);
  port (
    -- Interface to OCP Burst Profile
    burst_clk : in  std_logic;
    burst_m2s : out ocpad_burst64_m2sT;
    burst_s2m : in  ocpad_burst64_s2mT;
    -- Interface to OCP Register Profile
    reg_clk  : in  std_logic;    
    reg_m2s  : in  ocpad_register32_m2sT;
    reg_s2m  : out ocpad_register32_s2mT;
    -- Interface to Data Flow
    data_clk : in  std_logic;
    data : out std_logic_vector(data_width-1 downto 0);
    data_valid : out std_logic;
    rfd : in std_logic;
    -- Interrupt out
    irq : out std_logic);

end data_source_b64r32;

architecture mixed of data_source_b64r32 is

  constant number_of_registers  : integer := 5;
  constant burst_width : integer := 64;
  constant reg_width : integer := 32;
  constant shift_count : integer := burst_width/data_shift;
  constant burst_length : integer := 64;
  
  
  signal registers_out,registers_back : register32_array(0 to number_of_registers-1);
  signal registers_written : std_logic_vector(number_of_registers-1 downto 0);
  
  signal start_address,word_limit : std_logic_vector(31 downto 0);
  signal continuous : std_logic;
  signal running : std_logic;
  signal ctrl_written : std_logic;
  signal ctrl_start_bit : std_logic;
  signal start : std_logic;
  signal rst : std_logic;
  
  signal interrupt_count : std_logic_vector(31 downto 0) := (others => '0');
  signal rate_count : std_logic_vector(15 downto 0) := (others => '0');
  signal interrupt_limit : std_logic_vector(31 downto 0) := (others => '0');
  signal rate_limit : std_logic_vector(15 downto 0) := (others => '0');  
  signal interrupt_limit_m1 : std_logic_vector(31 downto 0):= (others => '0');
  signal icount_active : std_logic := '0';
  signal rcount_active : std_logic := '0';
  signal rate_limit_m1 : std_logic_vector(15 downto 0) := (others => '0');  
  
  type state_type is (idle,gen_read,wait_read);
  signal state : state_type := idle;
  signal next_burst : std_Logic_vector(7 downto 0):= (others => '0');
  signal addr : std_logic_vector(31 downto 0):= (others => '0');
  signal word_count : std_logic_vector(31 downto 0):= (others => '0');

  signal burst_s2m_dv : std_logic;

  signal fifo_nfull : std_logic;
  signal fifo_empty : std_logic;
  signal fifo_data_out : std_logic_vector(burst_width-1 downto 0);
  signal fifo_radv : std_logic;
  
  signal data_ce : std_logic;
  signal rate_divide_pulse : std_logic;
  
  signal shift_counter : std_logic_vector(shift_count_width-1 downto 0):= (others => '0');
  signal shift_reg : std_logic_vector(burst_width-1 downto 0):= (others => '0');
  
  signal dv_reg : std_logic := '0';

  component async_pulse
    port (
      clk : in std_logic;
      I   : in std_Logic;
      O   : out std_Logic);
  end component;
  
begin  -- mixed

  -- Register Interface
  ocp_ra0: ocp_register_interface32
      generic map (
        number_of_registers => number_of_registers)
      port map (
        ocp_clk           => reg_clk,
        ocp_m2s           => reg_m2s,
        ocp_s2m           => reg_s2m,
        registers_out     => registers_out,
        registers_back    => registers_back,
        register_written => registers_written);

  -- Resample Registers into Burst Clock Domain
  retime_regs_bclk: process (burst_clk)
  begin  -- process retime_regs_bclk
    if burst_clk'event and burst_clk = '1' then  -- rising clock edge
      start_address <= registers_out(0)(31 downto 0);
      word_limit <= registers_out(1)(31 downto 0);
      continuous <= registers_out(3)(1);
      ctrl_start_bit <= registers_out(3)(0); 
      start <= ctrl_written and ctrl_start_bit;
    end if;
  end process retime_regs_bclk;

  -- Test when CTRL (reg 3) is written
  ap0: async_pulse
    port map (
      clk => burst_clk,
      I   => registers_written(3),
      O   => ctrl_written);

  -- Resample Registers into Data Clock Domain
  retime_regs_dclk: process (data_clk)
  begin  -- process retime_regs_bclk
    if data_clk'event and data_clk = '1' then  -- rising clock edge
      interrupt_limit <= registers_out(2)(31 downto 0);
      rate_limit <= registers_out(4)(15 downto 0);
    end if;
  end process retime_regs_dclk;

  -- Make registers readable
  registers_back(0) <= registers_out(0);
  registers_back(1) <= registers_out(1);
  registers_back(2) <= registers_out(2);
  registers_back(3) <= EXT(registers_out(3)(1) & running,reg_width); 
  registers_back(4) <= registers_out(4);
  


    
  -- Burst Read State Machine
  rst <= not reg_m2s.Reset_n;
    
  burst_reader: process (burst_clk, rst)
  begin  -- process burst_reader
    if rst = '1' then             -- asynchronous reset
      addr <= (others => '0');
      word_count <= (others => '0');
      state <= idle;
      next_burst <= CONV_STD_LOGIC_VECTOR(burst_length,8);
      running <= '0';
    elsif burst_clk'event and burst_clk = '1' then  -- rising clock edge
      if running = '0' then
        if start = '1' then
          addr <= start_address;
          word_count <= word_limit; 
          running <= '1';         
        end if;
        state <= idle;
      else
        case state is
          when idle =>
            if word_count < burst_length  then
              next_burst <= word_count(7 downto 0);
            else
              next_burst <= CONV_STD_LOGIC_VECTOR(burst_length,8);
            end if;
            if OR_reduce(word_count) = '1' then
              if fifo_nfull = '0' then
                state <= gen_read;          
              end if;
            else
              if continuous = '1' then
                addr <= start_address;
                word_count <= word_limit;
              else
                running <= '0';
              end if;
            end if;
 
          when gen_read =>
            if burst_s2m.CmdAccept = '1' then
              addr <= addr + burst_length*(burst_width/8);  -- Increment
                                                            -- address by burst data width in bytes
              state <= wait_read;
              if word_count < burst_length then
                word_count <= (others => '0');
              else
                word_count <= word_count - burst_length;  -- purpose: 

              end if; 
            end if;
          when wait_read =>
            if burst_s2m.RespLast = '1' and burst_s2m.Resp = ocp_Resp_VALID then
              state <= idle;
            end if;
          when others =>
            state <= idle;
        end case;
        if burst_s2m.error = '1' then
          running <= '0';
        end if;
      end if;
    end if;
  end process burst_reader;
  

  burst_m2s.Cmd <= ocp_Cmd_READ when state = gen_read else ocp_Cmd_IDLE;
  burst_m2s.Addr <= EXT(addr,32);
  burst_m2s.BurstLength <= EXT(next_burst,ocpad_burst_width);
  burst_m2s.BurstSingleReq <= '1';
  burst_m2s.Data <= (others => '0');
  burst_m2s.DataByteEn <= (others => '0');
  burst_m2s.DataLast <= '0';
  burst_m2s.DataValid <= '0';
  burst_m2s.RespAccept <= '1';
  burst_m2s.Reset_n <= not rst;


  burst_s2m_dv <= '1' when burst_s2m.Resp =  ocp_Resp_VALID else '0';
  
  data_fifo : afifo
    generic map(
      width  => burst_width,
      order  => fifo_order,
      wpfl   => burst_length,
      wpfl2  => 0,
      rpel   => 2,
      family => family,
      safety => false,
      fflush => false)
    port map(
      rst    => rst,
      wk     => burst_clk,
      wd     => burst_s2m.Data,
      wadv   => burst_s2m_dv,
      we     => open,
      wpf    => fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => data_clk,
      rq     => fifo_data_out,
      radv   => fifo_radv,
      rflush => '0',
      re     => fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  -- Generate Data Side Clock Enable
  rate_counter: process (data_clk)
  begin  -- process rate_counter
    if data_clk'event and data_clk = '1' then  -- rising clock edge
      -- Interpret Rate Limit of 0, as 1
      rcount_active <= OR_reduce(rate_limit);
      rate_limit_m1 <= rate_limit-1;
     
      if rcount_active = '1' then
        if rate_count = rate_limit_m1 then
          rate_count <= (others => '0');
          rate_divide_pulse <= '1';
        else
          rate_count <= rate_count+1;
          rate_divide_pulse <= '0';
        end if;
      else
        rate_divide_pulse <= '1';
      end if;

    end if;
  end process rate_counter;
  
  gen_always_clock: if not use_rfd and not use_rate_control generate
    data_ce <= '1';
  end generate gen_always_clock;

  gen_rfd: if use_rfd and not use_rate_control generate
    data_ce <= rfd;
  end generate gen_rfd;

  gen_rate_control: if not use_rfd and use_rate_control generate
    data_ce <= rate_divide_pulse;
  end generate gen_rate_control;

  gen_rfd_rate_control: if use_rfd and use_rate_control generate
    data_ce <= rate_divide_pulse and rfd;
  end generate gen_rfd_rate_control;

  -- Data Shifter
  shifter: process (data_clk, rst)
  begin  -- process shifter
    if rst = '1' then               -- asynchronous reset (active low)
      shift_counter <= (others => '0');
      shift_reg <= (others => '0');
      dv_reg <= '0';
    elsif data_clk'event and data_clk = '1' then  -- rising clock edge
      if data_ce = '1' and fifo_empty = '0' then
        if OR_reduce(shift_counter) = '0' then
          shift_reg <= fifo_data_out;
        else
          shift_reg(burst_width-data_shift-1 downto 0) <= shift_reg(burst_width-1 downto data_shift);
        end if;
        if shift_counter = CONV_STD_LOGIC_VECTOR(shift_count-1,shift_count_width) then
          shift_counter <= (others => '0');
        else
          shift_counter <= shift_counter+1;
        end if;
      end if;
      dv_reg <= data_ce and not fifo_empty;
      data <= shift_reg(data_width-1 downto 0);
      data_valid <= dv_reg;
    end if;
  end process shifter;

  fifo_radv <= data_ce and not fifo_empty when shift_counter = CONV_STD_LOGIC_VECTOR(shift_count-1,shift_count_width) else '0';

  
  -- Interrupt counter
  -- Generates an interrupt for every "interrupt_limit" valid samples output

  icount0: process (data_clk)
  begin  -- process icount0
    if data_clk'event and data_clk = '1' then  -- rising clock edge
      interrupt_limit_m1 <= interrupt_limit-1;
      icount_active <= OR_reduce(interrupt_limit);
      if icount_active = '1' and dv_reg = '1' then
        if interrupt_count = interrupt_limit_m1 then
          irq <= '1';
          interrupt_count <= (others => '0');
        else
          interrupt_count <= interrupt_count+1;
          irq <= '0';
        end if;
      else
        irq <= '0';
      end if;
    end if;
  end process icount0;
  
  
end mixed;
