/*
** data_capture.c 
**
** Example using ADPKGOCP Bridga and Data Sink to
** capture data in an SDRAM buffer and transfer it to host memory
**
** Single threaded example, so can lock up if data clock rate
** exceeds PCI bandwidth
**
**
** Example for:
**
**   o ADM-XRC-4LX
**   o ADM-XRC-4FX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2
**
*/

#include <windows.h>
#include <stdio.h>

#include <admxrc2.h>

#include "common.h"
#include "lb2ocp.h"
#include "ocpsink.h"


void check_data(void *buffer, int size)
{
  /* Checks captured data against test pattern,
  assumes that data source is connected to incrementing 16 bit counter */
  int errs =0;
  short *buf = buffer;
  short act,exp;
  int j;
  
  exp= buf[0];
  for (j = 1; j < size; j++) {
    exp++;
    act = buf[j];
    if (act != exp) {
      if (errs<32)
        printf("Check data : error at index %ld, exp/act=0x%8.8lx/0x%8.8lx\n",
        (unsigned long) j,
       (unsigned long) exp, (unsigned long) act);
      exp = act;
      errs++;
    }
  }  
  
}

void capture_data(ADMXRC2_HANDLE card,
                  ADMXRC2_DMADESC bufHandle,
                  void *buffer,
                  PLB2OCP bridge,	
                  POCPSINK sink, 
                  int chunk_size, 
                  int nchunks)
{
  HANDLE			event = NULL;
  int i;

  event = CreateEvent(NULL, FALSE, FALSE,	NULL);

  ADMXRC2_RegisterInterruptEvent(card, event);

  LB2OCP_EnableInterrupts(bridge,sink->irq_mask);
  LB2OCP_ClearInterrupts(bridge,sink->irq_mask);
  
  printf("%8.8x\n",bridge->fpgaSpace[LB2OCPB_ISR]);
 
  if (OCPSINK_StartCapture(sink,
                       0,                        // Start Address
                       nchunks*chunk_size/16,    // Number of 128bit words to capture
                       chunk_size/2,             // Generate 1 IRQ per chunk
                       0))                       // Single Shot Mode
                       {  
  for (i=0;i<nchunks;i++)
    {
      
      
      WaitForSingleObject(event, INFINITE);
      printf("%8.8x\n",bridge->fpgaSpace[LB2OCPB_ISR]);
      printf("Captured section %d\n",i);
      LB2OCP_ClearInterrupts(bridge,sink->irq_mask);
      LB2OCP_TranferDataToHost(bridge, 
                       sink->burst_engine,
                       sink->burst_target,
                       chunk_size,
                       i*chunk_size,
                       0,
                       bufHandle); 
      printf("Read back section %d\n",i); 
      printf("%8.8x\n",bridge->fpgaSpace[LB2OCPB_ISR]);
      check_data(buffer,chunk_size/2);
      printf("Checked data\n"); 
      
    }
  }
  else
  printf("Failed to start data capture\n");
  
  
  ADMXRC2_UnregisterInterruptEvent(card, event);
  CloseHandle(event);
}

int main(int argc, char *argv[])
{
	int ret;
	ADMXRC2_STATUS		status;
	ADMXRC2_CARD_INFO	cardInfo;
	ADMXRC2_SPACE_INFO	spaceInfo;
	ADMXRC2_HANDLE		card = NULL;
	int recvbufValid= 0;
	ADMXRC2_DMADESC recvbufHandle;
	int bufSize;
	
	char *recvbuf,*sendbuf;
	
	int					i;
	DWORD				start;
	HANDLE				event = NULL;
	const char*			filename;
	Arguments			arguments;
	Option				options[7];
	float				clock_freq;
        float				mclock_freq;
	BOOLEAN				use64bit;
	double	s, g, t;
	DWORD	now, delta1,delta2;
	
	unsigned int chunk_size = 1024*1024;
	unsigned int nchunks = 1;
	PLB2OCP bridge;
	POCPSINK sink;
  int system_status;

	options[0].key = "card";
	options[0].type = Option_uint;
	options[0].def.uintVal = 0;
	options[0].meaning = "ID of card to use";
	options[1].key = "index";
	options[1].type = Option_uint;
	options[1].def.uintVal = 0;
	options[1].meaning = "index of card to use";
	options[2].key = "size";
	options[2].type = Option_uint;
	options[2].def.uintVal = 1024*1024;
	options[2].meaning = "send/receive buffer size (bytes)";
	options[3].key = "lclk";
	options[3].type = Option_float;
	options[3].def.floatVal = 33.0;
	options[3].meaning = "local bus clock frequency (MHz)";
  options[4].key = "mclk";
	options[4].type = Option_float;
	options[4].def.floatVal = 200.0;
	options[4].meaning = "memory clock frequency (MHz)";
	options[5].key = "bitstream";
	options[5].type = Option_string;
	options[5].def.stringVal = "data_capture.bit";
	options[5].meaning = "Bitstream to configure FPGA";
  options[6].key = "chunks";
	options[6].type = Option_uint;
	options[6].def.uintVal = 1;
	options[6].meaning = "Number of buffers of data to receive";
	arguments.nOption = 7;
	arguments.option = options;
	arguments.nParamType = 0;
	arguments.minNParam = 0;

	parse_command_line(
		"dma",
		argc,
		argv,
		&arguments);

	if (options[0].specified && options[1].specified) {
		printf("Do not specify both /card and /index\n");
		ret = -1;
		goto done;
	}

	if (!options[1].specified) {
		ADMXRC2_CARDID cardID = options[0].value.uintVal;

		status = ADMXRC2_OpenCard(cardID, &card);
		if (status != ADMXRC2_SUCCESS) {
			printf("Failed to open card with ID %ld: %s\n",
				(unsigned long) cardID, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	} else {
		unsigned int index = options[1].value.uintVal;

		status = ADMXRC2_OpenCardByIndex(index, &card);
		if (status != ADMXRC2_SUCCESS) {
			printf("Failed to open card with index %ld: %s\n",
				(unsigned long) index, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	}

	/* Allocate memory for the DMA data buffers */
	bufSize = options[2].value.uintVal;
	chunk_size = bufSize;
  nchunks = options[6].value.uintVal;
	recvbuf = (BYTE*) malloc(bufSize  + bufSize );
	if (recvbuf == NULL) {
		printf("Failed to allocate buffers for threads\n");
		ret = -1;
		goto done;
	}
	sendbuf = recvbuf + bufSize;

	clock_freq = options[3].value.floatVal * 1.0e6f;

  mclock_freq = options[4].value.floatVal * 1.0e6f;
       

	/* Get card information, for device fitted */
	status = ADMXRC2_GetCardInfo(card, &cardInfo);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	switch (cardInfo.BoardType) {

	case ADMXRC2_BOARD_ADMXRC4LX:
	    use64bit = 0;
	  break;
      case ADMXRC2_BOARD_ADMXRC4FX:
	    case ADMXRC2_BOARD_ADMXRC5T1:
	    case ADMXRC2_BOARD_ADMXRC5T2:
	      use64bit = 1;
		  break;
	  default:
	    printf("This example must be run on one of the following cards:\n\n");	
		  printf("\tADM-XRC-4LX\n");
      printf("\tADM-XRC-4FX\n");
		  printf("\tADM-XRC-5T1\n");
		  printf("\tADM-XRC-5T2\n");
		  ret = -1;
		  goto done;
	  }

	status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, clock_freq, NULL);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to set LCLK to %.1fMHz: %s\n",
			clock_freq / 1.0e6f, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
	status = ADMXRC2_SetClockRate(card, 1, mclock_freq, NULL);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to set MCLK to %.1fMHz: %s\n",
			mclock_freq / 1.0e6f, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	filename = options[5].value.stringVal;

	status = ADMXRC2_ConfigureFromFile(card, filename);
	if (status != ADMXRC2_SUCCESS) {
		printf ("Failed to load the bitstream '%s': %s\n",
			filename, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	/* Get a DMA descriptor for the buffer used for DMA transfers from the card */
	status = ADMXRC2_SetupDMA(card, recvbuf, bufSize, 0, &recvbufHandle);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to set up receive DMA buffer: %s\n", ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
	recvbufValid = TRUE;

  bridge = LB2OCP_Create(card,use64bit,use64bit);
  LB2OCP_SetNumberDMAEngines(bridge,1);       
  sink = OCPSINK_Create(bridge,
                        0, // Burst Engine
                        0, // Burst Target
                        1, // Register Bank
                        0, // Offset in register bank
                        1, // IRQ Mask
                        use64bit); // Data sink reg profile width
             // Pulse Reset
      
  Sleep(1000);          
             
	system_status = LB2OCP_Status(bridge);       
  printf("System Status : %x\n",system_status);
  
  if ((system_status&1)==0)
  {
    printf("Local Bus Clock DCM not locked ... exiting\n");
    goto done;
  }
  if ((system_status&2)==0)
  {
    printf("Memory Clock DCM not locked ... exiting\n");
    goto done;
  }
  if ((system_status&4)==0)
  {
    printf("IDELAY Control not ready ... exiting\n");
    goto done;
  }
  if ((system_status&16)==0)
  {
    printf("Memory Controller not online ... exiting\n");
    goto done;
  }
  
	printf("Ready to capture data.  Press enter to begin\n");
	getchar();
  start = GetTickCount();
  
	capture_data(
	             card,
	             recvbufHandle,
	             recvbuf,
	             bridge,
	             sink,
	             chunk_size,
	             nchunks);
  
	now = GetTickCount();
  delta1 = now - start;
      
  g = (double) (nchunks*chunk_size);   	
	g = g / 1048576.0f;
 
	printf("Received %.2fMB\n",
		g);
	
	printf("Receive rate %.2fMB/s\n",		
			g * 1000.0 / ((double) delta1));
        
  
done:
	/* Free the DMA descriptors */
	
	if (recvbufValid) {
		status = ADMXRC2_UnsetupDMA(card, recvbufHandle);
		if (status != ADMXRC2_SUCCESS) {
			printf("Failed to delete DMA receive buffer: %s\n", ADMXRC2_GetStatusString(status));
		}
	}



	/* Free the Win32 Event */
	if (event != NULL) {
		CloseHandle(event);
	}

	/* Close the device */
	if (card != NULL) {
		ADMXRC2_CloseCard(card);
	}
	
	/* Free the memory used for the DMA data buffers */
	if (recvbuf != NULL) {
		free(recvbuf);
	}

	return ret;
}
