--
-- data_capture-xrc5t1.vhd 
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

library unisim;
use unisim.vcomponents.all;

library work;
use work.memif.all;
use work.ocpad.all;
use work.fifo.all;
use work.localbus_if_types.all;
use work.ocp_ddma_components.all;
use work.ocpad_dataflow_components.all;
use work.ocpad_components.all;

entity data_capture is
  generic(
    -- Memory bank parameters
    bank0             : in    bank_t := (enable => true, ra_width => 21, rc_width => 8, rd_width => 32);
    num_ramclk        : in    natural := 1);
  port(
    -- Local bus
    lclk              : in    std_logic;
    lreset_l          : in    std_logic; 
    lwrite            : in    std_logic;
    lads_l            : in    std_logic;
    lblast_l          : in    std_logic;
    lbe_l             : in    std_logic_vector(3 downto 0);
    la                : in    std_logic_vector(23 downto 2);
    ld                : inout std_logic_vector(31 downto 0);
    lbterm_l          : inout std_logic;
    lready_l          : inout std_logic; 
    fholda            : in    std_logic;
    ldreq_l  : out   std_logic_vector(1 downto 0);
    ldack_l  : in    std_logic_vector(1 downto 0);
    finti_l  : out   std_logic;
    -- General purpose clock
    mclk_n            : in    std_logic;
    mclk_p            : in    std_logic;
    -- Reference clock
    refclk_n          : in    std_logic;
    refclk_p          : in    std_logic;
    -- Memory banks 
    ra0      : out   std_logic_vector(bank0.ra_width - 1 downto 0);
    rc0      : inout std_logic_vector(bank0.rc_width - 1 downto 0);
    rd0      : inout std_logic_vector(bank0.rd_width - 1 downto 0);
    ramclki           : in    std_logic_vector(num_ramclk - 1 downto 0);
    ramclko           : out   std_logic_vector(num_ramclk - 1 downto 0);
    -- External Inputs - Assuming 8 bit A2D with Clock
    dclk     : in    std_logic;
    data     : in    std_logic_vector(7 downto 0));

end entity;

architecture struct of data_capture is

  -- Set this flag false to use external data input and clock
  constant use_dummy_internal_data : boolean := true;
  
  constant number_of_ocp_nonburst         : integer := 1;
  constant number_of_ddma_engines         : integer := 1;
  constant number_of_ocp_burst_per_engine : integer := 1;

  
  signal rst          : std_logic;
  signal clk          : std_logic;
  signal clk_locked   : std_logic;
  signal mclk_ibufg  : std_logic;
  signal refclk_ibufg : std_logic;
  signal refclk_bufg  : std_logic;



 

  signal localbus_inputs  : localbus_inputs_type;
  signal localbus_outputs : localbus_outputs_type;

  signal ocp_m2s_nonburst_array : ocpad_register32_m2sT_array(0 to number_of_ocp_nonburst-1);
  signal ocp_s2m_nonburst_array : ocpad_register32_s2mT_array(0 to number_of_ocp_nonburst-1);
  signal ocp_m2s_burst_array    : ocpad_burst32_m2sT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);
  signal ocp_s2m_burst_array    : ocpad_burst32_s2mT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);

  signal system_status : std_logic_vector(31 downto 0);

  signal ocp_m2s_ddr1   : ocpad_burst32_m2sT;
  signal ocp_s2m_ddr1   : ocpad_burst32_s2mT;
  signal sink_burst_m2s : ocpad_burst32_m2sT;
  signal sink_burst_s2m : ocpad_burst32_s2mT;


  component lclk_dcm
    port(
      rst    : in  std_logic;
      refclk : in  std_logic;
      lclk   : in  std_logic;
      clk    : out std_logic;
      locked : out std_logic);
  end component;

  --
  -- If the synthesizer replicates an asynchronous reset signal due high fanout,
  -- this can prevent flip-flops being mapped into IOBs. We set the maximum
  -- fanout for such nets to a high enough value that replication never occurs.
  --
  attribute MAX_FANOUT        : string;
  attribute MAX_FANOUT of rst : signal is "1000000";

  
   component memclk
      generic(
        num_clock     : in    natural);
      port(
        rst           : in    std_logic;
        mclk_i        : in    std_logic;
        memclk0       : out   std_logic;
        memclk90      : out   std_logic;
        memclk180     : out   std_logic;
        memclk270     : out   std_logic;
        ramclko       : out   std_logic_vector(num_clock - 1 downto 0);
        ramclki       : in    std_logic_vector(num_clock - 1 downto 0);
        locked        : out   std_logic_vector(1 downto 0);
        memrst        : out   std_logic);
    end component;

    component ocp2memif32
      generic (
        use_chipscope : boolean := false;
        addr_width    : integer);
      port (
        clk                     : in  std_logic;
        -- OCP Port
        ocp_m2s                 : in ocpad_burst32_m2sT;
        ocp_s2m                 : out ocpad_burst32_s2mT;
        -- MemIF Port (32 bits wide)
        sr                      : out std_logic;
        ce                      : out std_logic;
        tag                     : out std_logic_vector(1 downto 0);
        w                       : out std_logic;
        a                       : out std_logic_vector(addr_width-1 downto 0);
        d                       : out std_logic_vector(31 downto 0);
        be                      : out std_logic_vector(3 downto 0);
        q                       : in  std_logic_vector(31 downto 0);
        qtag                    : in  std_logic_vector(1 downto 0);
        valid                   : in  std_logic;
        ready                   : in  std_logic);
    end component;
  

 
  
  constant port_width : natural := 32;
  constant port_be_width : natural := port_width / 8;
  constant port_addr_width : natural := 24;

  signal w0     : std_logic;
  signal ce0    : std_logic;
  signal r0     : std_logic;
  signal be0    : std_logic_vector(port_width/8-1 downto 0);
  signal a0     : std_logic_vector(port_addr_width-1 downto 0);
  signal d0     : std_logic_vector(port_width-1 downto 0);
  signal q0     : std_logic_vector(port_width-1 downto 0);
  signal valid0 : std_logic;
  signal ready0 : std_logic;
  signal tag0   : std_logic_vector(1 downto 0);
  signal qtag0  : std_logic_vector(1 downto 0);

  signal memclk0 : std_logic;
  signal memclk90 : std_logic;
  signal memclk180 : std_logic;
  signal memclk270 : std_logic;
  signal memrst : std_logic;


  signal locked : std_logic_vector(4 downto 0);

  signal mode_reg : std_logic_vector(31 downto 0);


  signal ocpmp_status0 : std_logic_vector(3 downto 0);
  signal system_irq    : std_logic_vector(31 downto 0) := (others => '0');
  signal irq           : std_logic;
  

  signal   dclk_i     :    std_logic;
  signal   data_i     :    std_logic_vector(7 downto 0);
  
begin

   ibufg_mclk : IBUFGDS
        port map(
            I  => mclk_p,
            IB => mclk_n,
            O  => mclk_ibufg);
    
    ibufg_refclk : IBUFGDS
        port map(
            I  => refclk_p,
            IB => refclk_n,
            O  => refclk_ibufg);
    
    bufg_refclk : BUFG
        port map(
            I  => refclk_ibufg,
            O  => refclk_bufg);
    
    lclk_dcm_0 : lclk_dcm
        port map(
            rst => rst,
            refclk => refclk_bufg,
            lclk => lclk,
            clk => clk,
            locked => clk_locked);


   memclk_0 : memclk
     generic map(
       num_clock => num_ramclk)
     port map(
       rst => rst,
       mclk_i => mclk_ibufg,
       memclk0 => memclk0,
       memclk90 => memclk90,
            memclk180 => memclk180,
            memclk270 => memclk270,
            ramclki => ramclki,
            ramclko => ramclko,
            locked => locked(1 downto 0),
            memrst => memrst);

  --
  -- Instantiate SDK memory modules
  --
      sram_port_0 : zbtsram_port
        generic map(
            pinout => zbtsram_pinout_admxrc4lx,
            ra_width => bank0.ra_width,
            rc_width => bank0.rc_width,
            rd_width => bank0.rd_width,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => 2)
        port map(
            clk0 => memclk0,
            rst => memrst,
            sr => '0',
            ce => ce0,
            w => w0,
            tag => tag0,
            a => a0,
            d => d0,
            be => be0,
            q => q0,
            qtag => qtag0,
            valid => valid0,
            ready => ready0,
            pipeline => '1',
            ra => ra0,
            rc => rc0,
            rd => rd0);

 
  --
  -- Instantiate OCP bridge to memif ports
  --
  
  sram_brigde0 : ocp2memif32
    generic map (
      use_chipscope => false,
      addr_width => port_addr_width)
    port map (
      clk     => memclk0,
      ocp_m2s => ocp_m2s_ddr1,
      ocp_s2m => ocp_s2m_ddr1,
      sr      => open,
      ce      => ce0,
      tag     => tag0,
      w       => w0,
      a       => a0,
      d       => d0,
      be      => be0,
      q       => q0,
      qtag    => qtag0,
      valid   => valid0,
      ready   => ready0);

  
  mp_ocp1 : MULTIPORT_OCP_IF2_b32
    generic map (
      round_robin_arb => true)
    port map (
      ocp_m2s_memif => ocp_m2s_ddr1,
      ocp_s2m_memif => ocp_s2m_ddr1,
      ocp_m2s0      => sink_burst_m2s,
      ocp_s2m0      => sink_burst_s2m,
      ocp_m2s1      => ocp_m2s_burst_array(0),
      ocp_s2m1      => ocp_s2m_burst_array(0),
      status        => ocpmp_status0,
      clk           => memclk0
      );


   lbio0 : localbus_io_plx
    port map (
      localbus_inputs  => localbus_inputs,
      localbus_outputs => localbus_outputs,
      lreset_l         => lreset_l,
      lwrite           => lwrite,
      lads_l           => lads_l,
      lblast_l         => lblast_l,
      lbterm_l         => lbterm_l,
      ld               => ld,
      la               => la,
      lready_l         => lready_l,
      lbe_l            => lbe_l,
      fholda           => fholda,
      ldreq_l          => ldreq_l,
      ldack_l          => ldack_l);


  lb_ddma_if0 : localbus2ocp_bridge_r32b32
    generic map (
      la_top                         => 23,
      family                         => family_virtex4,
      number_of_ocp_nonburst         => number_of_ocp_nonburst,
      number_of_ddma_engines         => number_of_ddma_engines,
      number_of_ocp_burst_per_engine => number_of_ocp_burst_per_engine)
    port map (
      lclk                           => clk,
      ocp_clk_b                      => memclk0,
      ocp_clk_nb                     => clk,
      rst                            => rst,
      localbus_inputs                => localbus_inputs,
      localbus_outputs               => localbus_outputs,
      ocp_nonburst_bus_outputs_array => ocp_m2s_nonburst_array,
      ocp_nonburst_bus_inputs_array  => ocp_s2m_nonburst_array,
      ocp_burst_bus_outputs_array    => ocp_m2s_burst_array,
      ocp_burst_bus_inputs_array     => ocp_s2m_burst_array,
      system_status                  => system_status,
      system_irq                     => system_irq,
      irq                            => irq);

  finti_l <= not irq;

  sink0 : data_sink_b32r32
    generic map (
      data_width => 8,
      data_shift => 8)
    port map (
      burst_clk    => memclk0,
      burst_m2s    => sink_burst_m2s,
      burst_s2m    => sink_burst_s2m,
      reg_clk      => clk,
      reg_m2s      => ocp_m2s_nonburst_array(0),
      reg_s2m      => ocp_s2m_nonburst_array(0),
      data_clk     => dclk_i,
      data         => data_i,
      data_valid   => '1',
      rfd          => open,
      flush_packet => '0',
      irq          => system_irq(0));

  
 


  system_status <= EXT(ocpmp_status0 & locked & clk_locked, 32);

  map_data_in_signals: if not use_dummy_internal_data generate
    dclk_i <= dclk;
    data_i <= data;
  end generate map_data_in_signals;

  gen_dummy_data: if use_dummy_internal_data generate
    dclk_i <= clk;
    process (clk,rst)
    begin  -- process
      if rst = '1' then         
        data_i <= (others => '0');
      elsif clk'event and clk = '1' then  
        data_i <= data_i + 1;
      end if;
    end process;
  end generate gen_dummy_data;
  
end architecture;
