--
-- data_capture-xrc5t1.vhd 
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

library unisim;
use unisim.vcomponents.all;

library work;
use work.memif.all;
use work.ocpad.all;
use work.fifo.all;
use work.localbus_if_types.all;
use work.ocp_ddma_components.all;
use work.ocpad_dataflow_components.all;
use work.ocpad_components.all;

entity data_capture is
  generic(
    -- Memory bank parameters
    bank0 : in bank_t := (enable => true, ra_width => 16, rc_width => 22, rd_width => 32));
  port(
    -- Local bus
    lclk     : in    std_logic;
    lreset_l : in    std_logic;
    lwrite   : in    std_logic;
    lads_l   : in    std_logic;
    lblast_l : in    std_logic;
    lbe_l    : in    std_logic_vector(7 downto 0);
    lad      : inout std_logic_vector(63 downto 0);
    lbterm_l : inout std_logic;
    lready_l : inout std_logic;
    fholda   : in    std_logic;
    ldreq_l  : out   std_logic_vector(3 downto 0);
    ldack_l  : in    std_logic_vector(3 downto 0);
    finti_l  : out   std_logic;
    -- General purpose clock
    mclka_n  : in    std_logic;
    mclka_p  : in    std_logic;
    -- Reference clock
    refclk_n : in    std_logic;
    refclk_p : in    std_logic;
    -- Memory banks (up to 16 supported by this design)
    ra0      : out   std_logic_vector(bank0.ra_width - 1 downto 0);
    rc0      : inout std_logic_vector(bank0.rc_width - 1 downto 0);
    rd0      : inout std_logic_vector(bank0.rd_width - 1 downto 0);
    -- External Inputs - Assuming 16 bit A2D with Clock
    dclk     : in    std_logic;
    data     : in    std_logic_vector(15 downto 0));

end entity;

architecture struct of data_capture is

  -- Set this flag false to use external data input and clock
  constant use_dummy_internal_data : boolean := true;
  
  constant number_of_ocp_nonburst         : integer := 1;
  constant number_of_ddma_engines         : integer := 1;
  constant number_of_ocp_burst_per_engine : integer := 1;

  
  signal rst          : std_logic;
  signal clk          : std_logic;
  signal clk_locked   : std_logic;
  signal mclka_i      : std_logic;
  signal mclka_ibufg  : std_logic;
  signal mclka_bufg   : std_logic;
  signal refclk_ibufg : std_logic;
  signal refclk_bufg  : std_logic;



 

  signal localbus_inputs  : localbus_inputs_type;
  signal localbus_outputs : localbus_outputs_type;

  signal ocp_m2s_nonburst_array : ocpad_register64_m2sT_array(0 to number_of_ocp_nonburst-1);
  signal ocp_s2m_nonburst_array : ocpad_register64_s2mT_array(0 to number_of_ocp_nonburst-1);
  signal ocp_m2s_burst_array    : ocpad_burst128_m2sT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);
  signal ocp_s2m_burst_array    : ocpad_burst128_s2mT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);

  signal system_status : std_logic_vector(31 downto 0);

  signal ocp_m2s_ddr1   : ocpad_burst128_m2sT;
  signal ocp_s2m_ddr1   : ocpad_burst128_s2mT;
  signal sink_burst_m2s : ocpad_burst128_m2sT;
  signal sink_burst_s2m : ocpad_burst128_s2mT;


  component mclk_gtp_wrap
    port(
      clkin  : in  std_logic;
      clkout : out std_logic);
  end component;

  component lclk_dcm
    port(
      rst    : in  std_logic;
      refclk : in  std_logic;
      lclk   : in  std_logic;
      clk    : out std_logic;
      locked : out std_logic);
  end component;

  --
  -- If the synthesizer replicates an asynchronous reset signal due high fanout,
  -- this can prevent flip-flops being mapped into IOBs. We set the maximum
  -- fanout for such nets to a high enough value that replication never occurs.
  --
  attribute MAX_FANOUT        : string;
  attribute MAX_FANOUT of rst : signal is "1000000";

  component ocp2memif128
    generic (
      use_chipscope : boolean := false);
    port (
      clk     : in  std_logic;
      -- OCP Port
      ocp_m2s : in  ocpad_burst128_m2sT;
      ocp_s2m : out ocpad_burst128_s2mT;
      -- MemIF Port (128 bits wide)
      sr      : out std_logic;
      ce      : out std_logic;
      tag     : out std_logic_vector(1 downto 0);
      w       : out std_logic;
      a       : out std_logic_vector(27 downto 0);
      d       : out std_logic_vector(127 downto 0);
      be      : out std_logic_vector(15 downto 0);
      q       : in  std_logic_vector(127 downto 0);
      qtag    : in  std_logic_vector(1 downto 0);
      valid   : in  std_logic;
      ready   : in  std_logic);
  end component;

  component memclk
    port (
      mclk          : in  std_logic;
      rst           : in  std_logic;
      refclk200     : in  std_logic;
      sel_refclk220 : in  std_logic;
      clk0          : out std_logic;
      clk45         : out std_logic;
      clk2x0        : out std_logic;
      clk2x90       : out std_logic;
      locked        : out std_logic_vector(2 downto 0);
      memrst        : out std_logic);
  end component;


  constant port_addr_width_dram : integer := 28;
  constant port_width_dram      : integer := 128;


  signal w0     : std_logic;
  signal ce0    : std_logic;
  signal r0     : std_logic;
  signal be0    : std_logic_vector(port_width_dram/8-1 downto 0);
  signal a0     : std_logic_vector(port_addr_width_dram-1 downto 0);
  signal d0     : std_logic_vector(port_width_dram-1 downto 0);
  signal q0     : std_logic_vector(port_width_dram-1 downto 0);
  signal valid0 : std_logic;
  signal ready0 : std_logic;
  signal tag0   : std_logic_vector(1 downto 0);
  signal qtag0  : std_logic_vector(1 downto 0);

  signal memclk0    : std_logic;
  signal memclk45   : std_logic;
  signal memclk2x0  : std_logic;
  signal memclk2x90 : std_logic;
  signal memrst     : std_logic;

  signal locked : std_logic_vector(4 downto 0);

  signal mode_reg : std_logic_vector(31 downto 0);


  signal ocpmp_status0 : std_logic_vector(3 downto 0);
  signal system_irq    : std_logic_vector(31 downto 0) := (others => '0');
  signal irq           : std_logic;
  

  signal   dclk_i     :    std_logic;
  signal   data_i     :    std_logic_vector(15 downto 0);
  
begin



  --
  -- Input MCLK through a normal differential input buffer, and then
  -- use a GTP wrapper to distribute it to the fabric.
  --

  ibuf_mclk : IBUFDS
    port map(
      I  => mclka_p,
      IB => mclka_n,
      O  => mclka_i);

  mclk_gtp_wrap_a : mclk_gtp_wrap
    port map(
      clkin  => mclka_i,
      clkout => mclka_ibufg);

  bufg_mclka : BUFG
    port map(
      I => mclka_ibufg,
      O => mclka_bufg);

  --
  -- Input 200MHz reference clock for IDELAY / ODELAY elements
  --

  ibufg_refclk : IBUFGDS_LVPECL_25
    port map(
      I  => refclk_p,
      IB => refclk_n,
      O  => refclk_ibufg);

  bufg_refclk : BUFG
    port map(
      I => refclk_ibufg,
      O => refclk_bufg);

  --
  -- Input and distribute local bus clock using a DCM.
  --

  lclk_dcm_0 : lclk_dcm
    port map(
      rst    => rst,
      refclk => refclk_bufg,
      lclk   => lclk,
      clk    => clk,
      locked => clk_locked);



  --
  -- Instantiate logic to generate memory clocks
  --

  memclk_0 : memclk
    port map(
      mclk          => mclka_bufg,
      rst           => rst,
      refclk200     => refclk_bufg,
      sel_refclk220 => '0',
      clk0          => memclk0,
      clk45         => memclk45,
      clk2x0        => memclk2x0,
      clk2x90       => memclk2x90,
      locked        => locked(2 downto 0),
      memrst        => memrst);



  --
  -- Instantiate SDK memory modules
  --
  
  dram_port_0 : ddr2sdram_port
    generic map(
       pinout => ddr2sdram_pinout_admxrc5t2,
       timing => ddr2sdram_timing_266,
       ra_width => bank0.ra_width,
       rc_width => bank0.rc_width,
       rd_width => bank0.rd_width,
       a_width => port_addr_width_dram,
       d_width => port_width_dram,
       tag_width => 2)
    port map(
      rst     => memrst,
      clk0    => memclk0,
      clk45   => memclk45,
      clk2x0  => memclk2x0,
      clk2x90 => memclk2x90,
      sr      => '0',
      ce      => ce0,
      w       => w0,
      tag     => tag0,
      a       => a0,
      d       => d0,
      be      => be0,
      q       => q0,
      qtag    => qtag0,
      ready   => ready0,
      valid   => valid0,
      row     => mode_reg(32 * 0 + 3 downto 32 * 0 + 2),
      col     => mode_reg(32 * 0 + 5 downto 32 * 0 + 4),
      bank    => mode_reg(32 * 0 + 7 downto 32 * 0 + 6),
      pbank   => mode_reg(32 * 0 + 9 downto 32 * 0 + 8),
      trained => locked(3),
      ra      => ra0,
      rc      => rc0,
      rd      => rd0);



  --
  -- Instantiate OCP bridge to memif ports
  --
  
  dram_brigde0 : ocp2memif128
    generic map (
      use_chipscope => false)
    port map (
      clk     => memclk0,
      ocp_m2s => ocp_m2s_ddr1,
      ocp_s2m => ocp_s2m_ddr1,
      sr      => open,
      ce      => ce0,
      tag     => tag0,
      w       => w0,
      a       => a0,
      d       => d0,
      be      => be0,
      q       => q0,
      qtag    => qtag0,
      valid   => valid0,
      ready   => ready0);

  
  mp_ocp1 : MULTIPORT_OCP_IF2_b128
    generic map (
      round_robin_arb => true)
    port map (
      ocp_m2s_memif => ocp_m2s_ddr1,
      ocp_s2m_memif => ocp_s2m_ddr1,
      ocp_m2s0      => sink_burst_m2s,
      ocp_s2m0      => sink_burst_s2m,
      ocp_m2s1      => ocp_m2s_burst_array(0),
      ocp_s2m1      => ocp_s2m_burst_array(0),
      status        => ocpmp_status0,
      clk           => memclk0
      );

--  ocp_csp0: ocp_chipscope128
--    port map (
--      ocp_m2s => ocp_m2s_ddr1,
--      ocp_s2m => ocp_s2m_ddr1,
--      clk     => memclk0);


  
  lbio0 : localbus_io
    port map (
      localbus_inputs  => localbus_inputs,
      localbus_outputs => localbus_outputs,
      lreset_l         => lreset_l,
      lwrite           => lwrite,
      lads_l           => lads_l,
      lblast_l         => lblast_l,
      lbterm_l         => lbterm_l,
      lad              => lad,
      lready_l         => lready_l,
      lbe_l            => lbe_l,
      fholda           => fholda,
      ldreq_l          => ldreq_l,
      ldack_l          => ldack_l);

  lb_ddma_if0 : localbus2ocp_bridge_r64b128
    generic map (
      la_top                         => 23,
      family                         => family_virtex4,
      number_of_ocp_nonburst         => number_of_ocp_nonburst,
      number_of_ddma_engines         => number_of_ddma_engines,
      number_of_ocp_burst_per_engine => number_of_ocp_burst_per_engine)
    port map (
      lclk                           => clk,
      ocp_clk_b                      => memclk0,
      ocp_clk_nb                     => clk,
      rst                            => rst,
      localbus_inputs                => localbus_inputs,
      localbus_outputs               => localbus_outputs,
      ocp_nonburst_bus_outputs_array => ocp_m2s_nonburst_array,
      ocp_nonburst_bus_inputs_array  => ocp_s2m_nonburst_array,
      ocp_burst_bus_outputs_array    => ocp_m2s_burst_array,
      ocp_burst_bus_inputs_array     => ocp_s2m_burst_array,
      system_status                  => system_status,
      system_irq                     => system_irq,
      irq                            => irq);

  finti_l <= not irq;

  sink0 : data_sink_b128r64
    generic map (
      data_width => 16,
      data_shift => 16)
    port map (
      burst_clk    => memclk0,
      burst_m2s    => sink_burst_m2s,
      burst_s2m    => sink_burst_s2m,
      reg_clk      => clk,
      reg_m2s      => ocp_m2s_nonburst_array(0),
      reg_s2m      => ocp_s2m_nonburst_array(0),
      data_clk     => dclk_i,
      data         => data_i,
      data_valid   => '1',
      rfd          => open,
      flush_packet => '0',
      irq          => system_irq(0));

  
  mode_reg(31 downto 0) <= X"000000D4";


  system_status <= EXT(ocpmp_status0 & locked & clk_locked, 32);

  map_data_in_signals: if not use_dummy_internal_data generate
    dclk_i <= dclk;
    data_i <= data;
  end generate map_data_in_signals;

  gen_dummy_data: if use_dummy_internal_data generate
    dclk_i <= clk;
    process (clk,rst)
    begin  -- process
      if rst = '1' then         
        data_i <= (others => '0');
      elsif clk'event and clk = '1' then  
        data_i <= data_i + 1;
      end if;
    end process;
  end generate gen_dummy_data;
  
end architecture;
