/*
** test_ocp_memory.c 
**
**
**
** Data is transferred using demand mode DMA to the FPGA, stored
** in memory and transferred out again also using demand mode DMA.
** Also demonstrated is configuring the FPGA using DMA.
**
** Example for:
**
**   o ADM-XRC-4LX
**   o ADM-XRC-5T1
**

*/

#include <windows.h>
#include <stdio.h>

#include <admxrc2.h>

#include "common.h"
#include "lb2ocp.h"


static int PERFORM_CHECK=1;


static ADMXRC2_HANDLE		card = NULL;
static ULONG				mode;




static double				got[1], sent[1];

static ULONG				bufSize;

static ADMXRC2_DMADESC		sendbufHandle;
static BOOLEAN				sendbufValid = FALSE;
static BYTE*				sendbuf = NULL;
static ADMXRC2_DMADESC		recvbufHandle;
static BOOLEAN				recvbufValid = FALSE;
static BYTE*				recvbuf = NULL;

static volatile DWORD*		fpgaSpace;


void blower(
      PLB2OCP bridge,
	    int target,
	    int start_addr,
	    int bank)
{
  ULONG				j;
  HANDLE				event;
  ULONG*				buf = (ULONG*) (sendbuf);
  ULONG				offs = 0;
  DWORD               val32 = 0x0U;

  event = CreateEvent(NULL, TRUE, FALSE, NULL);

  /* Set up the data */
  if (PERFORM_CHECK) {
    val32 = (DWORD) 0;
    val32 |= val32 << 8;
    val32 |= val32 << 16;
    for (j = 0; j < (bufSize >> 2); j++) {
      buf[j] = val32++;
    }
  }
  if (LB2OCP_TranferDataToFPGA(bridge, 
                               bank,
                               target,
                               bufSize,
                               start_addr,
                               offs,
                               sendbufHandle))
    sent[0] += bufSize;                    
  else
    printf("Blower failed to do DMA\n");
  
}

void sucker(PLB2OCP bridge,
	    int target,
	    int start_addr,
	    int bank)
{

  HANDLE				event;
  ULONG*				buf = (ULONG*) (recvbuf);
  ULONG				offs = 0;
  ULONG				n = 0;
  int errs=0;
 
 
  ULONG	exp, act;
  ULONG	j;
		
  event = CreateEvent(NULL, TRUE, FALSE, NULL);

 if (LB2OCP_TranferDataToHost(bridge, 
                               bank,
                               target,
                               bufSize,
                               start_addr,
                               offs,
                               recvbufHandle))
 { 
   if (PERFORM_CHECK) {
     /* Check the data */
     exp = buf[0];
     for (j = 1; j < (bufSize >> 2); j++) {
       exp++;
       act = buf[j];
       if (act != exp) {
         if (errs<32)
	        printf("Sucker : error at block %ld index %ld, exp/act=0x%8.8lx/0x%8.8lx\n",
	                (unsigned long) n, (unsigned long) j,
	                (unsigned long) exp, (unsigned long) act);
	       errs++;
        }
      }
    }  
    got[0] += bufSize;
    n++;
  }
  else
    printf("Sucker failed to do DMA\n");


 
}

int main(int argc, char *argv[])
{
	int					ret = 0;
	ADMXRC2_STATUS		status;
	ADMXRC2_CARD_INFO	cardInfo;
	ADMXRC2_SPACE_INFO	spaceInfo;
	int					i;
	DWORD				start;
	HANDLE				event = NULL;
	const char*			filename;
	Arguments			arguments;
	Option				options[7];
	float				clock_freq;
        float				mclock_freq;
	BOOLEAN				use64bit;
	double	s, g, t;
	DWORD	now, delta1,delta2;
  PLB2OCP bridge;
  int system_status;
  int banks = 2;
  
	options[0].key = "card";
	options[0].type = Option_uint;
	options[0].def.uintVal = 0;
	options[0].meaning = "ID of card to use";
	options[1].key = "index";
	options[1].type = Option_uint;
	options[1].def.uintVal = 0;
	options[1].meaning = "index of card to use";
	options[2].key = "size";
	options[2].type = Option_uint;
	options[2].def.uintVal = 64*1024;
	options[2].meaning = "send/receive buffer size (bytes)";
	options[3].key = "lclk";
	options[3].type = Option_float;
	options[3].def.floatVal = 33.0;
	options[3].meaning = "local bus clock frequency (MHz)";
  options[4].key = "mclk";
	options[4].type = Option_float;
	options[4].def.floatVal = 200.0;
	options[4].meaning = "memory clock frequency (MHz)";
	options[5].key = "bitstream";
	options[5].type = Option_string;
	options[5].def.stringVal = "test_ocp_bridge-v5sx.bit";
	options[5].meaning = "Bitstream to configure FPGA";
	options[6].key = "banks";
	options[6].type = Option_uint;
	options[6].def.uintVal = 2;
	options[6].meaning = "number of banks to test";
	arguments.nOption = 7;
	arguments.option = options;
	arguments.nParamType = 0;
	arguments.minNParam = 0;

	parse_command_line(
		"dma",
		argc,
		argv,
		&arguments);

	if (options[0].specified && options[1].specified) {
		printf("Do not specify both /card and /index\n");
		ret = -1;
		goto done;
	}

	if (!options[1].specified) {
		ADMXRC2_CARDID cardID = options[0].value.uintVal;

		status = ADMXRC2_OpenCard(cardID, &card);
		if (status != ADMXRC2_SUCCESS) {
			printf("Failed to open card with ID %ld: %s\n",
				(unsigned long) cardID, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	} else {
		unsigned int index = options[1].value.uintVal;

		status = ADMXRC2_OpenCardByIndex(index, &card);
		if (status != ADMXRC2_SUCCESS) {
			printf("Failed to open card with index %ld: %s\n",
				(unsigned long) index, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	}

	/* Allocate memory for the DMA data buffers */
	bufSize = options[2].value.uintVal;
	banks = options[6].value.uintVal;
	recvbuf = (BYTE*) malloc(bufSize  + bufSize );
	if (recvbuf == NULL) {
		printf("Failed to allocate buffers for threads\n");
		ret = -1;
		goto done;
	}
	sendbuf = recvbuf + bufSize;

	clock_freq = options[3].value.floatVal * 1.0e6f;
	PERFORM_CHECK = 1;
        
  mclock_freq = options[4].value.floatVal * 1.0e6f;
       

	/*
	** Create a manual reset event for use when configuring the FPGA
	** using DMA.
	*/
	event = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (event == NULL) {
		printf("Failed to create Win32 Event\n");
		ret = -1;
		goto done;
	}

	/* Get card information, for device fitted */
	status = ADMXRC2_GetCardInfo(card, &cardInfo);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to get card info: %s\n", ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	switch (cardInfo.BoardType) {

	case ADMXRC2_BOARD_ADMXRC4LX:
	  use64bit = 0;
	  break;
  case ADMXRC2_BOARD_ADMXRC4FX:
	case ADMXRC2_BOARD_ADMXRC5T1:
	case ADMXRC2_BOARD_ADMXRC5T2:
	  use64bit = 1;
		break;
	default:
	  printf("This example must be run on one of the following cards:\n\n");	
		printf("\tADM-XRC-4LX\n");
		printf("\tADM-XRC-5T1\n");
		ret = -1;
		goto done;
	}

	status = ADMXRC2_SetClockRate(card, ADMXRC2_CLOCK_LCLK, clock_freq, NULL);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to set LCLK to %.1fMHz: %s\n",
			clock_freq / 1.0e6f, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
	status = ADMXRC2_SetClockRate(card, 1, mclock_freq, NULL);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to set MCLK to %.1fMHz: %s\n",
			mclock_freq / 1.0e6f, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
	status = ADMXRC2_GetSpaceInfo(card, 0, &spaceInfo);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to get information about FPGA space\n");
		ret = -1;
		goto done;
	}
	fpgaSpace = (volatile DWORD*) spaceInfo.VirtualBase;
	if (fpgaSpace == NULL) {
		printf("*** Failed to retrieve pointer to FPGA space\n");
		return -1;
	}

	if (use64bit) {
		DWORD	config;

		config = ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
		status = ADMXRC2_SetSpaceConfig(card, 0, config);
		if (status != ADMXRC2_SUCCESS) {
			printf ("*** Failed to set space configuration: %s\n",
				ADMXRC2_GetStatusString(status));
			return -1;
		}
	}


	filename = options[5].value.stringVal;

	status = ADMXRC2_ConfigureFromFile(card, filename);
	if (status != ADMXRC2_SUCCESS) {
		printf ("Failed to load the bitstream '%s': %s\n",
			filename, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}


	/*
	** Build the DMA mode word. We want bursting, demand-mode,
	** use ready, use bterm and fixed local bus addressing.
	*/
	mode = ADMXRC2_BuildDMAModeWord(
		cardInfo.BoardType,
		use64bit ? ADMXRC2_IOWIDTH_64 : ADMXRC2_IOWIDTH_32,
		0,
		ADMXRC2_DMAMODE_FIXEDLOCAL |
		ADMXRC2_DMAMODE_USEREADY |
		ADMXRC2_DMAMODE_USEBTERM |
		ADMXRC2_DMAMODE_BURSTENABLE |
		ADMXRC2_DMAMODE_DEMAND);


	/* Get a DMA descriptor for the buffer used for DMA transfers to the card */
	status = ADMXRC2_SetupDMA(card, sendbuf, bufSize, 0, &sendbufHandle);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to set up send DMA buffer: %s\n", ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
	sendbufValid = TRUE;

	/* Get a DMA descriptor for the buffer used for DMA transfers from the card */
	status = ADMXRC2_SetupDMA(card, recvbuf, bufSize, 0, &recvbufHandle);
	if (status != ADMXRC2_SUCCESS) {
		printf("Failed to set up receive DMA buffer: %s\n", ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
	recvbufValid = TRUE;

  bridge = LB2OCP_Create(card,use64bit,use64bit);
  LB2OCP_SetNumberDMAEngines(bridge,2);       
      
  Sleep(1000);          
             
	system_status = LB2OCP_Status(bridge);       
  printf("System Status : %x\n",system_status);

  if (banks>0)
  {
    PERFORM_CHECK = 1;
    printf("Testing Bank 0\n");
    
    printf("Writing data\n");
   
    blower(bridge,0,0,0);
    printf("Reading data\n");
    sucker(bridge,0,0,0);
  } 
  if (banks>1)
  {  
  	printf("Testing Bank 1\n");
    printf("Writing data\n");
  
    blower(bridge,1,0,0);
    printf("Reading data\n");      
  	sucker(bridge,1,0,0);
  }
  PERFORM_CHECK = 0;

  if (banks>0)
  {
    printf("Write speed test\n");
    
    LB2OCP_SetBurstSize(bridge, 0, 32);
    LB2OCP_SetBurstSize(bridge, 1, 32);
    
    start = GetTickCount();
  
    for (i=0;i<100;i++)
    {
      printf(".");
  	  blower(bridge,0,0,0);
    }
          
    now = GetTickCount();
    delta1 = now - start;
    printf(".\n");
    printf("Read speed test\n");
    
    start = GetTickCount();
  
    for (i=0;i<100;i++)
     {
        printf(".");
  	    sucker(bridge,0,0,0);  
     }
    printf(".\n");
    now = GetTickCount();
    delta2 = now - start;
  
  	s = sent[0];       
    g = got[0];
          	
    s = s / 1048576.0f;
  	g = g / 1048576.0f;
  	t = s + g;
   
  	printf("Sent %.2fMB, received %.2fMB, total %.2fMB\n",
  		s, g, t);
  	
  	printf("Send rate %.2fMB/s, receive rate %.2fMB/s, total rate %.2fMB/s\n",
  			s * 1000.0 / ((double) delta1),
  			g * 1000.0 / ((double) delta2),
  			t * 1000.0 / ((double) delta1+delta2));
          
  }
  
  
done:
	/* Free the DMA descriptors */
	if (sendbufValid) {
		status = ADMXRC2_UnsetupDMA(card, sendbufHandle);
		if (status != ADMXRC2_SUCCESS) {
			printf("Failed to delete DMA send buffer: %s\n", ADMXRC2_GetStatusString(status));
		}
	}
	if (recvbufValid) {
		status = ADMXRC2_UnsetupDMA(card, recvbufHandle);
		if (status != ADMXRC2_SUCCESS) {
			printf("Failed to delete DMA receive buffer: %s\n", ADMXRC2_GetStatusString(status));
		}
	}



	/* Free the Win32 Event */
	if (event != NULL) {
		CloseHandle(event);
	}

	/* Close the device */
	if (card != NULL) {
		ADMXRC2_CloseCard(card);
	}
	
	/* Free the memory used for the DMA data buffers */
	if (recvbuf != NULL) {
		free(recvbuf);
	}

	return ret;
}
