--
-- memory-xrc4lx.vhd
--
-- Example demonstrating OCP connection to 4 banks of 4 LX memory

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

library work;
use work.memif.all;
use work.ocpad.all;
use work.fifo.all;
use work.localbus_if_types.all;
use work.ocp_ddma_components.all;


entity memory is
    generic(
        -- Memory bank parameters
        bank0             : in    bank_t := (enable => true, ra_width => 21, rc_width => 8, rd_width => 32);
        bank1             : in    bank_t := (enable => true, ra_width => 21, rc_width => 8, rd_width => 32);
        bank2             : in    bank_t := (enable => true, ra_width => 21, rc_width => 8, rd_width => 32);
        bank3             : in    bank_t := (enable => true, ra_width => 21, rc_width => 8, rd_width => 32);
        num_ramclk        : in    natural := 4);
    port(
        -- Local bus
        lclk              : in    std_logic;
        lreset_l          : in    std_logic; 
        lwrite            : in    std_logic;
        lads_l            : in    std_logic;
        lblast_l          : in    std_logic;
        lbe_l             : in    std_logic_vector(3 downto 0);
        la                : in    std_logic_vector(23 downto 2);
        ld                : inout std_logic_vector(31 downto 0);
        lbterm_l          : inout std_logic;
        lready_l          : inout std_logic; 
        fholda            : in    std_logic;
        ldreq_l  : out   std_logic_vector(1 downto 0);
        ldack_l  : in    std_logic_vector(1 downto 0);
        -- General purpose clock
        mclk_n            : in    std_logic;
        mclk_p            : in    std_logic;
        -- Reference clock
        refclk_n          : in    std_logic;
        refclk_p          : in    std_logic;
        -- Memory banks (up to 16 supported by this design)
        ra0               : out   std_logic_vector(bank0.ra_width - 1 downto 0);
        rc0               : inout std_logic_vector(bank0.rc_width - 1 downto 0);
        rd0               : inout std_logic_vector(bank0.rd_width - 1 downto 0);
        ra1               : out   std_logic_vector(bank1.ra_width - 1 downto 0);
        rc1               : inout std_logic_vector(bank1.rc_width - 1 downto 0);
        rd1               : inout std_logic_vector(bank1.rd_width - 1 downto 0);
        ra2               : out   std_logic_vector(bank2.ra_width - 1 downto 0);
        rc2               : inout std_logic_vector(bank2.rc_width - 1 downto 0);
        rd2               : inout std_logic_vector(bank2.rd_width - 1 downto 0);
        ra3               : out   std_logic_vector(bank3.ra_width - 1 downto 0);
        rc3               : inout std_logic_vector(bank3.rc_width - 1 downto 0);
        rd3               : inout std_logic_vector(bank3.rd_width - 1 downto 0);
        
        ramclki           : in    std_logic_vector(num_ramclk - 1 downto 0);
        ramclko           : out   std_logic_vector(num_ramclk - 1 downto 0));
end entity;

architecture struct of memory is

    signal rst : std_logic;
    signal clk : std_logic;
    signal clk_locked : std_logic;
    signal mclk_ibufg : std_logic;
    signal refclk_ibufg : std_logic;
    signal refclk_bufg : std_logic;

    signal ld_o : std_logic_vector(31 downto 0);
    signal ld_oe_l : std_logic_vector(31 downto 0);

    signal scl : std_logic;
    signal sda : std_logic;

    signal logicx : std_logic;

    signal memclk0 : std_logic;
    signal memclk90 : std_logic;
    signal memclk180 : std_logic;
    signal memclk270 : std_logic;
    signal memrst : std_logic;

    constant number_of_ocp_nonburst         : integer := 1;
    constant number_of_ddma_engines         : integer := 1;
    constant number_of_ocp_burst_per_engine : integer := 4;


    signal localbus_inputs  : localbus_inputs_type;
    signal localbus_outputs : localbus_outputs_type;

    signal ocp_m2s_nonburst_array : ocpad_register32_m2sT_array(0 to number_of_ocp_nonburst-1);
    signal ocp_s2m_nonburst_array  : ocpad_register32_s2mT_array(0 to number_of_ocp_nonburst-1);
    signal ocp_m2s_burst_array    : ocpad_burst32_m2sT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);
    signal ocp_s2m_burst_array     : ocpad_burst32_s2mT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);

    signal system_status : std_logic_vector(31 downto 0);

    signal locked : std_logic_vector(1 downto 0);
    
    component memclk
      generic(
        num_clock     : in    natural);
      port(
        rst           : in    std_logic;
        mclk_i        : in    std_logic;
        memclk0       : out   std_logic;
        memclk90      : out   std_logic;
        memclk180     : out   std_logic;
        memclk270     : out   std_logic;
        ramclko       : out   std_logic_vector(num_clock - 1 downto 0);
        ramclki       : in    std_logic_vector(num_clock - 1 downto 0);
        locked        : out   std_logic_vector(1 downto 0);
        memrst        : out   std_logic);
    end component;

    component ocp2memif32
      generic (
        use_chipscope : boolean := false;
        addr_width    : integer);
      port (
        clk                     : in  std_logic;
        -- OCP Port
        ocp_m2s                 : in ocpad_burst32_m2sT;
        ocp_s2m                 : out ocpad_burst32_s2mT;
        -- MemIF Port (32 bits wide)
        sr                      : out std_logic;
        ce                      : out std_logic;
        tag                     : out std_logic_vector(1 downto 0);
        w                       : out std_logic;
        a                       : out std_logic_vector(addr_width-1 downto 0);
        d                       : out std_logic_vector(31 downto 0);
        be                      : out std_logic_vector(3 downto 0);
        q                       : in  std_logic_vector(31 downto 0);
        qtag                    : in  std_logic_vector(1 downto 0);
        valid                   : in  std_logic;
        ready                   : in  std_logic);
    end component;
    
    component lclk_dcm
        port(
            rst           : in    std_logic;
            refclk        : in    std_logic;
            lclk          : in    std_logic;
            clk           : out   std_logic;
            locked        : out   std_logic);
    end component;

    component IBUFGDS
        port(
            I  : in  std_logic;
            IB : in  std_logic;
            O  : out std_logic);
    end component;

    component BUFG
        port(
            I  : in  std_logic;
            O  : out std_logic);
    end component;

    --
    -- If the synthesizer replicates an asynchronous reset signal due high fanout,
    -- this can prevent flip-flops being mapped into IOBs. We set the maximum
    -- fanout for such nets to a high enough value that replication never occurs.
    --
    attribute MAX_FANOUT : string;
    attribute MAX_FANOUT of rst : signal is "1000000";

    constant num_bank : integer := 4;
    constant port_width : natural := 32;
    constant port_be_width : natural := port_width / 8;
    constant port_addr_width : natural := 24;

    constant tag_width : natural := 2;
    subtype tag_t is std_logic_vector(tag_width - 1 downto 0);
    type tag_vector_t is array(natural range <>) of tag_t;

    signal tag_base : std_logic_vector(tag_width - 1 downto 0);
    signal tag_mask : std_logic_vector(tag_width - 1 downto 0);

    type port_pq_t is array(0 to num_bank - 1) of std_logic_vector(port_width - 1 downto 0);
    signal port_pce : std_logic_vector(num_bank - 1 downto 0);
    signal port_pcw : std_logic_vector(num_bank - 1 downto 0);
    signal port_pterm : std_logic_vector(num_bank - 1 downto 0);
    signal port_padv : std_logic_vector(num_bank - 1 downto 0);
    signal port_pwr : std_logic_vector(num_bank - 1 downto 0);
    signal port_pq : port_pq_t;
    signal port_pwf : std_logic_vector(num_bank - 1 downto 0);
    signal port_pwpf : std_logic_vector(num_bank - 1 downto 0);
    signal port_pre : std_logic_vector(num_bank - 1 downto 0);
    signal port_prpe : std_logic_vector(num_bank - 1 downto 0);

    signal port_plast : std_logic;
    signal port_pq_muxed : std_logic_vector(port_width - 1 downto 0);

    type port_sa_t is array(0 to num_bank - 1) of std_logic_vector(port_addr_width - 1 downto 0);
    type port_sd_t is array(0 to num_bank - 1) of std_logic_vector(port_width - 1 downto 0);
    type port_sbe_t is array(0 to num_bank - 1) of std_logic_vector(port_be_width - 1 downto 0);
    type port_stag_t is array(0 to num_bank - 1) of std_logic_vector(tag_width - 1 downto 0);
    signal port_sce : std_logic_vector(num_bank - 1 downto 0);
    signal port_sw : std_logic_vector(num_bank - 1 downto 0);
    signal port_stag : port_stag_t;
    signal port_sa : port_sa_t;
    signal port_sd : port_sd_t;
    signal port_sbe : port_sbe_t;
    signal port_sq : port_sd_t;
    signal port_sqtag : port_stag_t;
    signal port_svalid : std_logic_vector(num_bank - 1 downto 0);
    signal port_sready : std_logic_vector(num_bank - 1 downto 0);

   

    signal registers_out,registers_back : register32_array(0 to 1);

   
    
begin

    logicx <= '-';

    
    ibufg_mclk : IBUFGDS
        port map(
            I  => mclk_p,
            IB => mclk_n,
            O  => mclk_ibufg);
    
    ibufg_refclk : IBUFGDS
        port map(
            I  => refclk_p,
            IB => refclk_n,
            O  => refclk_ibufg);
    
    bufg_refclk : BUFG
        port map(
            I  => refclk_ibufg,
            O  => refclk_bufg);
    
    lclk_dcm_0 : lclk_dcm
        port map(
            rst => rst,
            refclk => refclk_bufg,
            lclk => lclk,
            clk => clk,
            locked => clk_locked);


     memclk_0 : memclk
        generic map(
            num_clock => num_ramclk)
        port map(
            rst => rst,
            mclk_i => mclk_ibufg,
            memclk0 => memclk0,
            memclk90 => memclk90,
            memclk180 => memclk180,
            memclk270 => memclk270,
            ramclki => ramclki,
            ramclko => ramclko,
            locked => locked(1 downto 0),
            memrst => memrst);



    lbio0 : localbus_io_plx
    port map (
      localbus_inputs  => localbus_inputs,
      localbus_outputs => localbus_outputs,
      lreset_l         => lreset_l,
      lwrite           => lwrite,
      lads_l           => lads_l,
      lblast_l         => lblast_l,
      lbterm_l         => lbterm_l,
      ld               => ld,
      la               => la,
      lready_l         => lready_l,
      lbe_l            => lbe_l,
      fholda           => fholda,
      ldreq_l          => ldreq_l,
      ldack_l          => ldack_l);

  lb_ddma_if0 : localbus2ocp_bridge_r32b32
    generic map (
      la_top                         => 23,
      family                         => family_virtex4,
      number_of_ocp_nonburst         => number_of_ocp_nonburst,
      number_of_ddma_engines         => number_of_ddma_engines,
      number_of_ocp_burst_per_engine => number_of_ocp_burst_per_engine)
    port map (
      lclk                           => clk,
      ocp_clk_b                      => memclk0,
      ocp_clk_nb                     => clk,
      rst                            => rst,
      localbus_inputs                => localbus_inputs,
      localbus_outputs               => localbus_outputs,
      ocp_nonburst_bus_outputs_array => ocp_m2s_nonburst_array,
      ocp_nonburst_bus_inputs_array  => ocp_s2m_nonburst_array,
      ocp_burst_bus_outputs_array    => ocp_m2s_burst_array,
      ocp_burst_bus_inputs_array     => ocp_s2m_burst_array,
      system_status => system_status,
      system_irq => X"00000000",
      irq => open);


    system_status <= EXT(locked & clk_locked,32);
    
    ocp_ra0: ocp_register_interface32
      generic map (
        number_of_registers => 2)
      port map (
        ocp_clk           => clk,
        ocp_m2s           => ocp_m2s_nonburst_array(0),
        ocp_s2m           => ocp_s2m_nonburst_array(0),
        registers_out     => registers_out,
        registers_back    => registers_back,
        register_written => open);
    

    registers_back <=   registers_out;

    

    
    gen_ocp_memif: for i in 0 to num_bank-1 generate
      sram_brigde_i: ocp2memif32
        generic map (
          use_chipscope => false,
          addr_width => port_addr_width)
        port map (
          clk     => memclk0,
          ocp_m2s => ocp_m2s_burst_array(i),
          ocp_s2m => ocp_s2m_burst_array(i),
          sr      => open,
          ce      => port_sce(i),
          tag     => port_stag(i),
          w       => port_sw(i),
          a       => port_sa(i),
          d       => port_sd(i),
          be      => port_sbe(i),
          q       => port_sq(i),
          qtag    => port_sqtag(i),
          valid   => port_svalid(i),
          ready   => port_sready(i));      
    end generate gen_ocp_memif;

    sram_port_0 : zbtsram_port
       generic map(
            pinout => zbtsram_pinout_admxrc4lx,
            ra_width => bank0.ra_width,
            rc_width => bank0.rc_width,
            rd_width => bank0.rd_width,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => tag_width)
        port map(
            clk0 => memclk0,
            rst => memrst,
            sr => '0',
            ce => port_sce(0),
            w => port_sw(0),
            tag => port_stag(0),
            a => port_sa(0),
            d => port_sd(0),
            be => port_sbe(0),
            q => port_sq(0),
            qtag => port_sqtag(0),
            valid => port_svalid(0),
            ready => port_sready(0),
            pipeline => '1',
            ra => ra0,
            rc => rc0,
            rd => rd0);

    sram_port_1 : zbtsram_port
       generic map(
            pinout => zbtsram_pinout_admxrc4lx,
            ra_width => bank1.ra_width,
            rc_width => bank1.rc_width,
            rd_width => bank1.rd_width,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => tag_width)
        port map(
            clk0 => memclk0,
            rst => memrst,
            sr => '0',
            ce => port_sce(1),
            w => port_sw(1),
            tag => port_stag(1),
            a => port_sa(1),
            d => port_sd(1),
            be => port_sbe(1),
            q => port_sq(1),
            qtag => port_sqtag(1),
            valid => port_svalid(1),
            ready => port_sready(1),
            pipeline => '1',
            ra => ra1,
            rc => rc1,
            rd => rd1);

    sram_port_2 : zbtsram_port
         generic map(
            pinout => zbtsram_pinout_admxrc4lx,
            ra_width => bank2.ra_width,
            rc_width => bank2.rc_width,
            rd_width => bank2.rd_width,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => tag_width)
        port map(
            clk0 => memclk0,
            rst => memrst,
            sr => '0',
            ce => port_sce(2),
            w => port_sw(2),
            tag => port_stag(2),
            a => port_sa(2),
            d => port_sd(2),
            be => port_sbe(2),
            q => port_sq(2),
            qtag => port_sqtag(2),
            valid => port_svalid(2),
            ready => port_sready(2),
            pipeline => '1',
            ra => ra2,
            rc => rc2,
            rd => rd2);

    sram_port_3 : zbtsram_port
        generic map(
            pinout => zbtsram_pinout_admxrc4lx,
            ra_width => bank3.ra_width,
            rc_width => bank3.rc_width,
            rd_width => bank3.rd_width,
            a_width => port_addr_width,
            d_width => port_width,
            tag_width => tag_width)
        port map(
            clk0 => memclk0,
            rst => memrst,
            sr => '0',
            ce => port_sce(3),
            w => port_sw(3),
            tag => port_stag(3),
            a => port_sa(3),
            d => port_sd(3),
            be => port_sbe(3),
            q => port_sq(3),
            qtag => port_sqtag(3),
            valid => port_svalid(3),
            ready => port_sready(3),
            pipeline => '1',
            ra => ra3,
            rc => rc3,
            rd => rd3);
    
end architecture;
