--
-- memory-xrc5t1.vhd 
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_unsigned.all;

-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

library work;
use work.memif.all;
use work.ocpad.all;
use work.fifo.all;
use work.localbus_if_types.all;
use work.ocp_ddma_components.all;

entity memory is
    generic(
        -- Memory bank parameters
        bank0             : in    bank_t := (enable => true, ra_width => 19, rc_width => 22, rd_width => 32);
        bank1             : in    bank_t := (enable => true, ra_width => 19, rc_width => 22, rd_width => 32));
    port(
        -- Local bus
        lclk              : in    std_logic;
        lreset_l          : in    std_logic; 
        lwrite            : in    std_logic;
        lads_l            : in    std_logic;
        lblast_l          : in    std_logic;
        lbe_l             : in    std_logic_vector(7 downto 0);
        lad               : inout std_logic_vector(63 downto 0);
        lbterm_l          : inout std_logic;
        lready_l          : inout std_logic; 
        fholda            : in    std_logic;
        ldreq_l  : out   std_logic_vector(3 downto 0);
        ldack_l  : in    std_logic_vector(3 downto 0);
        -- General purpose clock
        mclka_n           : in    std_logic;
        mclka_p           : in    std_logic;
        -- Reference clock
        refclk_n          : in    std_logic;
        refclk_p          : in    std_logic;
        -- Memory banks (up to 16 supported by this design)
        ra0               : out   std_logic_vector(bank0.ra_width - 1 downto 0);
        rc0               : inout std_logic_vector(bank0.rc_width - 1 downto 0);
        rd0               : inout std_logic_vector(bank0.rd_width - 1 downto 0);
        ra1               : out   std_logic_vector(bank1.ra_width - 1 downto 0);
        rc1               : inout std_logic_vector(bank1.rc_width - 1 downto 0);
        rd1               : inout std_logic_vector(bank1.rd_width - 1 downto 0));
end entity;

architecture struct of memory is

    signal rst : std_logic;
    signal clk : std_logic;
    signal clk_locked : std_logic;
    signal mclka_i : std_logic;
    signal mclka_ibufg : std_logic;
    signal mclka_bufg : std_logic;
    signal refclk_ibufg : std_logic;
    signal refclk_bufg : std_logic;

   

    constant number_of_ocp_nonburst         : integer := 1;
    constant number_of_ddma_engines         : integer := 1;
    constant number_of_ocp_burst_per_engine : integer := 2;


    signal localbus_inputs  : localbus_inputs_type;
    signal localbus_outputs : localbus_outputs_type;

    signal ocp_m2s_nonburst_array : ocpad_register64_m2sT_array(0 to number_of_ocp_nonburst-1);
    signal ocp_s2m_nonburst_array  : ocpad_register64_s2mT_array(0 to number_of_ocp_nonburst-1);
    signal ocp_m2s_burst_array    : ocpad_burst128_m2sT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);
    signal ocp_s2m_burst_array     : ocpad_burst128_s2mT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);

    signal system_status : std_logic_vector(31 downto 0);


    component mclk_gtp_wrap
        port(
            clkin  : in    std_logic;
            clkout : out   std_logic);
    end component;
    
    component lclk_dcm
        port(
            rst           : in    std_logic;
            refclk        : in    std_logic;
            lclk          : in    std_logic;
            clk           : out   std_logic;
            locked        : out   std_logic);
    end component;

    component BUFG
        port(
            I : in  std_logic;
            O : out std_logic);
    end component;

    component IBUFDS
        port(
            I  : in  std_logic;
            IB : in  std_logic;
            O  : out std_logic);
    end component;

    component IBUFGDS_LVPECL_25
        port(
            I  : in  std_logic;
            IB : in  std_logic;
            O  : out std_logic);
    end component;

    --
    -- If the synthesizer replicates an asynchronous reset signal due high fanout,
    -- this can prevent flip-flops being mapped into IOBs. We set the maximum
    -- fanout for such nets to a high enough value that replication never occurs.
    --
    attribute MAX_FANOUT : string;
    attribute MAX_FANOUT of rst : signal is "1000000";

    component ocp2memif128
      generic (
        use_chipscope : boolean := false);
      port (
        clk                     : in  std_logic;
        -- OCP Port
        ocp_m2s                 : in ocpad_burst128_m2sT;
        ocp_s2m                 : out ocpad_burst128_s2mT;
        -- MemIF Port (128 bits wide)
        sr                      : out std_logic;
        ce                      : out std_logic;
        tag                     : out std_logic_vector(1 downto 0);
        w                       : out std_logic;
        a                       : out std_logic_vector(27 downto 0);
        d                       : out std_logic_vector(127 downto 0);
        be                      : out std_logic_vector(15 downto 0);
        q                       : in  std_logic_vector(127 downto 0);
        qtag                    : in  std_logic_vector(1 downto 0);
        valid                   : in  std_logic;
        ready                   : in  std_logic);
    end component;

    component memclk
      port (
        mclk              : in  std_logic;
        rst               : in  std_logic;
        refclk200         : in  std_logic;
        sel_refclk220     : in  std_logic;
        clk0              : out std_logic;
        clk45             : out std_logic;
        clk2x0            : out std_logic;
        clk2x90           : out std_logic;
        locked            : out std_logic_vector(2 downto 0);
        memrst            : out std_logic);
    end component;
    

    constant port_addr_width_dram : integer := 28;
    constant port_width_dram : integer := 128;
    

    signal w0 : std_logic;
    signal ce0 : std_logic;
    signal r0 : std_logic;
    signal be0 : std_logic_vector(port_width_dram/8-1 downto 0);
    signal a0 : std_logic_vector(port_addr_width_dram-1 downto 0);
    signal d0 : std_logic_vector(port_width_dram-1 downto 0);
    signal q0 : std_logic_vector(port_width_dram-1 downto 0);
    signal valid0 : std_logic;
    signal ready0 : std_logic;
    signal tag0 : std_logic_vector(1 downto 0);
    signal qtag0 : std_logic_vector(1 downto 0);

    signal w1 : std_logic;
    signal ce1 : std_logic;
    signal r1 : std_logic;
    signal be1 : std_logic_vector(port_width_dram/8-1 downto 0);
    signal a1 : std_logic_vector(port_addr_width_dram-1 downto 0);
    signal d1 : std_logic_vector(port_width_dram-1 downto 0);
    signal q1 : std_logic_vector(port_width_dram-1 downto 0);
    signal valid1 : std_logic;
    signal ready1 : std_logic;
    signal tag1 : std_logic_vector(1 downto 0);
    signal qtag1 : std_logic_vector(1 downto 0);
   


    
    signal memclk0 : std_logic;
    signal memclk45 : std_logic;
    signal memclk2x0 : std_logic;
    signal memclk2x90 : std_logic;
    signal memrst : std_logic;

    signal locked : std_logic_vector(4 downto 0);

    signal mode_reg : std_logic_vector(63 downto 0);

    signal registers_out,registers_back : register64_array(0 to 1);

  
  
    
begin


    --rst <= not lreset_l;

    --
    -- Input MCLK through a normal differential input buffer, and then
    -- use a GTP wrapper to distribute it to the fabric.
    --

    ibuf_mclk : IBUFDS
        port map(
            I => mclka_p,
            IB => mclka_n,
            O => mclka_i);

    mclk_gtp_wrap_a : mclk_gtp_wrap
        port map(
            clkin => mclka_i,
            clkout => mclka_ibufg);

    bufg_mclka : BUFG
        port map(
            I  => mclka_ibufg,
            O  => mclka_bufg);
    
    --
    -- Input 200MHz reference clock for IDELAY / ODELAY elements
    --

    ibufg_refclk : IBUFGDS_LVPECL_25
        port map(
            I  => refclk_p,
            IB => refclk_n,
            O  => refclk_ibufg);
    
    bufg_refclk : BUFG
        port map(
            I  => refclk_ibufg,
            O  => refclk_bufg);
    
    --
    -- Input and distribute local bus clock using a DCM.
    --

    lclk_dcm_0 : lclk_dcm
        port map(
            rst => rst,
            refclk => refclk_bufg,
            lclk => lclk,
            clk => clk,
            locked => clk_locked);



    --
    -- Instantiate logic to generate memory clocks
    --

    memclk_0 : memclk
        port map(
            mclk => mclka_bufg,
            rst => rst,
            refclk200 => refclk_bufg,
            sel_refclk220 => '0',
            clk0 => memclk0,
            clk45 => memclk45,
            clk2x0 => memclk2x0,
            clk2x90 => memclk2x90,
            locked => locked(2 downto 0),
            memrst => memrst);

    

    --
    -- Instantiate SDK memory modules
    --
    
    dram_port_0 : ddr2sdram_port
       generic map(
            pinout => ddr2sdram_pinout_admxrc5t1,
            timing => ddr2sdram_timing_266,
            ra_width => bank0.ra_width,
            rc_width => bank0.rc_width,
            rd_width => bank0.rd_width,
            a_width => port_addr_width_dram,
            d_width => port_width_dram,
            tag_width => 2)
        port map(
            rst => memrst,
            clk0 => memclk0,
            clk45 => memclk45,
            clk2x0 => memclk2x0,
            clk2x90 => memclk2x90,
            sr => registers_out(0)(0),
            ce => ce0,
            w => w0,
            tag => tag0,
            a => a0,
            d => d0,
            be => be0,
            q => q0,
            qtag => qtag0,
            ready => ready0,
            valid => valid0,
            row => mode_reg(32 * 0 + 3 downto 32 * 0 + 2),
            col => mode_reg(32 * 0 + 5 downto 32 * 0 + 4),
            bank => mode_reg(32 * 0 + 7 downto 32 * 0 + 6),
            pbank => mode_reg(32 * 0 + 9 downto 32 * 0 + 8),
            trained => locked(3),
            ra => ra0,
            rc => rc0,
            rd => rd0);

    dram_port_1 : ddr2sdram_port
        generic map(
            pinout => ddr2sdram_pinout_admxrc5t1,
            timing => ddr2sdram_timing_266,
            ra_width => bank1.ra_width,
            rc_width => bank1.rc_width,
            rd_width => bank1.rd_width,
            a_width => port_addr_width_dram,
            d_width => port_width_dram,
            tag_width => 2)
        port map(
            rst => memrst,
            clk0 => memclk0,
            clk45 => memclk45,
            clk2x0 => memclk2x0,
            clk2x90 => memclk2x90,
            sr => registers_out(1)(0),
            ce => ce1,
            w => w1,
            tag => tag1,
            a => a1,
            d => d1,
            be => be1,
            q => q1,
            qtag => qtag1,
            ready => ready1,
            valid => valid1,
            row => mode_reg(32 * 1 + 3 downto 32 * 1 + 2),
            col => mode_reg(32 * 1 + 5 downto 32 * 1 + 4),
            bank => mode_reg(32 * 1 + 7 downto 32 * 1 + 6),
            pbank => mode_reg(32 * 1 + 9 downto 32 * 1 + 8),
            trained => locked(4),
            ra => ra1,
            rc => rc1,
            rd => rd1);
    

    --
    -- Instantiate OCP bridge to memif ports
    --
    
    dram_brigde0: ocp2memif128
      generic map (
        use_chipscope => false)
      port map (
        clk     => memclk0,
        ocp_m2s => ocp_m2s_burst_array(0),
        ocp_s2m => ocp_s2m_burst_array(0),
        sr      => open,
        ce      => ce0,
        tag     => tag0,
        w       => w0,
        a       => a0,
        d       => d0,
        be      => be0,
        q       => q0,
        qtag    => qtag0,
        valid   => valid0,
        ready   => ready0);

    dram_brigde1: ocp2memif128
      generic map (
        use_chipscope => false)
      port map (
        clk     => memclk0,
        ocp_m2s => ocp_m2s_burst_array(1),
        ocp_s2m => ocp_s2m_burst_array(1),
        sr      => open,
        ce      => ce1,
        tag     => tag1,
        w       => w1,
        a       => a1,
        d       => d1,
        be      => be1,
        q       => q1,
        qtag    => qtag1,
        valid   => valid1,
        ready   => ready1);



  lbio0 : localbus_io
    port map (
      localbus_inputs  => localbus_inputs,
      localbus_outputs => localbus_outputs,
      lreset_l         => lreset_l,
      lwrite           => lwrite,
      lads_l           => lads_l,
      lblast_l         => lblast_l,
      lbterm_l         => lbterm_l,
      lad              => lad,
      lready_l         => lready_l,
      lbe_l            => lbe_l,
      fholda           => fholda,
      ldreq_l          => ldreq_l,
      ldack_l          => ldack_l);

  lb_ddma_if0 : localbus2ocp_bridge_r64b128
    generic map (
      la_top                         => 23,
      family                         => family_virtex4,
      number_of_ocp_nonburst         => number_of_ocp_nonburst,
      number_of_ddma_engines         => number_of_ddma_engines,
      number_of_ocp_burst_per_engine => number_of_ocp_burst_per_engine)
    port map (
      lclk                           => clk,
      ocp_clk_b                      => memclk0,
      ocp_clk_nb                     => clk,
      rst                            => rst,
      localbus_inputs                => localbus_inputs,
      localbus_outputs               => localbus_outputs,
      ocp_nonburst_bus_outputs_array => ocp_m2s_nonburst_array,
      ocp_nonburst_bus_inputs_array  => ocp_s2m_nonburst_array,
      ocp_burst_bus_outputs_array    => ocp_m2s_burst_array,
      ocp_burst_bus_inputs_array     => ocp_s2m_burst_array,
      system_status => system_status,
      system_irq => X"00000000",
      irq => open);


    ocp_ra0: ocp_register_interface64
      generic map (
        number_of_registers => 2)
      port map (
        ocp_clk           => clk,
        ocp_m2s           => ocp_m2s_nonburst_array(0),
        ocp_s2m           => ocp_s2m_nonburst_array(0),
        registers_out     => registers_out,
        registers_back    => registers_back,
        register_written => open);
    

    registers_back <=   registers_out;

    --mode_reg(31 downto 0) <= registers_out(0)(31 downto 0);
    --mode_reg(63 downto 32) <= registers_out(1)(31 downto 0);
    mode_reg(31 downto 0) <= X"000000D4";
    mode_reg(63 downto 32) <= X"000000D4";
    
    system_status <= EXT(locked & clk_locked,32);
    
end architecture;
