--
-- ddma_rfsm_b128.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
-- DDMA RFSM
--
-- Demand mode DMA Read State Machine

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on


entity ddma_rfsm_b128 is
  
  port (
    clk          : in  std_logic;
    -- burst parameters
    burst_size   : in  std_logic_vector(7 downto 0);
    xfer_size    : in  std_logic_vector(27 downto 0);
    row_size     : in  std_logic_vector(19 downto 0);
    n_rows       : in  std_logic_vector(19 downto 0);
    col_inc      : in  std_logic_vector(19 downto 0);
    start_addr   : in  std_logic_vector(31 downto 0);
    fgsg_size    : in  std_logic_vector(3 downto 0);
    fgsg_pattern : in  std_logic_vector(11 downto 0);
    -- control signals
    start        : in  std_logic;
    read_cmd    : out std_logic;
    cmd_addr     : out std_logic_vector(31 downto 0);
    cmd_accept   : in  std_logic;
    resp_valid   : in  std_logic;
    resp_accept  : out std_logic;
    burst_length : out std_logic_vector(7 downto 0);
    resp_last    : in  std_logic;
    fifo_level   : in  std_logic_vector(15 downto 0);
    fifo_wadv    : out std_logic;
    -- transfer status
    xfer_status  : out std_logic_vector(63 downto 0)
    );

end ddma_rfsm_b128;

architecture rtl of ddma_rfsm_b128 is

  type state_type is (idle, cmd_wait, resp_wait, timeout_cleanup, end_of_burst);

  signal state : state_type := idle;

  signal xfer_level      : std_logic_vector(27 downto 0) := (others => '0');
  signal next_burst_size : std_logic_vector(7 downto 0) := (others => '0');
  signal burst_size_reg  : std_logic_vector(7 downto 0) := (others => '0');
  signal addr            : std_logic_vector(31 downto 0) := (others => '0');
  signal burst_size_i    : std_logic_vector(7 downto 0) := (others => '0');
  signal burst_count     : std_logic_vector(7 downto 0) := (others => '0');

  signal row_size_i : std_logic_vector(19 downto 0) := (others => '0');
  signal n_rows_i   : std_logic_vector(19 downto 0) := (others => '0');
  signal col_inc_i  : std_logic_vector(19 downto 0) := (others => '0');


  signal row_count : std_logic_vector(19 downto 0) := (others => '0');

  signal start_burst : std_logic := '0';

  signal next_addr_col_inc : std_logic_vector(31 downto 0) := (others => '0');
  signal next_addr_row_inc : std_logic_vector(31 downto 0) := (others => '0');
  signal next_xfer_level      : std_logic_vector(27 downto 0) := (others => '0');

  signal row_count_eq_n_rows : std_logic := '0';
  signal burst_size_reg_m1  : std_logic_vector(7 downto 0) := (others => '0');
  signal last_in_burst : std_logic := '0';


  signal fgsg_size_i  : std_logic_vector(3 downto 0) := (others => '0');
  signal fgsg_shift_reg : std_logic_vector(11 downto 0) := (others => '0');
  signal fgsg_active : std_logic := '0';
  
  signal resp_valid_q : std_Logic;
  signal timeout_count : std_logic_vector(15 downto 0);
  signal timeout :  std_logic;
  signal timeout_latch :  std_logic;

begin  -- rtl

  process (clk)
  begin  -- process
    if clk'event and clk = '1' then     -- rising clock edge
      
      case state is
        when idle =>
          if start_burst = '1' then
            state <= cmd_wait;
          end if;
          burst_size_reg <= next_burst_size;
        when cmd_wait =>
          if OR_reduce(xfer_level) = '0' then
            state       <= end_of_burst;
          elsif cmd_accept = '1' then
            state <= resp_wait;
          elsif timeout = '1' then
            state <= timeout_cleanup;
          end if;
          burst_size_reg_m1 <= burst_size_reg-1;
        when resp_wait =>
          if resp_valid = '1' then
            if last_in_burst = '1' then
              state       <= end_of_burst;
              burst_count <= CONV_STD_LOGIC_VECTOR(1, 8);
              if row_count_eq_n_rows = '1' then
                row_count <= CONV_STD_LOGIC_VECTOR(1, 20);
                addr      <= next_addr_col_inc;
              else
                row_count <= row_count+1;
                addr      <= next_addr_row_inc;
              end if;
              xfer_level <= next_xfer_level;
            else
              burst_count <= burst_count+1;
            end if;
          elsif timeout = '1' then
            state <= timeout_cleanup;
          end if;
        when timeout_cleanup =>
          if last_in_burst = '1' then
            state       <= end_of_burst;
            burst_count <= CONV_STD_LOGIC_VECTOR(1, 8);
            if row_count_eq_n_rows = '1' then
              row_count <= CONV_STD_LOGIC_VECTOR(1, 20);
              addr      <= next_addr_col_inc;
            else
              row_count <= row_count+1;
              addr      <= next_addr_row_inc;
            end if;
            xfer_level <= next_xfer_level;
          else
            burst_count <= burst_count+1;
          end if;          
        when end_of_burst =>
          state <= idle;
        when others =>
          state <= idle;
      end case;

      if (state = cmd_wait and cmd_accept = '0') or (state = resp_wait and resp_valid = '0') then
        timeout_count <= timeout_count+1;
      else
        timeout_count <= (others => '0');
      end if;
      timeout <= AND_reduce(timeout_count);



      if burst_size_reg = "00000001" and state = cmd_wait then
        last_in_burst <= '1';
      elsif burst_count = burst_size_reg_m1 then
        last_in_burst <= resp_valid;
      elsif resp_valid = '1' then
        last_in_burst <= '0';
      end if;
      
      if row_count = n_rows_i then
        row_count_eq_n_rows <= '1';
      else
        row_count_eq_n_rows <= '0';
      end if;
      
      -- Register address increments - only updates once per burst
      next_addr_col_inc <=  addr + EXT(col_inc_i & "0000", 32);
      next_addr_row_inc <=  addr + EXT(row_size_i & "0000", 32);
      next_xfer_level <= xfer_level - burst_size_reg;
      
      if EXT(burst_size_i, 29) < xfer_level then
        next_burst_size <= burst_size_i;
      else
        next_burst_size <= xfer_level(7 downto 0);
      end if;

      if state = idle and fifo_level > EXT(next_burst_size & "0", 16) and OR_reduce(next_burst_size) = '1' then
        start_burst <= '1';
      else
        start_burst <= '0';
      end if;

      for i in 0 to 10 loop
        if (i+1)<CONV_INTEGER(fgsg_size_i) then
          fgsg_shift_reg(i) <= fgsg_shift_reg(i+1);
        else
          fgsg_shift_reg(i) <= fgsg_shift_reg(0);
        end if;
      end loop;  -- i
      
      
      -- Initialise values on start = '1'
      -- Overrides any other changes to these registers
      if start = '1' then
        xfer_level   <= xfer_size;
        addr         <= start_addr;
        burst_size_i <= burst_size;
        row_size_i   <= row_size;
        n_rows_i     <= n_rows;
        col_inc_i    <= col_inc;
        row_count    <= CONV_STD_LOGIC_VECTOR(1, 20);
        burst_count  <= CONV_STD_LOGIC_VECTOR(1, 8);
        fgsg_size_i  <= fgsg_size;
        fgsg_shift_reg <= fgsg_pattern;
        fgsg_active  <= OR_reduce(fgsg_size);
        timeout_latch <= '0';
      end if;
    
      if timeout = '1' then
        timeout_latch <= '1';
      end if;
    end if;
  end process;

  resp_valid_q <= '1' when state = timeout_cleanup else resp_valid;  

  fifo_wadv    <= resp_valid_q and ((not fgsg_active) or fgsg_shift_reg(0));
  resp_accept  <= resp_valid; 
  
  cmd_addr     <= addr;
  read_cmd    <= '1' when state = cmd_wait else '0';
  burst_length <= burst_size_reg;

  xfer_status(31 downto 4) <= xfer_level;
  xfer_status(0) <= '1' when state = idle else '0';
  xfer_status(1) <= '1' when state = cmd_wait else '0';
  xfer_status(2) <= '1' when state = resp_wait else '0';
  xfer_status(3) <= timeout_latch;
  xfer_status(39 downto 32) <= burst_count;
  xfer_status(59 downto 40) <= row_count;
  
end rtl;
