--
-- ddma2ocp_rwmux_b32.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
-- DDMA2OCP_RWMUX
--
-- Demand mode DMA OCP Interface read/write MUX
--
-- Arbitrates between read and write commands


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.ocpad.all;

entity ddma2ocp_rwmux_b32 is
  
  port (
    clk          : in  std_logic;
    rst          : in  std_logic;
    timeout_disable    : in std_logic;
    -- write_port
    write_cmd    : in std_logic;
    write_addr   : in std_logic_vector(31 downto 0);
    write_cmd_accept   : out  std_logic;
    write_data   : in  std_logic_vector(31 downto 0);
    write_data_valid   : in std_logic;
    write_data_accept  : out std_logic;
    write_burst_length : in std_logic_vector(7 downto 0);
    write_data_last    : in std_logic;
    -- read_port
    read_cmd : in std_logic;
    read_addr : in std_logic_vector(31 downto 0);
    read_cmd_accept   : out  std_logic;
    read_data   : out  std_logic_vector(31 downto 0);
    read_data_valid   : out std_logic;
    read_data_accept  : in std_logic;
    read_burst_length : in std_logic_vector(7 downto 0);
    read_data_last    : out std_logic;
    -- OCP port
    ocp_bus_outputs         : out ocpad_burst32_m2sT := ocpad_burst32_default_m2s;
    ocp_bus_inputs          : in  ocpad_burst32_s2mT
    
    );

end ddma2ocp_rwmux_b32;

architecture rtl of ddma2ocp_rwmux_b32 is

  type state_type is (idle, reading, writing, err);

  signal state : state_type := idle;

  attribute max_fanout : string;
  attribute max_fanout of state : signal is "16";
  attribute max_fanout of ocp_bus_outputs : signal is "16";

  signal timeout : std_logic_vector(15 downto 0);
  
begin  -- rtl

  rst0: process (clk, rst)
  begin  -- process rst0
    if rst = '1' then      
      ocp_bus_outputs.reset_n <= '0';
    elsif clk'event and clk = '1' then  
      ocp_bus_outputs.reset_n <= not rst;
    end if;
  end process rst0;
  
  ocp_bus_outputs.Cmd <= "010" when state = reading and read_cmd = '1' else "001" when state = writing and write_cmd = '1' else "000";
  ocp_bus_outputs.Addr <= read_addr when state = reading else write_addr;
  gen_burst_length1: if read_burst_length'length < ocpad_burst_width generate
    ocp_bus_outputs.burstlength <= EXT(read_burst_length,ocpad_burst_width) when state = reading else EXT(write_burst_length,ocpad_burst_width);
  end generate gen_burst_length1;
  gen_burst_length2: if read_burst_length'length >= ocpad_burst_width generate
    ocp_bus_outputs.burstlength <= read_burst_length(ocpad_burst_width-1 downto 0) when state = reading else write_burst_length(ocpad_burst_width-1 downto 0);
  end generate gen_burst_length2;
  ocp_bus_outputs.burstsinglereq <= '1';
  ocp_bus_outputs.data <= write_data;
  ocp_bus_outputs.databyteen <= (others => '1');
  ocp_bus_outputs.datalast <= write_data_last when state = writing else '0';
  ocp_bus_outputs.datavalid <= write_data_valid when state = writing else '0';
  ocp_bus_outputs.respaccept <= read_data_accept when state = reading else '0';

  write_cmd_accept  <= ocp_bus_inputs.CmdAccept when state = writing and write_cmd = '1' else '0';
  write_data_accept <= ocp_bus_inputs.DataAccept when state = writing else '0';
  
  read_cmd_accept  <= ocp_bus_inputs.CmdAccept when state = reading and read_cmd = '1' else '0';
  read_data <= ocp_bus_inputs.Data;
  read_data_valid <= '1' when ocp_bus_inputs.Resp = "01" and state = reading else '0';
  read_data_last <= ocp_bus_inputs.RespLast when state = reading else '0'; 
  
  process (clk,rst)
  begin  -- process
    if rst = '1' then
      state <= idle;
      timeout <= (others => '0');
    elsif clk'event and clk = '1' then     -- rising clock edge      
      case state is
        when idle =>
          if read_cmd = '1' then
            state <= reading;
          elsif write_cmd = '1' then
            state <= writing;
          end if;
          timeout <= (others => '0');
        when reading =>
          if (ocp_bus_inputs.RespLast = '1' and read_data_accept = '1') or AND_reduce(timeout) = '1' then
            state <= idle;
          end if;
          if timeout_disable = '0' then
            timeout <= timeout+1;  
          end if;
        when writing =>
          if (write_data_last = '1' and ocp_bus_inputs.DataAccept = '1') or AND_reduce(timeout) = '1' then
            state <= idle;
          end if;
          if timeout_disable = '0' then
            timeout <= timeout+1;  
          end if;
        when err =>
          state <= err;
        when others =>
          state <= idle;
      end case;  
    end if;
  end process;

  
end rtl;
