--
-- localbus2ocp_bridge_r32b32.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
-- LocalBus to OCP Bridge
--
-- Connects the Local Bus to multiple OCP ports
--
-- Allows up to 7 memory mapped simple non-bursting ports
-- for configuration regsiters and low data transfer rates. (32 bit)
--
-- Also allows 2 banks of up to 4 DDMA controlled
-- high performance OCP interfaces (32 bit)


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.fifo.all;
use work.ocpad.all;
use work.fifo_if_types.all;
use work.ocp_ddma_components.all;
use work.localbus_if_types.all;
use work.slave_bus_if_types.all;

entity localbus2ocp_bridge_r32b32 is
  generic (
    number_of_ocp_nonburst : integer := 7;
    number_of_ddma_engines : integer := 2;
    number_of_ocp_burst_per_engine : integer := 4;
    la_top           : natural := 23;
    family           : family_t := family_virtex4);
  port (
    lclk             : in    std_logic;  -- Local Bus Clock
    ocp_clk_nb       : in    std_logic;  -- System Clock for non-burst
    ocp_clk_b        : in    std_logic;  -- System Clock for burst
    rst              : out   std_logic;  -- Global Reset (LRESET)
    -- OCP Interfaces
    ocp_nonburst_bus_outputs_array         : out ocpad_register32_m2sT_array(0 to number_of_ocp_nonburst-1);
    ocp_nonburst_bus_inputs_array          : in  ocpad_register32_s2mT_array(0 to number_of_ocp_nonburst-1);
    ocp_burst_bus_outputs_array         : out ocpad_burst32_m2sT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);
    ocp_burst_bus_inputs_array          : in  ocpad_burst32_s2mT_array(0 to number_of_ocp_burst_per_engine*number_of_ddma_engines-1);
    -- Local Bus Interface
    localbus_inputs  : in  localbus_inputs_type;
    localbus_outputs : out localbus_outputs_type;
    system_status    : in  std_logic_vector(31 downto 0);
    system_irq       : in  std_logic_vector(31 downto 0);
    irq              : out std_logic);

end localbus2ocp_bridge_r32b32;

architecture rtl of localbus2ocp_bridge_r32b32 is

  constant dma_ctrl_reg_base : integer := 0;
  constant dma_status_reg_base : integer := 2;
  constant ocp_nonburst_reg_base : integer := 4;
  constant ocp_burst_reg_base : integer := 12;
  constant ocp_burst_status_reg_base : integer := 6;
  constant irq_enb_addr : integer := 32;
  constant irq_clr_addr : integer := 33;
  constant irq_status_addr : integer := 1;
  
  constant number_of_control_registers : integer := 4+8+20+2;
  constant number_of_status_registers : integer := 4+8+2;
  
  signal slave_bus_multiple_outputs : slave_bus_outputs_array(0 to number_of_ocp_nonburst);
  signal slave_bus_multiple_inputs : slave_bus_inputs_array(0 to number_of_ocp_nonburst);
  signal slave_bus_inputs : slave_bus_inputs_type;
  signal slave_bus_outputs : slave_bus_outputs_type;

  signal slave_bus_rom_out,slave_bus_reg_out : slave_bus_inputs_type;
  signal sbrom_addressed : std_logic;
  
  signal rf0_ip,rf2_ip : fifo_read_port_inputs_type;
  signal rf0_op,rf2_op : fifo_read_port_outputs_type;
  signal wf1_ip,wf3_ip : fifo_write_port_inputs_type;
  signal wf1_op,wf3_op : fifo_write_port_outputs_type;

  signal control_regs : register_array(0 to number_of_control_registers-1);
  signal status_regs : register_array(0 to number_of_status_registers-1);
  signal control_write : std_logic_vector(number_of_control_registers-1 downto 0);

  signal w0_status,w1_status,r0_status,r1_status : std_logic_vector(63 downto 0);

  signal rst_i : std_logic;

  component async_pulse
    port (
      clk : in std_logic;
      I   : in std_Logic;
      O   : out std_Logic);
  end component;


  signal write_start0,write_start1,read_start0,read_start1 : std_logic;

  signal irq_status : std_logic_vector(31 downto 0);
  signal irq_rising_edge : std_logic_vector(31 downto 0);
  signal irq_enable : std_Logic_vector(31 downto 0);
  signal irq_clear : std_Logic_vector(31 downto 0);
  
    
  
  
begin  -- rtl
  assert number_of_ocp_nonburst<8 report "Maximum of 7 OCP non-burst buses" severity error;
  assert number_of_ocp_nonburst>0 report "Must have 1 or more OCP non-burst buses" severity error;
  assert number_of_ddma_engines = 1 or number_of_ddma_engines = 2 report "Must have 1 or 2 DMA engines" severity error;
  assert number_of_ocp_burst_per_engine<9 report "Maximum of 8 OCP burst buses per engine" severity error;
  assert number_of_ocp_burst_per_engine>0 report "Must have 1 or more OCP burst buses per engine" severity error;
  
  rst <= rst_i;
  
  lbif0: localbus_ddma32_interface_b32
    generic map (
      la_top => la_top,
      family => family)
    port map (
      lclk             => lclk,
      fifo_clk          => ocp_clk_b,
      rst              => rst_i,
      localbus_inputs  => localbus_inputs,
      localbus_outputs => localbus_outputs,
      slavebus_inputs  => slave_bus_inputs,
      slavebus_outputs => slave_bus_outputs,
      rfifo0_inputs      => rf0_ip,
      rfifo0_outputs     => rf0_op,
      wfifo1_inputs      => wf1_ip,
      wfifo1_outputs     => wf1_op,
      rfifo2_inputs      => rf2_ip,
      rfifo2_outputs     => rf2_op,
      wfifo3_inputs      => wf3_ip,
      wfifo3_outputs     => wf3_op,
      fifo_counter_set => control_write(dma_ctrl_reg_base+3 downto dma_ctrl_reg_base+0),
      fifo_counter0_init => control_regs(dma_ctrl_reg_base+0),
      fifo_counter1_init => control_regs(dma_ctrl_reg_base+1),
      fifo_counter2_init => control_regs(dma_ctrl_reg_base+2),
      fifo_counter3_init => control_regs(dma_ctrl_reg_base+3),
      fifo_counter0      => status_regs(dma_status_reg_base+0),
      fifo_counter1      => status_regs(dma_status_reg_base+1),
      fifo_counter2      => status_regs(dma_status_reg_base+2),
      fifo_counter3      => status_regs(dma_status_reg_base+3));


  sbm0: slavebus_mux
    generic map (
      number_of_output_buses => number_of_ocp_nonburst+1)
    port map (
      clk                        => lclk,
      slave_bus_outputs          => slave_bus_outputs,
      slave_bus_inputs           => slave_bus_inputs,
      slave_bus_multiple_outputs => slave_bus_multiple_outputs,
      slave_bus_multiple_inputs  => slave_bus_multiple_inputs);
  
  
  sbrb0: slavebus_register_bank
    generic map (
      number_of_control_registers => number_of_control_registers,
      control_base_address        => 0,
      number_of_status_registers  => number_of_status_registers,
      status_base_address         => number_of_control_registers*4)
    port map (
      clk                     => lclk,
      slave_bus_in            => slave_bus_multiple_outputs(0),
      slave_bus_out           => slave_bus_reg_out,
      control_register_values => control_regs,
      control_register_writes => control_write,
      status_register_values  => status_regs);

  sbrom0: slavebus_rom
    port map (
      clk           => lclk,
      slave_bus_in  => slave_bus_multiple_outputs(0),
      slave_bus_out => slave_bus_rom_out,
      rom_addressed => sbrom_addressed);
  
  slave_bus_multiple_inputs(0) <= slave_bus_rom_out when sbrom_addressed = '1' else slave_bus_reg_out;
  

  
  gen_slave_ocp_if: for i in 0 to number_of_ocp_nonburst-1 generate
    sb2ocp_i: slavebus2ocp32
      port map (
        lclk              => lclk,
        ocp_clk           => ocp_clk_nb,
        rst               => rst_i,
        slave_bus_outputs => slave_bus_multiple_outputs(i+1),
        slave_bus_inputs  => slave_bus_multiple_inputs(i+1),
        ocp_bus_outputs   => ocp_nonburst_bus_outputs_array(i),
        ocp_bus_inputs    => ocp_nonburst_bus_inputs_array(i),
        page_register     => control_regs(i+1+ocp_nonburst_reg_base));
  end generate gen_slave_ocp_if;
  

  dma0: ocp_ddma_engine_b32
    generic map (
      number_of_ocp_masters => number_of_ocp_burst_per_engine)
    port map (
      clk                   => ocp_clk_b,
      rst                   => rst_i,
      ocp_bus_outputs_array => ocp_burst_bus_outputs_array(0 to number_of_ocp_burst_per_engine-1),
      ocp_bus_inputs_array  => ocp_burst_bus_inputs_array(0 to number_of_ocp_burst_per_engine-1),
      rfifo_inputs          => rf0_ip,
      rfifo_outputs         => rf0_op,
      wfifo_inputs          => wf1_ip,
      wfifo_outputs         => wf1_op,
      timeout_disable       => control_regs(ocp_burst_reg_base+0)(31),
      write_burst_size      => control_regs(ocp_burst_reg_base+0)(7 downto 0),   
      write_xfer_size   => control_regs(dma_ctrl_reg_base+0)(31 downto 2), 
      write_row_size    => control_regs(ocp_burst_reg_base+1)(19 downto 0), 
      write_n_rows      => control_regs(ocp_burst_reg_base+2)(19 downto 0), 
      write_col_inc     => control_regs(ocp_burst_reg_base+3)(19 downto 0), 
      write_start_addr  => control_regs(ocp_burst_reg_base+4)(31 downto 0), 
      write_start       => write_start0, 
      write_status      => w0_status, 
      write_target      => control_regs(ocp_burst_reg_base+0)(10 downto 8), 

      read_burst_size   => control_regs(ocp_burst_reg_base+5)(7 downto 0),
      read_xfer_size    => control_regs(dma_ctrl_reg_base+1)(31 downto 2),
      read_row_size     => control_regs(ocp_burst_reg_base+6)(19 downto 0),
      read_n_rows       => control_regs(ocp_burst_reg_base+7)(19 downto 0),
      read_col_inc      => control_regs(ocp_burst_reg_base+8)(19 downto 0),
      read_start_addr   => control_regs(ocp_burst_reg_base+9)(31 downto 0), 
      read_start        => read_start0,
      read_status       => r0_status, 
      read_target       => control_regs(ocp_burst_reg_base+5)(10 downto 8),
      read_fgsg_size    => control_regs(ocp_burst_reg_base+5)(19 downto 16),
      read_fgsg_pattern => control_regs(ocp_burst_reg_base+5)(31 downto 20));

  -- Start pulses, latch data and start FSM after write strobe on address register
  ap0: async_pulse
    port map (
      clk => ocp_clk_b,
      I   => control_write(ocp_burst_reg_base+4),
      O   => write_start0);

  ap1: async_pulse
    port map (
      clk => ocp_clk_b,
      I   => control_write(ocp_burst_reg_base+9),
      O   => read_start0);
  

  gen_2nd_engine: if number_of_ddma_engines = 2 generate
    dma1: ocp_ddma_engine_b32
    generic map (
      number_of_ocp_masters => number_of_ocp_burst_per_engine)
    port map (
      clk                   => ocp_clk_b,
      rst                   => rst_i,
      ocp_bus_outputs_array => ocp_burst_bus_outputs_array(number_of_ocp_burst_per_engine to number_of_ocp_burst_per_engine*2-1),
      ocp_bus_inputs_array  => ocp_burst_bus_inputs_array(number_of_ocp_burst_per_engine to number_of_ocp_burst_per_engine*2-1),
      rfifo_inputs          => rf2_ip,
      rfifo_outputs         => rf2_op,
      wfifo_inputs          => wf3_ip,
      wfifo_outputs         => wf3_op,
      timeout_disable       => control_regs(ocp_burst_reg_base+10)(31),
      write_burst_size      => control_regs(ocp_burst_reg_base+10)(7 downto 0),   
      write_xfer_size   => control_regs(dma_ctrl_reg_base+2)(31 downto 2), 
      write_row_size    => control_regs(ocp_burst_reg_base+11)(19 downto 0), 
      write_n_rows      => control_regs(ocp_burst_reg_base+12)(19 downto 0), 
      write_col_inc     => control_regs(ocp_burst_reg_base+13)(19 downto 0), 
      write_start_addr  => control_regs(ocp_burst_reg_base+14)(31 downto 0), 
      write_start       => write_start1,
      write_status      => w1_status, 
      write_target      => control_regs(ocp_burst_reg_base+10)(10 downto 8), 
      read_burst_size   => control_regs(ocp_burst_reg_base+15)(7 downto 0),
      read_xfer_size    => control_regs(dma_ctrl_reg_base+3)(31 downto 2),
      read_row_size     => control_regs(ocp_burst_reg_base+16)(19 downto 0),
      read_n_rows       => control_regs(ocp_burst_reg_base+17)(19 downto 0),
      read_col_inc      => control_regs(ocp_burst_reg_base+18)(19 downto 0),
      read_start_addr   => control_regs(ocp_burst_reg_base+19)(31 downto 0), 
      read_start        => read_start1,
      read_status       => r1_status, 
      read_target       => control_regs(ocp_burst_reg_base+15)(10 downto 8),
      read_fgsg_size    => control_regs(ocp_burst_reg_base+15)(19 downto 16),
      read_fgsg_pattern => control_regs(ocp_burst_reg_base+15)(31 downto 20));

     -- Start pulses, latch data and start FSM after write strobe on address register
    ap2: async_pulse
      port map (
        clk => ocp_clk_b,
        I   => control_write(ocp_burst_reg_base+14),
        O   => write_start1);

    ap3: async_pulse
      port map (
        clk => ocp_clk_b,
        I   => control_write(ocp_burst_reg_base+19),
        O   => read_start1);

    
  end generate gen_2nd_engine;
  
  status_regs(ocp_burst_status_reg_base+0) <= w0_status(31 downto 0);
  status_regs(ocp_burst_status_reg_base+1) <= w0_status(63 downto 32);
  status_regs(ocp_burst_status_reg_base+2) <= r0_status(31 downto 0);
  status_regs(ocp_burst_status_reg_base+3) <= r0_status(63 downto 32);
  status_regs(ocp_burst_status_reg_base+4) <= w1_status(31 downto 0);
  status_regs(ocp_burst_status_reg_base+5) <= w1_status(63 downto 32);
  status_regs(ocp_burst_status_reg_base+6) <= r1_status(31 downto 0);
  status_regs(ocp_burst_status_reg_base+7) <= r1_status(63 downto 32);
  
  status_regs(0) <= system_status;
  status_regs(irq_status_addr) <= irq_status;

  irq_clear <= control_regs(irq_clr_addr);
  irq_enable <= control_regs(irq_enb_addr);

  
  edge_detect_irq: for i in 0 to 31 generate
    ap_i: async_pulse
      port map (
        clk => lclk,
        I   => system_irq(i),
        O   => irq_rising_edge(i));
  end generate edge_detect_irq;

  latch_irq: process (lclk, rst_i)
  begin  -- process latch_irq
    if rst_i = '1' then                   -- asynchronous reset
      irq_status <= (others => '0');
      irq <= '0';
    elsif lclk'event and lclk = '1' then  -- rising clock edge
      for i in 0 to 31 loop
        if control_write(irq_clr_addr) = '1' and irq_clear(i) = '1' then
          irq_status(i) <= '0';
        elsif irq_rising_edge(i) = '1' and irq_enable(i) = '1' then
          irq_status(i) <= '1';
        end if;
      end loop;  -- i

      -- Bring IRQ line low when clearing status bits
      -- Will generate a new interrupt if other source is active
      if control_write(irq_clr_addr) = '1' then
        irq <= '0';
      else
        irq <= OR_reduce(irq_status);
      end if;           
    end if;
  end process latch_irq;

  

  
end rtl;
