--
-- localbus_ddma64_interface_b64.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
-- localbus_ddma64_interface
--
-- Provides an interface between the Local bus and
-- Asynchronous DMA FIFO ports and a simple non-bursting
-- slave bus
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.fifo.all;
use work.localbus.all;
use work.localbus_if_types.all;
use work.slave_bus_if_types.all;
use work.fifo_if_types.all;

entity localbus_ddma64_interface_b64 is
  generic(
    la_top           :     natural := 23;
    family           :     family_t := family_virtex4;
    dma_mapping      :     integer := 0;
    bus_width        :     integer := 64);
  port(
    -- Clocking
    lclk             : in    std_logic;  -- Local Bus Clock
    fifo_clk         : in    std_logic;  -- FIFO Clock
    rst              : out   std_logic;  -- Global Reset (LRESET)
    -- Local Bus Interface
    localbus_inputs  : in  localbus_inputs_type;
    localbus_outputs : out localbus_outputs_type;
    -- Slave Bus Interface - synchronous to LCLK
    slavebus_outputs : out slave_bus_outputs_type;
    slavebus_inputs  : in  slave_bus_inputs_type;
    -- 4 x FIFO Interfaces connected to 4 DMA channels
    rfifo0_inputs    : in  fifo_read_port_inputs_type;
    rfifo0_outputs   : out fifo_read_port_outputs_type;
    wfifo1_inputs    : in  fifo_write_port_inputs_type;
    wfifo1_outputs   : out fifo_write_port_outputs_type;
    rfifo2_inputs    : in  fifo_read_port_inputs_type;
    rfifo2_outputs   : out fifo_read_port_outputs_type;
    wfifo3_inputs    : in  fifo_write_port_inputs_type;
    wfifo3_outputs   : out fifo_write_port_outputs_type;
    fifo_counter_set : in  std_logic_vector(3 downto 0);
    fifo_counter0_init : in std_logic_vector(31 downto 0);
    fifo_counter1_init : in std_logic_vector(31 downto 0);
    fifo_counter2_init : in std_logic_vector(31 downto 0);
    fifo_counter3_init : in std_logic_vector(31 downto 0);
    fifo_counter0 : out std_logic_vector(31 downto 0);
    fifo_counter1 : out std_logic_vector(31 downto 0);
    fifo_counter2 : out std_logic_vector(31 downto 0);
    fifo_counter3 : out std_logic_vector(31 downto 0));      
end entity;

architecture mixed of localbus_ddma64_interface_b64 is
    --
    -- This value is the number of words / free locations that the design attempts to aggregate
    -- in the FIFO before attempting a demand-mode DMA burst. This is intended to increase the
    -- efficiency of the local bus by increasing the average burst length for demand-mode DMA
    -- transfers.
    --

    constant threshold : natural := 32;

    signal rst_i : std_logic;
    signal lads_i : std_logic;
    signal lblast_i : std_logic;
    signal lbe_iq_l : std_logic_vector(7 downto 0);
    signal lwrite_i : std_logic;
    signal ldack_i : std_logic_vector(3 downto 0);
    signal ldack_q : std_logic_vector(3 downto 0);
    signal ld_iq : std_logic_vector(63 downto 0);
    signal ld_o : std_logic_vector(63 downto 0);
    signal la_i : std_logic_vector(la_top downto 2);
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal lbterm_i : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal ldreq_o_l : std_logic_vector(3 downto 0);
    signal ld_oe_l : std_logic;

    signal fholda : std_logic;
    
    signal la_q : std_logic_vector(la_top downto 2);

    signal dd_qlads : std_logic_vector(3 downto 0);
    signal ds_qlads : std_logic;
    signal ds_idle : std_logic;
    signal ds_decode : std_logic;
    signal ds_write : std_logic;
    signal ds_xfer : std_logic;
    signal dd_idle : std_logic_vector(3 downto 0);
    signal dd_idle_q : std_logic_vector(3 downto 0);
    signal dd_idle_qq : std_logic_vector(3 downto 0);
    signal ds_ready : std_logic;
    signal ds_stop : std_logic;

    signal fifo0_e : std_logic;
    signal fifo0_ne : std_logic;
    signal fifo0_read : std_logic;
    signal fifo0_radvt : std_logic;
    signal fifo0_rev : std_logic;
    signal fifo0_f : std_logic;
    signal fifo0_nf : std_logic;
    signal fifo0_write : std_logic;
    signal fifo0_level : std_logic_vector(15 downto 0);
    signal fifo0_q : std_logic_vector(63 downto 0);
    
    signal fifo1_e : std_logic;
    signal fifo1_ne : std_logic;
    signal fifo1_read : std_logic;
    signal fifo1_radvt : std_logic;
    signal fifo1_rev : std_logic;
    signal fifo1_f : std_logic;
    signal fifo1_nf : std_logic;
    signal fifo1_write : std_logic;
    signal fifo1_level : std_logic_vector(15 downto 0);
    signal fifo1_q : std_logic_vector(63 downto 0);
    signal fifo1_data : std_logic_vector(63 downto 0);
    signal afifo1_empty : std_logic;
    
    signal fifo2_e : std_logic;
    signal fifo2_ne : std_logic;
    signal fifo2_read : std_logic;
    signal fifo2_radvt : std_logic;
    signal fifo2_rev : std_logic;
    signal fifo2_f : std_logic;
    signal fifo2_nf : std_logic;
    signal fifo2_write : std_logic;
    signal fifo2_level : std_logic_vector(15 downto 0);
    signal fifo2_q : std_logic_vector(63 downto 0);

    signal fifo3_e : std_logic;
    signal fifo3_ne : std_logic;
    signal fifo3_read : std_logic;
    signal fifo3_radvt : std_logic;
    signal fifo3_rev : std_logic;
    signal fifo3_f : std_logic;
    signal fifo3_nf : std_logic;
    signal fifo3_write : std_logic;
    signal fifo3_level : std_logic_vector(15 downto 0);
    signal fifo3_q : std_logic_vector(63 downto 0);
    signal fifo3_data  : std_logic_vector(63 downto 0);  
    signal afifo3_empty : std_logic;
    
    signal dd_req : std_logic_vector(3 downto 0);
    signal dd_burst : std_logic_vector(3 downto 0);

    
    signal i0count : std_logic_vector(28 downto 0);
    signal i0count0 : std_logic;
    signal i0countn : std_logic;
    signal wrote_i0count : std_logic;
    signal o1count : std_logic_vector(28 downto 0);
    signal o1count0 : std_logic;
    signal o1countn : std_logic;
    signal wrote_o1count: std_logic;

    signal i2count : std_logic_vector(28 downto 0);
    signal i2count0 : std_logic;
    signal i2countn : std_logic;
    signal wrote_i2count : std_logic;
    signal o3count : std_logic_vector(28 downto 0);
    signal o3count0 : std_logic;
    signal o3countn : std_logic;
    signal wrote_o3count: std_logic;

    signal logic0 : std_logic;
    signal logic1 : std_logic;

    signal ready_timeout : std_logic;
    signal ready_timeout_count : std_logic_vector(7 downto 0);

    signal ld_o_slavebus : std_logic_vector(63 downto 0);
    signal slavebus_rd  : std_logic;
    signal slavebus_wr : std_logic;

    signal fifo_counter_set_remap :  std_logic_vector(3 downto 0);
    signal fifo_counter0_init_remap :  std_logic_vector(31 downto 0);
    signal fifo_counter1_init_remap :  std_logic_vector(31 downto 0);
    signal fifo_counter2_init_remap :  std_logic_vector(31 downto 0);
    signal fifo_counter3_init_remap : std_logic_vector(31 downto 0);
    signal fifo_counter0_remap :  std_logic_vector(31 downto 0);
    signal fifo_counter1_remap :  std_logic_vector(31 downto 0);
    signal fifo_counter2_remap :  std_logic_vector(31 downto 0);
    signal fifo_counter3_remap :  std_logic_vector(31 downto 0);

    signal ldack_i_remap,ldreq_o_l_remap : std_logic_vector(3 downto 0);

    signal fifo0_ld_iq : std_logic_vector(63 downto 0);
    signal fifo2_ld_iq : std_logic_vector(63 downto 0);
    signal fifo0_write_q : std_logic;
    signal fifo2_write_q : std_logic;
    signal fifo0_write_toggle : std_logic := '0';
    signal fifo2_write_toggle : std_logic := '0';
 
    signal fifo0_ld_iq_reg : std_logic_vector(31 downto 0);
    signal fifo2_ld_iq_reg : std_logic_vector(31 downto 0);
    signal afifo1_read : std_logic;
    signal afifo3_read : std_logic;
    signal afifo1_qdata : std_logic_vector(63 downto 0);
    signal afifo3_qdata : std_logic_vector(63 downto 0);

    signal fifo1_read_toggle : std_logic := '0';
    signal fifo3_read_toggle : std_logic := '0';

    attribute iob : string;
    attribute iob of lbe_iq_l : signal is "true";
    attribute iob of ld_iq : signal is "true";
    
begin

  assert bus_width = 64 or bus_width = 32 report "Bus Width must be 32 or 64 bits" severity error;
  
  assert dma_mapping = 0 or dma_mapping = 1 or dma_mapping = 2 report "dma_mapping must be 0,1 or 2" severity error;
  -- DMA FIFO remapping
  -- Mapping  DMA Channel      0    1    2    3
  --    0     FIFO             W0   R1   W2   R3
  --    1     FIFO             R1   R3   W0   W2
  --    2     FIFO             W0   W2   R1   R3
  --
  -- LDREQ and LDACK lines are remapped
  -- Control registers are also remapped, so that  fifo_counter0_init
  -- always corresponds to DMA channel 0, etc.
  
  gen_mapping0: if dma_mapping = 0 generate
    ldack_i <= ldack_i_remap;
    ldreq_o_l_remap <= ldreq_o_l;
    fifo_counter_set_remap <= fifo_counter_set;
    fifo_counter0_init_remap <= fifo_counter0_init;
    fifo_counter1_init_remap <= fifo_counter1_init;
    fifo_counter2_init_remap <= fifo_counter2_init;
    fifo_counter3_init_remap <= fifo_counter3_init;
    fifo_counter0 <= fifo_counter0_remap;
    fifo_counter1 <= fifo_counter1_remap;
    fifo_counter2 <= fifo_counter2_remap;
    fifo_counter3 <= fifo_counter3_remap;
  end generate gen_mapping0;

  gen_mapping1: if dma_mapping = 1 generate
    ldack_i <= ldack_i_remap(1) & ldack_i_remap(3) & ldack_i_remap(0) & ldack_i_remap(2); 
    ldreq_o_l_remap <= ldreq_o_l(2) & ldreq_o_l(0) & ldreq_o_l(3) & ldreq_o_l(1);
    fifo_counter_set_remap <= fifo_counter_set(1) & fifo_counter_set(3) & fifo_counter_set(0) & fifo_counter_set(2);
    fifo_counter0_init_remap <= fifo_counter2_init;
    fifo_counter1_init_remap <= fifo_counter0_init;
    fifo_counter2_init_remap <= fifo_counter3_init;
    fifo_counter3_init_remap <= fifo_counter1_init;
    fifo_counter0 <= fifo_counter1_remap;
    fifo_counter1 <= fifo_counter3_remap;
    fifo_counter2 <= fifo_counter0_remap;
    fifo_counter3 <= fifo_counter2_remap;
  end generate gen_mapping1;

  gen_mapping2: if dma_mapping = 2 generate
    ldack_i <= ldack_i_remap(3) & ldack_i_remap(1) & ldack_i_remap(2) & ldack_i_remap(0);
    ldreq_o_l_remap <= ldreq_o_l(3) & ldreq_o_l(1) & ldreq_o_l(2) & ldreq_o_l(0); 
    fifo_counter_set_remap <= fifo_counter_set(3) & fifo_counter_set(1) & fifo_counter_set(2) & fifo_counter_set(0);
    fifo_counter0_init_remap <= fifo_counter0_init;
    fifo_counter1_init_remap <= fifo_counter2_init;
    fifo_counter2_init_remap <= fifo_counter1_init;
    fifo_counter3_init_remap <= fifo_counter3_init;
    fifo_counter0 <= fifo_counter0_remap;
    fifo_counter1 <= fifo_counter2_remap;
    fifo_counter2 <= fifo_counter1_remap;
    fifo_counter3 <= fifo_counter3_remap;
  end generate gen_mapping2;

  
  
    logic0 <= '0';
    logic1 <= '1';

    rst <= rst_i;
    
    --
    -- Convert the inputs to active high.
    --

    rst_i      <= not localbus_inputs.lreset_l;
    lads_i     <= not localbus_inputs.lads_l;
    lblast_i   <= not localbus_inputs.lblast_l;
    lbterm_i   <= not localbus_inputs.lbterm_l_i;
    lwrite_i   <= localbus_inputs.lwrite;
    la_i       <= localbus_inputs.la_i(la_top downto 2);
    ldack_i_remap    <= not localbus_inputs.ldack_l;
    fholda     <= localbus_inputs.fholda;    
  
    --
    -- Register LD and LBE# (should be mapped into IOBs)
    --

    gen_lbe_iq_l : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            lbe_iq_l <= (others => '0');
        elsif lclk'event and lclk = '1' then
            lbe_iq_l <= localbus_inputs.lbe_l;
        end if;
    end process;

    gen_ld_iq : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            ld_iq <= (others => '0');
        elsif lclk'event and lclk = '1' then
            ld_iq <= localbus_inputs.lad_i;
        end if;
    end process;

    --
    -- Output active low signals.
    --

    localbus_outputs.ldreq_l <= ldreq_o_l_remap;

    --
    -- Generate qualified versions of 'lads_l'.
    --
    -- For direct slave access, this is when 'lads_l' is asserted AND the FPGA
    -- is being addressed AND the FPGA is not the local bus master AND the
    -- cycle is not demand-mode DMA.
    --
    -- For demand-mode DMA access, this is when 'lads_l' is asserted AND the
    -- FPGA is being addressed and the FPGA is not the local bus master. The
    -- demand-mode DMA acknowledge signals are checked within the 'plxddsm'
    -- module.
    --

    ds_qlads <= lads_i and not la_i(la_top) and not fholda;
    gen_dd_qlads : for i in ldack_i'range generate
        dd_qlads(i) <= lads_i and not la_i(la_top) and not fholda and ldack_i(i);
    end generate;

    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    localbus_outputs.lbterm_l_o <= lbterm_o_l;
    localbus_outputs.lbterm_l_t <= lbterm_oe_l;

    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --

    localbus_outputs.lready_l_o <= lready_o_l;
    localbus_outputs.lready_l_t <= lready_oe_l;
   
    --
    -- Drive the local data bus on a read.
    --

    localbus_outputs.lad_o <= ld_o;
    localbus_outputs.lad_t <= (others => ld_oe_l);

    --
    -- Latch 'lad' on assertion of 'lads_l', in order to improve timing. There
    -- is no need to increment the address, because we don't allow bursting for
    -- register addresses, and DMA transfers that burst are expected to use
    -- constant local addressing.
    --

    latch_address : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            la_q <= (others => '0');
            ldack_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
                ldack_q <= ldack_i;
            end if;
        end if;
    end process;

    --
    -- Generate 'ld_o', data to be output on local bus.
    --
    -- Depending on whether this is a demand-mode DMA burst or a normal direct
    -- slave burst, and the local bus address, we output one of
    --
    --   o FIFO data 'fifo_q'
    --   o 'icount' and 'ocount' register values
    --

    gen_ld_o : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            ld_o <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if ldack_q(1) = '1' then
              ld_o <= fifo1_q;
            elsif ldack_q(3) = '1' then
              ld_o <= fifo3_q;
            else
              ld_o <= ld_o_slavebus;
            end if;
        end if;
    end process;

    --
    -- This process implements the DMA transfer count registers and flags for
    -- each direction.
    --
    --    icount          number of words remaining in PCI-to-local DMA transfer
    --    icount0         '1' => 'icount' is zero
    --    icountn         '1' => 'icount' is less than or equal to 'threshold'
    --    ocount          number of words remaining in local-to-PCI DMA transfer
    --    ocount0         '1' => 'ocount' is zero
    --    ocountn         '1' => 'ocount' is less than or equal to 'threshold'
    --
    -- Prior to the host kicking off a demand-mode DMA transfer, the host must
    -- program the transfer count registers must be programmed with the length
    -- of the DMA transfer being performed.
    --

    gen_counts: process(rst_i, lclk)
    begin
      if rst_i = '1' then

        i0count  <= (others => '0');
        i0countn <= '1';
        i0count0 <= '1';
        wrote_i0count <= '0';

        o1count  <= (others => '0');
        o1countn <= '1';
        o1count0 <= '1';
        wrote_o1count <= '0';
        
      elsif lclk'event and lclk = '1' then
        if fifo_counter_set_remap(0) = '1' then
          if bus_width = 64 then
            i0count(28 downto 0) <= fifo_counter0_init_remap(31 downto 3);
          else
            i0count(28 downto 0) <= fifo_counter0_init_remap(30 downto 2);
          end if;          
        else
          if fifo0_write = '1' then
            i0count <= i0count - 1;
          end if;
        end if;

        wrote_i0count <= fifo_counter_set_remap(0);

        
        if fifo_counter_set_remap(1) = '1' then
          if bus_width = 64 then
            o1count(28 downto 0) <= fifo_counter1_init_remap(31 downto 3);  
          else
            o1count(28 downto 0) <= fifo_counter1_init_remap(30 downto 2);  
          end if;                
        else
          if fifo1_read = '1' then
            o1count <= o1count - 1;
          end if;
        end if;

        wrote_o1count <= fifo_counter_set_remap(1);
        if wrote_i0count = '1' then
          if i0count <= threshold then
            i0countn <= '1';
          else
            i0countn <= '0';
          end if;
          if i0count = 0 then
            i0count0 <= '1';
          else
            i0count0 <= '0';
          end if;
        else
          if fifo0_write = '1' then
            if i0count = (threshold + 1) then
              i0countn <= '1';
            end if;
            if i0count = 1 then
              i0count0 <= '1';
            end if;
          end if;
        end if;

        if wrote_o1count = '1' then
          if o1count <= threshold then
            o1countn <= '1';
          else
            o1countn <= '0';
          end if;
          if o1count = 0 then
            o1count0 <= '1';
          else
            o1count0 <= '0';
          end if;
        else
          if fifo1_read = '1' then
            if o1count = (threshold + 1) then
              o1countn <= '1';
            end if;
            if o1count = 1 then
              o1count0 <= '1';
            end if;
          end if;
        end if;
      end if;
    end process;


    -- Idenitical process for counters for DMA channels 2 and 3

    gen_counts32: process(rst_i, lclk)
    begin
      if rst_i = '1' then

        i2count  <= (others => '0');
        i2countn <= '1';
        i2count0 <= '1';
        wrote_i2count <= '0';

        o3count  <= (others => '0');
        o3countn <= '1';
        o3count0 <= '1';
        wrote_o3count <= '0';
        
      elsif lclk'event and lclk = '1' then
        if fifo_counter_set_remap(2) = '1' then
          if bus_width = 64 then
            i2count(28 downto 0) <= fifo_counter2_init_remap(31 downto 3);
          else
            i2count(28 downto 0) <= fifo_counter2_init_remap(30 downto 2);
          end if;        
        else
          if fifo2_write = '1' then
            i2count <= i2count - 1;
          end if;
        end if;

        wrote_i2count <= fifo_counter_set_remap(2);

        
        if fifo_counter_set_remap(3) = '1' then
          if bus_width =64 then
            o3count(28 downto 0) <= fifo_counter3_init_remap(31 downto 3);    
          else
            o3count(28 downto 0) <= fifo_counter3_init_remap(30 downto 2);               
          end if;
        else
          if fifo3_read = '1' then
            o3count <= o3count - 1;
          end if;
        end if;

        wrote_o3count <= fifo_counter_set_remap(3);
        if wrote_i2count = '1' then
          if i2count <= threshold then
            i2countn <= '1';
          else
            i2countn <= '0';
          end if;
          if i2count = 0 then
            i2count0 <= '1';
          else
            i2count0 <= '0';
          end if;
        else
          if fifo2_write = '1' then
            if i2count = (threshold + 1) then
              i2countn <= '1';
            end if;
            if i2count = 1 then
              i2count0 <= '1';
            end if;
          end if;
        end if;

        if wrote_o3count = '1' then
          if o3count <= threshold then
            o3countn <= '1';
          else
            o3countn <= '0';
          end if;
          if o3count = 0 then
            o3count0 <= '1';
          else
            o3count0 <= '0';
          end if;
        else
          if fifo3_read = '1' then
            if o3count = (threshold + 1) then
              o3countn <= '1';
            end if;
            if o3count = 1 then
              o3count0 <= '1';
            end if;
          end if;
        end if;
      end if;
    end process;


    fifo_counter0_remap <= i0count & "0" & i0countn & i0count0;
    fifo_counter1_remap <= o1count & "0" & o1countn & o1count0;
    fifo_counter2_remap <= i2count & "0" & i2countn & i2count0;
    fifo_counter3_remap <= o3count & "0" & o3countn & o3count0;

    
    --
    -- Delay the 'idle' signals from each of the demand-mode DMA state machines
    -- by one and two cycles.
    --

    gen_dd_idle_qq : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            dd_idle_q <= (others => '1');
            dd_idle_qq <= (others => '1');
        elsif lclk'event and lclk = '1' then
            for i in ldreq_o_l'range loop
               dd_idle_q(i) <= dd_idle(i);
               dd_idle_qq(i) <= dd_idle_q(i);
            end loop;
        end if;
    end process;

    --
    -- For the PCI->local direction, we wait until the FIFO has space to hold a
    -- decent-sized burst before requesting a demand-mode DMA burst. We need to
    -- pause if the FIFO becomes nearly-full.
    --
    -- The 'dd_idle_qq(0)' term takes care of a corner case. At the end of a demand-
    -- mode DMA local bus cycle, it may take a couple of clock cycles for the FIFO
    -- flags and level to be fully updated to reflect how much data has just been
    -- entered into the FIFO. Therefore, we force 'dd_req(0)' deasserted for two
    -- cycles at the end of each demand-mode DMA local bus cycle.
    --
    -- This could be optimized by only forcing 'dd_req(0)' deasserted when the FIFO
    -- level is known to be above a certain threshold, but for simplicity we do it
    -- unconditionally.
    --

    --  N.B. async_fifo write fifo_level indicates number of spaces free in FIFO
    
    dd_req(0) <= '1' when i0count0 = '0' and dd_idle_qq(0) = '1' and fifo0_level > (threshold) else
                 '0';
    dd_burst(0) <= not fifo0_nf;

    dd_req(2) <= '1' when i2count0 = '0' and dd_idle_qq(2) = '1' and fifo2_level > (threshold) else
                 '0';
    dd_burst(2) <= not fifo2_nf;

    --
    -- For the local->PCI direction, we request a demand-mode DMA burst only if the
    -- DMA transfer is not complete ('ocount' is nonzero) and either:
    --
    --      (i) the FIFO contains enough data for a decent-sized burst, or
    --
    --     (ii) we're near the end of the DMA transfer AND the FIFO level
    --          is equal to or greater than the remaining transfer count.
    --
    -- The 'dd_idle_q(1)' term takes care of a corner case. At the end of a demand-
    -- mode DMA local bus cycle, it may take an extra clock cycle for the FIFO flags
    -- and level to be fully updated to reflect how much data has just been read out
    -- of the FIFO. Therefore, we force 'dd_req(1)' deasserted for one cycle at the
    -- end of each demand-mode DMA local bus cycle.
    --
    -- This could be optimized by only forcing 'dd_req(1)' deasserted when the FIFO
    -- level is known to be below a certain threshold, but for simplicity we do it
    -- unconditionally.
    --

    dd_req(1) <= '1' when (dd_idle_q(1) = '1' and o1count0 = '0' and fifo1_level >= threshold) or
                          (dd_idle_q(1) = '1' and o1count0 = '0' and o1countn = '1' and fifo1_level >= o1count(15 downto 0)) else
                 '0';
    dd_burst(1) <= not fifo1_ne;

    dd_req(3) <= '1' when (dd_idle_q(3) = '1' and o3count0 = '0' and fifo3_level >= threshold) or
                          (dd_idle_q(3) = '1' and o3count0 = '0' and o3countn = '1' and fifo3_level >= o3count(15 downto 0)) else
                 '0';
    dd_burst(3) <= not fifo3_ne;

    --
    -- Writing into the FIFO only happens when data is written on PLX DMA channel 0/2.
    -- Because the data and byte enables from the local bus are registered, the
    -- 'fifo_write' signal must also be registered.
    --

    gen_fifo_write : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            fifo0_write <= '0';
            fifo2_write <= '0';
        elsif lclk'event and lclk = '1' then
            fifo0_write <= ds_xfer and ds_write and ldack_q(0);
            fifo2_write <= ds_xfer and ds_write and ldack_q(2);
        end if;
    end process;

    --
    -- Advancing the FIFO permanent read pointer only happens when data is
    -- read on PLX DMA channel 1/3.
    --

    fifo1_read <= ds_xfer and not ds_write and ldack_q(1);
    fifo3_read <= ds_xfer and not ds_write and ldack_q(3);

    --
    -- Advance the FIFO temporary read pointer during a burst. Because we are
    -- pipelining 'ld_o' we require the 'ds_decode' term in order to clock data
    -- through to 'ld_o' at the beginning of a read burst.
    --

    fifo1_radvt <= (ds_xfer or ds_decode) and ldack_q(1);
    fifo3_radvt <= (ds_xfer or ds_decode) and ldack_q(3);

    --
    -- We make the FIFO's temporary read pointer revert to its permanent
    -- read pointer at the end of a burst.
    --

    fifo1_rev <= dd_idle(1) and not dd_idle_q(1);
    fifo3_rev <= dd_idle(3) and not dd_idle_q(3);

    --
    -- Generate 'ready' and 'stop' signals, which control the state machines
    -- that respond to local bus cycles.
    --

    --_ready <= '1';                               -- We're always ready
    ds_stop <= not OR_reduce(ldack_q) or           -- Prevent bursting if not demand-mode DMA
               (ldack_q(0) and not dd_burst(0)) or -- Stop when demand-mode DMA and write FIFO is nearly full
               (ldack_q(1) and not dd_burst(1)) or -- Stop when demand-mode DMA and read FIFO is nearly empty
               (ldack_q(2) and not dd_burst(2)) or -- Stop when demand-mode DMA and write FIFO is nearly full
               (ldack_q(3) and not dd_burst(3)); -- Stop when demand-mode DMA and read FIFO is nearly empty

    --
    -- Instantiate the state machines that control the LDREQ# lines.
    --

    gen_plxddsm : for i in ldreq_o_l'range generate
        U0 : plxddsm
            port map(
            clk => lclk,
            rst => rst_i,
            sr => logic0,
            qlads => dd_qlads(i),
            lblast => lblast_i,
            lbterm => lbterm_i,
            request => dd_req(i),
            ready => ds_ready, -- MUST be same as fed into 'plxdssm'
            stop => ds_stop,   -- MUST be same as fed into 'plxdssm'
            idle => dd_idle(i),
            ldreq_o_l => ldreq_o_l(i));
    end generate;

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    plxdssm0: plxdssm
        port map(
            clk => lclk,
            rst => rst_i,
            sr => logic0,
            qlads => ds_qlads,
            lblast => lblast_i,
            lbterm => lbterm_i,
            lwrite => lwrite_i,
            eld_oe => open,
            ld_oe_l => ld_oe_l,
            lready_o_l => lready_o_l,
            lready_oe_l => lready_oe_l,
            lbterm_o_l => lbterm_o_l,
            lbterm_oe_l => lbterm_oe_l,
            idle => ds_idle,
            transfer => ds_xfer,
            decode => ds_decode,
            write => ds_write,
            ready => ds_ready, -- MUST be same as fed into 'plxddsm'
            stop => ds_stop);  -- MUST be same as fed into 'plxddsm'

    --
    -- 'rpel' = 2 => nearly-empty is 2 or less words present the FIFO.
    -- 'wpfl' = 3 => nearly-full is 3 or less free locations remaining in the FIFO.
    --

    -- Use prefetch sychronous FIFO and asynchronous FIFO in chain in
    -- local->PCI direction (host read)
    -- Use single asynchoronous FIFO in PCI->local direction (host write)

  -- If bus width 64, no 64-32 bit conversion required
  gen_bw64: if bus_width = 64 generate
    -- Write side
    fifo0_ld_iq <= ld_iq;
    fifo2_ld_iq <= ld_iq;
    fifo0_write_q <= fifo0_write;
    fifo2_write_q <= fifo2_write;

    -- Read side
    fifo1_write <= not afifo1_empty and not fifo1_nf;
    fifo3_write <= not afifo3_empty and not fifo3_nf;
    fifo1_data <= afifo1_qdata;
    fifo3_data <= afifo3_qdata;
    afifo1_read <= fifo1_write;
    afifo3_read <= fifo3_write;
  end generate gen_bw64;

  -- if bus width 32, 32 - 64 bit conversion required
  -- Put conversion at input of write FIFO
  -- and between read FIFOs
  gen_bw32: if bus_width = 32 generate
    process (lclk, rst_i)
    begin  -- process
      if rst_i = '1' then   
        fifo0_write_toggle <= '0';
        fifo2_write_toggle <= '0';
        fifo0_ld_iq_reg <= (others => '0');
        fifo2_ld_iq_reg <= (others => '0');
        fifo1_read_toggle <= '0';
        fifo3_read_toggle <= '0';
      elsif lclk'event and lclk = '1' then  -- rising clock edge
        if fifo0_write = '1' then
          fifo0_write_toggle <= not fifo0_write_toggle;
          fifo0_ld_iq_reg <= ld_iq(31 downto 0);
        end if;
        if fifo2_write = '1' then
          fifo2_write_toggle <= not fifo2_write_toggle;
          fifo2_ld_iq_reg <= ld_iq(31 downto 0);
        end if;
        if fifo1_write = '1' then
          fifo1_read_toggle <= not fifo1_read_toggle;
        end if;
        if fifo3_write = '1' then
          fifo3_read_toggle <= not fifo3_read_toggle;
        end if;
      end if;
    end process;

    fifo0_ld_iq <= ld_iq(31 downto 0) & fifo0_ld_iq_reg;
    fifo2_ld_iq <= ld_iq(31 downto 0) & fifo2_ld_iq_reg;
    fifo0_write_q <= fifo0_write and fifo0_write_toggle;
    fifo2_write_q <= fifo2_write and fifo2_write_toggle;

    -- Always write 2 words into output FIFO
    fifo1_write <= (not afifo1_empty and not fifo1_nf) or fifo1_read_toggle;
    fifo3_write <= (not afifo3_empty and not fifo3_nf) or fifo3_read_toggle;
    fifo1_data(31 downto 0) <= afifo1_qdata(31 downto 0) when fifo1_read_toggle = '0' else afifo1_qdata(63 downto 32);
    fifo3_data(31 downto 0) <= afifo3_qdata(31 downto 0) when fifo3_read_toggle = '0' else afifo3_qdata(63 downto 32);
    afifo1_read <= fifo1_write and fifo1_read_toggle;
    afifo3_read <= fifo3_write and fifo3_read_toggle;   
    
  end generate gen_bw32;
  
    afifo0 : afifo
        generic map(
            width => 64,
            order => 9,
            wpfl => 3,
            wpfl2 => 0,
            rpel => 2,
            family => family,
            safety => false,
            fflush => false)
        port map(
            rst => rst_i,
            wk => lclk,
            wd => fifo0_ld_iq,
            wadv => fifo0_write_q,
            we => open,
            wpf => fifo0_nf,
            wpf2 => open,
            wf => fifo0_f,
            wovf => open,
            wlevel => fifo0_level,
            rk => fifo_clk,
            rq => rfifo0_outputs.data,
            radv => rfifo0_inputs.rd,
            rflush => '0',
            re => rfifo0_outputs.empty,
            rpe => rfifo0_outputs.nempty,
            rf => open,
            rudf => open,
            rlevel => rfifo0_outputs.level);

    
    sfifop1 : sfifop
	generic map(
            width      => 64,
            order      => 9,
            rpel       => 2,
            wpfl       => 3,
            family     => family,
            safety     => false)
        port map(
            rst        => rst_i,
            k          => lclk,
            wd         => fifo1_data,
            wadv       => fifo1_write,
            we         => open,
            wpf        => fifo1_nf,
            wf         => fifo1_f,
            wovf       => open,
            rq         => fifo1_q,
            radv       => fifo1_read,
            radvt      => fifo1_radvt,
            rrev       => fifo1_rev,
            rflush     => logic0,
            re         => fifo1_e,
            rf         => open,
            rpe        => fifo1_ne,
            rudf       => open,
            rlevel     => fifo1_level);

   
    
    afifo1 : afifo
        generic map(
            width => 64,
            order => 9,
            wpfl => 3,
            wpfl2 => 0,
            rpel => 2,
            family => family,
            safety => false,
            fflush => false)
        port map(
            rst => rst_i,
            wk => fifo_clk,
            wd => wfifo1_inputs.data,
            wadv => wfifo1_inputs.wr,
            we => open,
            wpf => wfifo1_outputs.nfull,
            wpf2 => open,
            wf => wfifo1_outputs.full,
            wovf => open,
            wlevel => wfifo1_outputs.level,
            rk => lclk,
            rq => afifo1_qdata,
            radv => afifo1_read,
            rflush => '0',
            re => afifo1_empty,
            rpe => open,
            rf => open,
            rudf => open,
            rlevel => open);


    

    afifo2 : afifo
        generic map(
            width => 64,
            order => 9,
            wpfl => 3,
            wpfl2 => 0,
            rpel => 2,
            family => family,
            safety => false,
            fflush => false)
        port map(
            rst => rst_i,
            wk => lclk,
            wd => fifo2_ld_iq,
            wadv => fifo2_write_q,
            we => open,
            wpf => fifo2_nf,
            wpf2 => open,
            wf => fifo2_f,
            wovf => open,
            wlevel => fifo2_level,
            rk => fifo_clk,
            rq => rfifo2_outputs.data,
            radv => rfifo2_inputs.rd,
            rflush => '0',
            re => rfifo2_outputs.empty,
            rpe => rfifo2_outputs.nempty,
            rf => open,
            rudf => open,
            rlevel => rfifo2_outputs.level);

    
    sfifop3 : sfifop
	generic map(
            width      => 64,
            order      => 9,
            rpel       => 2,
            wpfl       => 3,
            family     => family,
            safety     => false)
        port map(
            rst        => rst_i,
            k          => lclk,
            wd         => fifo3_data,
            wadv       => fifo3_write,
            we         => open,
            wpf        => fifo3_nf,
            wf         => fifo3_f,
            wovf       => open,
            rq         => fifo3_q,
            radv       => fifo3_read,
            radvt      => fifo3_radvt,
            rrev       => fifo3_rev,
            rflush     => logic0,
            re         => fifo3_e,
            rf         => open,
            rpe        => fifo3_ne,
            rudf       => open,
            rlevel     => fifo3_level);


    
    
    
    afifo3 : afifo
        generic map(
            width => 64,
            order => 9,
            wpfl => 3,
            wpfl2 => 0,
            rpel => 2,
            family => family,
            safety => false,
            fflush => false)
        port map(
            rst => rst_i,
            wk => fifo_clk,
            wd => wfifo3_inputs.data,
            wadv => wfifo3_inputs.wr,
            we => open,
            wpf => wfifo3_outputs.nfull,
            wpf2 => open,
            wf => wfifo3_outputs.full,
            wovf => open,
            wlevel => wfifo3_outputs.level,
            rk => lclk,
            rq => afifo3_qdata,
            radv => afifo3_read,
            rflush => '0',
            re => afifo3_empty,
            rpe => open,
            rf => open,
            rudf => open,
            rlevel => open);


    --
    -- Local Bus to slave bus conversion.
    -- 
    lb2slave0: process (lclk, rst_i)
    begin  -- process lb2slave0
      if rst_i = '1' then               -- asynchronous reset
        ds_ready <= '1';
        ready_timeout_count <= (others => '0');
        ld_o_slavebus <= (others => '0');
        slavebus_rd <= '0';
        slavebus_wr <= '0';
      elsif lclk'event and lclk = '1' then  -- rising clock edge
        -- Set ds_ready low on direct slave read
        if ds_qlads = '1' and ldack_i = "0000"  and lwrite_i = '0' then
          ds_ready <= '0';
        elsif slavebus_inputs.qv = '1' or ready_timeout = '1' then
          ds_ready <= '1';
        end if;

        -- Timeout of 256 cycles if slave does not respond
        if ds_ready = '0' then
          ready_timeout_count <= ready_timeout_count +1;
        else
          ready_timeout_count <= (others => '0');
        end if;
        ready_timeout <= AND_reduce(ready_timeout_count);

        -- Output single word
        if slavebus_inputs.qv = '1' then
          ld_o_slavebus <= slavebus_inputs.q;
        elsif ready_timeout = '1' then
          ld_o_slavebus <= X"55AA55AA55AADEAD";
        end if;

        -- Generate read request
        if ds_qlads = '1' and ldack_i = "0000"  and lwrite_i = '0' then
          slavebus_rd <= '1';
        else
          slavebus_rd <= '0';
        end if;    

        -- Generate write request
        if ds_xfer = '1' and ds_write = '1' and OR_reduce(ldack_q) = '0' then
          slavebus_wr <= '1';
        else
          slavebus_wr <= '0';
        end if;        
        slavebus_outputs.rd <= slavebus_rd;
        slavebus_outputs.wr <= slavebus_wr;      
        if slavebus_wr = '1' or slavebus_rd = '1' then
          slavebus_outputs.a <= la_q(21 downto 2) & "00";
          slavebus_outputs.d <= ld_iq;
          slavebus_outputs.be <= not lbe_iq_l; 
        end if; 
      end if;
    end process lb2slave0;


    

    
end mixed;
