--
-- ocp_ddma_engine_b64.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
-- OCP DDMA Engine
--
-- Demand mode DMA engine for OCP/Local Bus transfers

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.ocpad.all;
use work.fifo_if_types.all;
use work.ocp_ddma_components.all;


entity ocp_ddma_engine_b64 is
  generic (
    number_of_ocp_masters : integer := 2);
  port (
    clk          : in  std_logic;
    rst          : in  std_logic;
    -- OCP Interfaces
    ocp_bus_outputs_array         : out ocpad_burst64_m2sT_array(0 to number_of_ocp_masters-1);
    ocp_bus_inputs_array          : in  ocpad_burst64_s2mT_array(0 to number_of_ocp_masters-1);
    -- FIFO Interface
    rfifo_inputs    : out  fifo_read_port_inputs_type;
    rfifo_outputs   : in   fifo_read_port_outputs_type;
    wfifo_inputs    : out  fifo_write_port_inputs_type;
    wfifo_outputs   : in   fifo_write_port_outputs_type;
    timeout_disable    : in std_logic;
    -- write side parameters
    write_burst_size   : in  std_logic_vector(7 downto 0);
    write_xfer_size    : in  std_logic_vector(28 downto 0);
    write_row_size     : in  std_logic_vector(19 downto 0);
    write_n_rows       : in  std_logic_vector(19 downto 0);
    write_col_inc      : in  std_logic_vector(19 downto 0);
    write_start_addr   : in  std_logic_vector(31 downto 0);
    write_start        : in  std_Logic;
    write_status       : out std_logic_vector(63 downto 0);
    write_target       : in std_logic_vector(2 downto 0);
    -- read side parameters
    read_burst_size   : in  std_logic_vector(7 downto 0);
    read_xfer_size    : in  std_logic_vector(28 downto 0);
    read_row_size     : in  std_logic_vector(19 downto 0);
    read_n_rows       : in  std_logic_vector(19 downto 0);
    read_col_inc      : in  std_logic_vector(19 downto 0);
    read_start_addr   : in  std_logic_vector(31 downto 0);
    read_start        : in  std_Logic;
    read_status       : out std_logic_vector(63 downto 0);
    read_target       : in std_logic_vector(2 downto 0);
    read_fgsg_size    : in  std_logic_vector(3 downto 0);
    read_fgsg_pattern : in  std_logic_vector(11 downto 0));

end ocp_ddma_engine_b64;

architecture rtl of ocp_ddma_engine_b64 is
  
  signal write_cmd : std_logic;
  signal read_cmd : std_logic;
  signal write_cmd_addr: std_logic_vector(31 downto 0);
  signal read_cmd_addr: std_logic_vector(31 downto 0);
  signal write_cmd_accept : std_logic;
  signal read_cmd_accept : std_logic;
  signal write_data_valid : std_logic;
  signal write_data_accept : std_logic;
  signal write_burst_length : std_logic_vector(7 downto 0);
  signal write_data_last : std_logic;
  signal read_resp_valid : std_logic;
  signal read_resp_accept : std_logic;
  signal read_burst_length : std_logic_vector(7 downto 0);
  signal read_resp_last : std_logic;
  
  signal write_cmd_a : std_logic_vector(number_of_ocp_masters-1 downto 0);
  signal write_cmd_accept_a : std_logic_vector(number_of_ocp_masters-1 downto 0);
  signal write_data_valid_a : std_logic_vector(number_of_ocp_masters-1 downto 0);
  signal write_data_accept_a : std_logic_vector(number_of_ocp_masters-1 downto 0);
  signal write_data_last_a : std_logic_vector(number_of_ocp_masters-1 downto 0);

  signal read_cmd_a : std_logic_vector(number_of_ocp_masters-1 downto 0);
  signal read_cmd_accept_a : std_logic_vector(number_of_ocp_masters-1 downto 0);
  signal read_data_valid_a : std_logic_vector(number_of_ocp_masters-1 downto 0);
  signal read_data_accept_a : std_logic_vector(number_of_ocp_masters-1 downto 0);
  signal read_data_last_a : std_logic_vector(number_of_ocp_masters-1 downto 0);

  type read_data_array is array (0 to number_of_ocp_masters-1) of std_logic_vector(63 downto 0);
  signal read_data : read_data_array;

  signal fifo_wadv : std_logic;
  
  signal read_data_reg : std_logic_vector(63 downto 0);
  signal fifo_wadv_reg : std_logic;
  signal read_data_reg1 : std_logic_vector(63 downto 0);
  signal fifo_wadv_reg1 : std_logic;
   
  signal byte_level_fgsg : std_logic;
  signal byte_level_fgsg_pattern  : std_logic_vector(2 downto 0);
  
begin  -- rtl
  assert number_of_ocp_masters<9 report "Maximum of 8 OCP buses" severity error;
  assert number_of_ocp_masters>0 report "Must have 1 or more OCP buses" severity error;
  
  wsfms0: ddma_wfsm_b64
    port map (
      clk          => clk,
      burst_size   => write_burst_size,
      xfer_size    => write_xfer_size,
      row_size     => write_row_size,
      n_rows       => write_n_rows,
      col_inc      => write_col_inc,
      start_addr   => write_start_addr,
      start        => write_start,
      write_cmd    => write_cmd,
      cmd_addr     => write_cmd_addr,
      cmd_accept   => write_cmd_accept,
      data_valid   => write_data_valid,
      data_accept  => write_data_accept,
      burst_length => write_burst_length,
      data_last    => write_data_last,
      fifo_level   => rfifo_outputs.level,
      fifo_radv    => rfifo_inputs.rd,
      xfer_status  => write_status);

  rsfsm0: ddma_rfsm_b64
    port map (
      clk          => clk,
      burst_size   => read_burst_size,
      xfer_size    => read_xfer_size,
      row_size     => read_row_size,
      n_rows       => read_n_rows,
      col_inc      => read_col_inc,
      start_addr   => read_start_addr,
      fgsg_size    => read_fgsg_size,
      fgsg_pattern => read_fgsg_pattern,
      start        => read_start,
      read_cmd     => read_cmd,
      cmd_addr     => read_cmd_addr,
      cmd_accept   => read_cmd_accept,
      resp_valid   => read_resp_valid,
      resp_accept  => read_resp_accept,
      burst_length => read_burst_length,
      resp_last    => read_resp_last,
      fifo_level   => wfifo_outputs.level,
      fifo_wadv    => fifo_wadv,
      xfer_status  => read_status);
    
  

  gen_rwmux: for i in 0 to number_of_ocp_masters-1 generate
    rwmux_i: ddma2ocp_rwmux_b64
      port map (
        clk                => clk,
        rst                => rst,
        timeout_disable    => timeout_disable,
        write_cmd          => write_cmd_a(i),
        write_addr         => write_cmd_addr,
        write_cmd_accept   => write_cmd_accept_a(i),
        write_data         => rfifo_outputs.data,
        write_data_valid   => write_data_valid_a(i),
        write_data_accept  => write_data_accept_a(i),
        write_burst_length => write_burst_length,
        write_data_last    => write_data_last_a(i),
        read_cmd           => read_cmd_a(i),
        read_addr          => read_cmd_addr,
        read_cmd_accept    => read_cmd_accept_a(i),
        read_data          => read_data(i),
        read_data_valid    => read_data_valid_a(i),
        read_data_accept   => read_data_accept_a(i),
        read_burst_length  => read_burst_length,
        read_data_last     => read_data_last_a(i),
        ocp_bus_outputs    => ocp_bus_outputs_array(i),
        ocp_bus_inputs     => ocp_bus_inputs_array(i));

    -- Send write commands to selected target
    write_cmd_a(i) <= write_cmd when CONV_STD_LOGIC_VECTOR(i,3) = write_target else '0';
    write_data_valid_a(i) <= write_data_valid when CONV_STD_LOGIC_VECTOR(i,3) = write_target else '0';
    write_data_last_a(i) <= write_data_last when CONV_STD_LOGIC_VECTOR(i,3) = write_target else '0';
 
    
    -- Send read commands to selected target
    read_cmd_a(i) <= read_cmd when CONV_STD_LOGIC_VECTOR(i,3) = read_target else '0';
    read_data_accept_a(i) <= read_resp_accept when CONV_STD_LOGIC_VECTOR(i,3) = read_target else '0';
 
    
  end generate gen_rwmux;
  
  -- Only single target can be active at once so can OR accept signals
  write_cmd_accept <= OR_reduce(write_cmd_accept_a);
  write_data_accept <= OR_reduce(write_data_accept_a);

  -- Only single target can be active at once so can OR accept signals
  read_cmd_accept <= OR_reduce(read_cmd_accept_a);
  read_resp_last <= OR_reduce(read_data_last_a);
  read_resp_valid <= OR_reduce(read_data_valid_a);

  -- Regsiter Output FIFO writes
  -- Only starts bursts if space in FIFO > burst_size
  fifo_regs: process (clk)
  begin  -- process fifo_regs
    if clk'event and clk = '1' then  -- rising clock edge
      for i in 0 to number_of_ocp_masters-1 loop
        if read_target = CONV_STD_LOGIC_VECTOR(i,3) then
          read_data_reg <= read_data(i);         
        end if;
      end loop;  -- i

      fifo_wadv_reg <= fifo_wadv;

      if read_start = '1' then
        if read_fgsg_size = "0001" then
          byte_level_fgsg <= '1';
          byte_level_fgsg_pattern <= read_fgsg_pattern(2 downto 0);
        else
          byte_level_fgsg <= '0';
        end if;
      end if;
  
      if byte_level_fgsg = '0' then
        read_data_reg1 <= read_data_reg;
      else
        if fifo_wadv_reg = '0' then
          case byte_level_fgsg_pattern is
            when "000" =>  -- Lower byte of 16 bits
              for i in 0 to 3 loop
                read_data_reg1(8*i+7 downto 8*i) <= read_data_reg(16*i+7 downto 16*i);
              end loop;  -- i                      
            when "001" =>  -- Upper byte of 16 bits
              for i in 0 to 3 loop
                read_data_reg1(8*i+7 downto 8*i) <= read_data_reg(16*i+15 downto 16*i+8);
              end loop;  -- i
            when "010" =>  -- Lower 16bits of 32 bits
              for i in 0 to 1 loop
                read_data_reg1(16*i+15 downto 16*i) <= read_data_reg(32*i+15 downto 32*i);
              end loop;  -- i 
            when "011" =>  -- Upper 16bits of 32 bits
              for i in 0 to 1 loop
                read_data_reg1(16*i+15 downto 16*i) <= read_data_reg(32*i+31 downto 32*i+16);
              end loop;  -- i
            when "100" =>  -- Lower 32bits of 64 bits
              read_data_reg1(31 downto 0) <= read_data_reg(31 downto 0);
            when "101" =>  -- Upper 32bits of 64 bits
              read_data_reg1(31 downto 0) <= read_data_reg(63 downto 32);
            when others =>
              null;             
          end case;
        else
          case byte_level_fgsg_pattern is
            when "000" =>  -- Lower byte of 16 bits
              for i in 0 to 3 loop
                read_data_reg1(32+8*i+7 downto 32+8*i) <= read_data_reg(16*i+7 downto 16*i);
              end loop;  -- i                      
            when "001" =>  -- Upper byte of 16 bits
              for i in 0 to 3 loop
                read_data_reg1(32+8*i+7 downto 32+8*i) <= read_data_reg(16*i+15 downto 16*i+8);
              end loop;  -- i
            when "010" =>  -- Lower 16bits of 32 bits
              for i in 0 to 1 loop
                read_data_reg1(32+16*i+15 downto 32+16*i) <= read_data_reg(32*i+15 downto 32*i);
              end loop;  -- i 
            when "011" =>  -- Upper 16bits of 32 bits
              for i in 0 to 1 loop
                read_data_reg1(32+16*i+15 downto 32+16*i) <= read_data_reg(32*i+31 downto 32*i+16);
              end loop;  -- i
            when "100" =>  -- Lower 32bits of 64 bits
              read_data_reg1(63 downto 32) <= read_data_reg(31 downto 0);
            when "101" =>  -- Upper 32bits of 64 bits
              read_data_reg1(63 downto 32) <= read_data_reg(63 downto 32);
            when others =>
              null;             
          end case;
        end if;
      end if;
      fifo_wadv_reg1<= fifo_wadv_reg;
      
      wfifo_inputs.data <= read_data_reg1;
      wfifo_inputs.wr <= fifo_wadv_reg1;      
    end if;
  end process fifo_regs;
  
  
end rtl;
