--
-- localbus_if_types.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--  
--
-- Record data types to simplify local bus IO into
-- 2 bi-directional signals

library IEEE;
use IEEE.std_logic_1164.all;
use work.fifo.all;
use work.slave_bus_if_types.all;
use work.fifo_if_types.all;
package localbus_if_types is

  type localbus_inputs_type is record
    lreset_l   : std_logic;
    lwrite     : std_logic;
    lads_l     : std_logic;
    lblast_l   : std_logic;
    lbterm_l_i : std_logic;
    lad_i      : std_logic_vector(63 downto 0);
    la_i       : std_logic_vector(31 downto 0);
    lbe_l      : std_logic_vector(7 downto 0);
    fholda     : std_logic;
    ldack_l    : std_logic_vector(3 downto 0);
  end record;

  type localbus_outputs_type is record
    lbterm_l_o : std_logic;
    lbterm_l_t : std_logic;             -- Tri-state enable for lbterm
    lad_o      : std_logic_vector(63 downto 0);
    lad_t      : std_logic_vector(63 downto 0);
    lready_l_o : std_logic;
    lready_l_t : std_logic;             -- Tri-state enable for lready
    ldreq_l    : std_logic_vector(3 downto 0);
  end record;

  component localbus_io
    generic (
      bus_width    : integer := 64;
      dma_channels : integer := 4);
    port (
      localbus_inputs  : out   localbus_inputs_type;
      localbus_outputs : in    localbus_outputs_type;
      lreset_l         : in    std_logic;
      lwrite           : in    std_logic;
      lads_l           : in    std_logic;
      lblast_l         : in    std_logic;
      lbterm_l         : inout std_logic;
      lad              : inout std_logic_vector(bus_width-1 downto 0);
      lready_l         : inout std_logic;
      lbe_l            : in    std_logic_vector(bus_width/8-1 downto 0);
      fholda           : in    std_logic;
      ldreq_l          : out   std_logic_vector(dma_channels-1 downto 0);
      ldack_l          : in    std_logic_vector(dma_channels-1 downto 0));
  end component;

  component localbus_io_plx
    generic (
      bus_width    : integer := 32;
      dma_channels : integer := 2;
      la_top       : integer := 23);
    port (
      localbus_inputs  : out localbus_inputs_type;
      localbus_outputs : in  localbus_outputs_type;
      lreset_l         : in    std_logic;
      lwrite           : in    std_logic;
      lads_l           : in    std_logic;
      lblast_l         : in    std_logic;
      lbterm_l         : inout std_logic;
      ld               : inout std_logic_vector(bus_width-1 downto 0);
      la               : in    std_Logic_vector(la_top downto 2);
      lready_l         : inout std_logic;
      lbe_l            : in    std_logic_vector(bus_width/8-1 downto 0);
      fholda           : in    std_logic;
      ldreq_l          : out   std_logic_vector(dma_channels-1 downto 0);
      ldack_l          : in    std_logic_vector(dma_channels-1 downto 0));
  end component;
  
  component localbus_ddma64_interface_b64
    generic(
      la_top : natural := 23;
      family : family_t := family_virtex4;
      dma_mapping      :     integer := 0;
      bus_width        :     integer := 64);
    port(
      -- Clocking
      lclk               : in  std_logic;  -- Local Bus Clock
      fifo_clk           : in  std_logic;  -- FIFO Clock
      rst                : out std_logic;  -- Global Reset (LRESET)
      -- Local Bus Interface
      localbus_inputs    : in  localbus_inputs_type;
      localbus_outputs   : out localbus_outputs_type;
      -- Slave Bus Interface - synchronous to LCLK
      slavebus_outputs   : out slave_bus_outputs_type;
      slavebus_inputs    : in  slave_bus_inputs_type;
      -- 4 x FIFO Interfaces connected to 4 DMA channels
      rfifo0_inputs      : in  fifo_read_port_inputs_type;
      rfifo0_outputs     : out fifo_read_port_outputs_type;
      wfifo1_inputs      : in  fifo_write_port_inputs_type;
      wfifo1_outputs     : out fifo_write_port_outputs_type;
      rfifo2_inputs      : in  fifo_read_port_inputs_type;
      rfifo2_outputs     : out fifo_read_port_outputs_type;
      wfifo3_inputs      : in  fifo_write_port_inputs_type;
      wfifo3_outputs     : out fifo_write_port_outputs_type;
      fifo_counter_set   : in  std_logic_vector(3 downto 0);
      fifo_counter0_init : in  std_logic_vector(31 downto 0);
      fifo_counter1_init : in  std_logic_vector(31 downto 0);
      fifo_counter2_init : in  std_logic_vector(31 downto 0);
      fifo_counter3_init : in  std_logic_vector(31 downto 0);
      fifo_counter0      : out std_logic_vector(31 downto 0);
      fifo_counter1      : out std_logic_vector(31 downto 0);
      fifo_counter2      : out std_logic_vector(31 downto 0);
      fifo_counter3      : out std_logic_vector(31 downto 0));    
  end component;


  component localbus_ddma64_interface_b128
    generic(
      la_top : natural := 23;
      family : family_t := family_virtex4;
      dma_mapping      :     integer := 0;
      bus_width        :     integer := 64);
    port(
      -- Clocking
      lclk               : in  std_logic;  -- Local Bus Clock
      fifo_clk           : in  std_logic;  -- FIFO Clock
      rst                : out std_logic;  -- Global Reset (LRESET)
      -- Local Bus Interface
      localbus_inputs    : in  localbus_inputs_type;
      localbus_outputs   : out localbus_outputs_type;
      -- Slave Bus Interface - synchronous to LCLK
      slavebus_outputs   : out slave_bus_outputs_type;
      slavebus_inputs    : in  slave_bus_inputs_type;
      -- 4 x FIFO Interfaces connected to 4 DMA channels
      rfifo0_inputs      : in  fifo_read_port_inputs_type;
      rfifo0_outputs     : out fifo_read_port_outputs_type;
      wfifo1_inputs      : in  fifo_write_port_inputs_type;
      wfifo1_outputs     : out fifo_write_port_outputs_type;
      rfifo2_inputs      : in  fifo_read_port_inputs_type;
      rfifo2_outputs     : out fifo_read_port_outputs_type;
      wfifo3_inputs      : in  fifo_write_port_inputs_type;
      wfifo3_outputs     : out fifo_write_port_outputs_type;
      fifo_counter_set   : in  std_logic_vector(3 downto 0);
      fifo_counter0_init : in  std_logic_vector(31 downto 0);
      fifo_counter1_init : in  std_logic_vector(31 downto 0);
      fifo_counter2_init : in  std_logic_vector(31 downto 0);
      fifo_counter3_init : in  std_logic_vector(31 downto 0);
      fifo_counter0      : out std_logic_vector(31 downto 0);
      fifo_counter1      : out std_logic_vector(31 downto 0);
      fifo_counter2      : out std_logic_vector(31 downto 0);
      fifo_counter3      : out std_logic_vector(31 downto 0));    
  end component;

 

  component localbus_ddma32_interface_b32
    generic(
      la_top : natural := 23;
      family : family_t := family_virtex4;
      dma_mapping      :     integer := 0;
      bus_width        :     integer := 32);
    port(
      -- Clocking
      lclk               : in  std_logic;  -- Local Bus Clock
      fifo_clk           : in  std_logic;  -- FIFO Clock
      rst                : out std_logic;  -- Global Reset (LRESET)
      -- Local Bus Interface
      localbus_inputs    : in  localbus_inputs_type;
      localbus_outputs   : out localbus_outputs_type;
      -- Slave Bus Interface - synchronous to LCLK
      slavebus_outputs   : out slave_bus_outputs_type;
      slavebus_inputs    : in  slave_bus_inputs_type;
      -- 4 x FIFO Interfaces connected to 4 DMA channels
      rfifo0_inputs      : in  fifo_read_port_inputs_type;
      rfifo0_outputs     : out fifo_read_port_outputs_type;
      wfifo1_inputs      : in  fifo_write_port_inputs_type;
      wfifo1_outputs     : out fifo_write_port_outputs_type;
      rfifo2_inputs      : in  fifo_read_port_inputs_type;
      rfifo2_outputs     : out fifo_read_port_outputs_type;
      wfifo3_inputs      : in  fifo_write_port_inputs_type;
      wfifo3_outputs     : out fifo_write_port_outputs_type;
      fifo_counter_set   : in  std_logic_vector(3 downto 0);
      fifo_counter0_init : in  std_logic_vector(31 downto 0);
      fifo_counter1_init : in  std_logic_vector(31 downto 0);
      fifo_counter2_init : in  std_logic_vector(31 downto 0);
      fifo_counter3_init : in  std_logic_vector(31 downto 0);
      fifo_counter0      : out std_logic_vector(31 downto 0);
      fifo_counter1      : out std_logic_vector(31 downto 0);
      fifo_counter2      : out std_logic_vector(31 downto 0);
      fifo_counter3      : out std_logic_vector(31 downto 0));    
  end component;

  
end package localbus_if_types;
