--
-- slavebus2ocp32.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
--
-- Simple slavebus to OCP bridge
--
-- Converts simple slave bus to non-bursting OCP bus

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.slave_bus_if_types.all;
use work.ocpad.all;

entity slavebus2ocp32 is

  port (
    lclk                     : in  std_logic;
    ocp_clk                     : in  std_logic;
    rst                     : in  std_logic;
    slave_bus_outputs       : in  slave_bus_outputs_type;
    slave_bus_inputs        : out slave_bus_inputs_type;
    ocp_bus_outputs         : out ocpad_register32_m2sT := ocpad_register32_default_m2s;
    ocp_bus_inputs          : in  ocpad_register32_s2mT;
    page_register           : in  std_logic_vector(31 downto 0));

end slavebus2ocp32;

architecture rtl of slavebus2ocp32 is

  signal rd,wr,qv : std_logic;

  component async_pulse
    port (
      clk : in std_logic;
      I   : in std_Logic;
      O   : out std_Logic);
  end component;

  signal rd64,wr64 : std_logic := '0';
  signal addr64 : std_logic_vector(31 downto 0);
  signal rd32_busy : std_logic := '0';
  signal rd64_busy : std_logic := '0';
  signal wr32_busy : std_logic := '0';

  signal d64 : std_logic_vector(31 downto 0);
  signal be64 : std_logic_vector(3 downto 0);


  signal reset_n : std_logic;
  
begin  -- rtl

  ap0: async_pulse 
    port map (
      clk => ocp_clk,
      I   => slave_bus_outputs.rd,
      O   => rd);

  ap1: async_pulse 
    port map (
      clk => ocp_clk,
      I   => slave_bus_outputs.wr,
      O   => wr);

  ap2: async_pulse 
    port map (
      clk => lclk,
      I   => qv,
      O   => slave_bus_inputs.qv);

  
  process (ocp_clk)
  begin  -- process
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      reset_n <= not rst;
      ocp_bus_outputs.reset_n <= reset_n;
      if rd = '1' then
        if OR_reduce(slave_bus_outputs.be(3 downto 0)) = '1' then
          ocp_bus_outputs.cmd <= "010";
          rd32_busy <= '1';
        end if;
      elsif wr = '1' then
        if OR_reduce(slave_bus_outputs.be(3 downto 0)) = '1' then
          ocp_bus_outputs.cmd <= "001";
          wr32_busy <= '1';
        end if;
      elsif rd32_busy = '0' and rd64 = '1' then
        ocp_bus_outputs.cmd <= "010";
        rd64_busy <= '1';
      elsif wr32_busy = '0' and wr64 = '1' then
        ocp_bus_outputs.cmd <= "001";
      elsif ocp_bus_inputs.cmdaccept = '1' then
        ocp_bus_outputs.cmd <= "000";
        wr32_busy <= '0';
        if ocp_bus_inputs.resp = "01" then
          rd32_busy <= '0';
          rd64_busy <= '0';
        end if;
      elsif ocp_bus_inputs.resp = "01" then
        rd32_busy <= '0';
        rd64_busy <= '0';
      end if;

      if rd = '1' then
        rd64 <= OR_reduce(slave_bus_outputs.be(7 downto 4));
      elsif wr = '1' then
        wr64 <= OR_reduce(slave_bus_outputs.be(7 downto 4));
      elsif rd32_busy = '0' and rd64 = '1' then
        rd64 <= '0';
      elsif wr32_busy = '0' and wr64 = '1' then
        wr64 <= '0';
      end if;
      
      if wr = '1' or rd = '1' then
        ocp_bus_outputs.addr <= page_register(31 downto 19) & slave_bus_outputs.a(18 downto 0);
        addr64 <= page_register(31 downto 19) & slave_bus_outputs.a(18 downto 0);
      elsif rd64 = '1' or wr64 = '1' then
        ocp_bus_outputs.addr <= addr64+4;
      end if;
      
      if wr = '1' then
        ocp_bus_outputs.data <= slave_bus_outputs.d(31 downto 0);
        ocp_bus_outputs.databyteen <= slave_bus_outputs.be(3 downto 0);
        d64 <=  slave_bus_outputs.d(63 downto 32);
        be64 <=  slave_bus_outputs.be(7 downto 4);       
      elsif wr64 = '1' then
        ocp_bus_outputs.data <= d64;
        ocp_bus_outputs.databyteen <= be64;       
      end if;
      
      if ocp_bus_inputs.resp = "01" then
        -- Receive just first data word 
        qv <= not rd64;
        if rd64_busy = '0' then
          slave_bus_inputs.q(31 downto 0) <= ocp_bus_inputs.data;
        else
          slave_bus_inputs.q(63 downto 32) <= ocp_bus_inputs.data;
        end if;
      else
        qv <= '0';
      end if;       
    end if;
  end process;
 
  
  
  
end rtl;
