--
-- slavebus_mux.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
--
-- Multiplexer for slavebus
--
-- Provides a 1 to N (N=2..8)
-- fanout for slavebus outputs

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.slave_bus_if_types.all;

entity slavebus_mux is
  
  generic (
    number_of_output_buses : integer := 2); 

  port (
    clk                     : in  std_logic;
    slave_bus_outputs       : in  slave_bus_outputs_type;
    slave_bus_inputs        : out slave_bus_inputs_type;
    slave_bus_multiple_outputs : out slave_bus_outputs_array(0 to number_of_output_buses-1);
    slave_bus_multiple_inputs : in slave_bus_inputs_array(0 to number_of_output_buses-1));

end slavebus_mux;

architecture rtl of slavebus_mux is

  signal current_bus : std_logic_vector(2 downto 0) := "000";
  signal current_bus_onehot : std_logic_vector(number_of_output_buses-1 downto 0) := (0 => '1', others => '0');
  
begin  -- rtl

  assert number_of_output_buses<9 report "Maximum of 8 slave buses" severity error;
  assert number_of_output_buses>1 report "Minimum of 2 slave buses" severity error;

  gen_onehot: for i in 0 to number_of_output_buses-1 generate
    current_bus_onehot(i) <= '1' when CONV_STD_LOGIC_VECTOR(i,3) =  slave_bus_outputs.a(21 downto 19) else '0';
  end generate gen_onehot;
  
  process (clk)
  begin  -- process
    if clk'event and clk = '1' then  -- rising clock edge
      for i in 0 to number_of_output_buses-1 loop
        slave_bus_multiple_outputs(i).rd <= slave_bus_outputs.rd and current_bus_onehot(i);
        slave_bus_multiple_outputs(i).wr <= slave_bus_outputs.wr and current_bus_onehot(i);
        slave_bus_multiple_outputs(i).d <= slave_bus_outputs.d;
        slave_bus_multiple_outputs(i).a <= EXT(slave_bus_outputs.a(18 downto 0),22);
        slave_bus_multiple_outputs(i).be <= slave_bus_outputs.be;  
      end loop;  -- i
      if slave_bus_outputs.rd = '1' or slave_bus_outputs.wr = '1' then
        current_bus <= slave_bus_outputs.a(21 downto 19);
      end if;
      for i in 0 to number_of_output_buses-1 loop
        if CONV_STD_LOGIC_VECTOR(i,3) = current_bus then
          slave_bus_inputs.q <= slave_bus_multiple_inputs(i).q;
          slave_bus_inputs.qv <= slave_bus_multiple_inputs(i).qv;
        end if;
      end loop;  -- i
    end if;
  end process;
  
end rtl;
