--
-- slavebus_register_bank.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
--
-- Bank of readable registers on the slavebus

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.slave_bus_if_types.all;

entity slavebus_register_bank is
  
  generic (
    number_of_control_registers : integer := 4;
    control_base_address        : integer := 0;  -- in bytes
    number_of_status_registers  : integer := 4;
    status_base_address         : integer := 32);  -- in bytes

  port (
    clk                     : in  std_logic;
    slave_bus_in            : in  slave_bus_outputs_type;
    slave_bus_out           : out slave_bus_inputs_type;
    control_register_values : out register_array(0 to number_of_control_registers-1);
    control_register_writes : out std_logic_vector(number_of_control_registers-1 downto 0);
    status_register_values  : in  register_array(0 to number_of_status_registers-1);
    status_register_reads_done : out std_logic_vector(number_of_status_registers-1 downto 0));

end slavebus_register_bank;

architecture rtl of slavebus_register_bank is

  signal control_registers : register_array(0 to number_of_control_registers-1) := (others => (others => '0'));

  signal control_reg_decode : std_logic_vector(number_of_control_registers-1 downto 0) := (others => '0');
  signal status_reg_decode : std_logic_vector(number_of_status_registers-1 downto 0) := (others => '0') ;
  signal control_register_wr : std_logic_vector(number_of_control_registers-1 downto 0) := (others => '0');

  signal q_sig,q_out : std_logic_vector(63 downto 0) := (others => '0');
  signal qv,rd_reg : std_logic := '0';
  
  
begin  -- rtl

  control_regs: process (clk)
  begin  -- process control_regs
    if clk'event and clk = '1' then  -- rising clock edge
      for i in 0 to number_of_control_registers-1 loop
        if (i mod 2) = 0 then
          if slave_bus_in.a(2) = '0' and slave_bus_in.a = CONV_STD_LOGIC_VECTOR(control_base_address+i*4,22) then
            if slave_bus_in.wr = '1' and slave_bus_in.be(0) = '1' then
              control_registers(i)(7 downto 0) <= slave_bus_in.d(7 downto 0);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(1) = '1' then
              control_registers(i)(15 downto 8) <= slave_bus_in.d(15 downto 8);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(2) = '1' then
              control_registers(i)(23 downto 16) <= slave_bus_in.d(23 downto 16);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(3) = '1' then
              control_registers(i)(31 downto 24) <= slave_bus_in.d(31 downto 24);
            end if;
            control_register_wr(i) <= slave_bus_in.wr and OR_reduce(slave_bus_in.be(3 downto 0));
            control_reg_decode(i) <= slave_bus_in.rd;
          end if;
        else
          -- (odd addresses)
          if slave_bus_in.a(2) = '0' and slave_bus_in.a = CONV_STD_LOGIC_VECTOR(control_base_address+(i-1)*4,22) then
            -- 64 bit local bus mode
            if slave_bus_in.wr = '1' and slave_bus_in.be(4) = '1' then
              control_registers(i)(7 downto 0) <= slave_bus_in.d(39 downto 32);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(5) = '1' then
              control_registers(i)(15 downto 8) <= slave_bus_in.d(47 downto 40);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(6) = '1' then
              control_registers(i)(23 downto 16) <= slave_bus_in.d(55 downto 48);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(7) = '1' then
              control_registers(i)(31 downto 24) <= slave_bus_in.d(63 downto 56);
            end if;
            control_register_wr(i) <= slave_bus_in.wr and OR_reduce(slave_bus_in.be(7 downto 4));
            control_reg_decode(i) <= slave_bus_in.rd;
          elsif  slave_bus_in.a(2) = '1' and slave_bus_in.a = CONV_STD_LOGIC_VECTOR(control_base_address+i*4,22) then
            -- 32 bit local bus mode
            if slave_bus_in.wr = '1' and slave_bus_in.be(0) = '1' then
              control_registers(i)(7 downto 0) <= slave_bus_in.d(7 downto 0);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(1) = '1' then
              control_registers(i)(15 downto 8) <= slave_bus_in.d(15 downto 8);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(2) = '1' then
              control_registers(i)(23 downto 16) <= slave_bus_in.d(23 downto 16);
            end if;
            if slave_bus_in.wr = '1' and slave_bus_in.be(3) = '1' then
              control_registers(i)(31 downto 24) <= slave_bus_in.d(31 downto 24);
            end if;
            control_register_wr(i) <= slave_bus_in.wr and OR_reduce(slave_bus_in.be(3 downto 0));
            control_reg_decode(i) <= slave_bus_in.rd;
          end if;
        end if;
      end loop;  -- i
    end if;
  end process control_regs;

  control_register_values <= control_registers;
  control_register_writes <= control_register_wr;
  
  status_regs: process (clk)
  begin  -- process status_regs
    if clk'event and clk = '1' then  -- rising clock edge
      for i in 0 to number_of_status_registers-1 loop
        if (i mod 2) = 0 then
          if slave_bus_in.a(2) = '0' and slave_bus_in.a = CONV_STD_LOGIC_VECTOR(status_base_address+i*4,22) then
            status_reg_decode(i) <= slave_bus_in.rd;
          end if;
        else
          -- (odd addresses)
          if slave_bus_in.a(2) = '0' and slave_bus_in.a = CONV_STD_LOGIC_VECTOR(status_base_address+(i-1)*4,22) then
            -- 64 bit local bus mode
            status_reg_decode(i) <= slave_bus_in.rd;
          elsif  slave_bus_in.a(2) = '1' and slave_bus_in.a = CONV_STD_LOGIC_VECTOR(status_base_address+i*4,22) then
            -- 32 bit local bus mode
            status_reg_decode(i) <= slave_bus_in.rd;
          end if;
        end if;
      end loop;  -- i
      status_register_reads_done <= status_reg_decode;
    end if;
  end process status_regs;


  gen_control_read_q: for i in 0 to number_of_control_registers-1 generate
    gen_low_word: if (i mod 2) = 0 generate
      q_sig(31 downto 0) <= control_registers(i) when control_reg_decode(i) = '1' else (others => 'Z');
    end generate gen_low_word;
    gen_high_word: if (i mod 2) = 1 generate
      q_sig(31 downto 0) <= control_registers(i) when control_reg_decode(i) = '1' and control_reg_decode(i-1) = '0' else (others => 'Z');
      q_sig(63 downto 32) <= control_registers(i) when control_reg_decode(i) = '1' and control_reg_decode(i-1) = '1' else (others => 'Z');
    end generate gen_high_word;
    
  end generate gen_control_read_q;

  gen_status_read_q: for i in 0 to number_of_status_registers-1 generate
    gen_low_word2: if (i mod 2) = 0 generate
      q_sig(31 downto 0) <= status_register_values(i) when status_reg_decode(i) = '1' else (others => 'Z');
    end generate gen_low_word2;
    gen_high_word2: if (i mod 2) = 1 generate
      q_sig(31 downto 0) <= status_register_values(i) when status_reg_decode(i) = '1' and status_reg_decode(i-1) = '0' else (others => 'Z');
      q_sig(63 downto 32) <= status_register_values(i) when status_reg_decode(i) = '1' and status_reg_decode(i-1) = '1' else (others => 'Z');
    end generate gen_high_word2;
    
  end generate gen_status_read_q;


  register_output: process (clk)
  begin  -- process register_output
    if clk'event and clk = '1' then  -- rising clock edge
      rd_reg <=  slave_bus_in.rd;
      qv <= rd_reg;
      q_out <= q_sig;
    end if;
  end process register_output;

  slave_bus_out.qv <= qv;
  slave_bus_out.q <= q_out;
  
  
end rtl;
