--
-- localbus2ocp_bridge_r32.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
-- LocalBus to OCP Bridge
--
-- Connects the Local Bus to multiple OCP ports
--
-- Allows up to 7 memory mapped simple non-bursting ports
-- for configuration registers and low data transfer rates. (32 bit)
--
-- No Bursting Ports Supported


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.fifo.all;
use work.ocpad.all;
use work.fifo_if_types.all;
use work.localbus_if_types.all;
use work.slave_bus_if_types.all;

entity localbus2ocp_bridge_r32 is
  generic (
    number_of_ocp_nonburst : integer := 7;
    la_top           : natural := 23;
    family           : integer := fifo_family_virtex4);
  port (
    lclk             : in    std_logic;  -- Local Bus Clock
    ocp_clk_nb       : in    std_logic;  -- System Clock for non-burst
    rst              : out   std_logic;  -- Global Reset (LRESET)
    -- OCP Interfaces
    ocp_nonburst_bus_outputs_array         : out ocpad_register32_m2sT_array(0 to number_of_ocp_nonburst-1);
    ocp_nonburst_bus_inputs_array          : in  ocpad_register32_s2mT_array(0 to number_of_ocp_nonburst-1);
    -- Local Bus Interface
    localbus_inputs  : in  localbus_inputs_type;
    localbus_outputs : out localbus_outputs_type;
    system_status    : in  std_logic_vector(31 downto 0);
    system_irq       : in  std_logic_vector(31 downto 0);
    irq              : out std_logic);

end localbus2ocp_bridge_r32;

architecture rtl of localbus2ocp_bridge_r32 is


  constant ocp_nonburst_reg_base : integer := 4;
  constant irq_enb_addr : integer := 32;
  constant irq_clr_addr : integer := 33;
  constant irq_status_addr : integer := 1;
  
  constant number_of_control_registers : integer := 4+8+20+2;
  constant number_of_status_registers : integer := 4+8+2;
  
  signal slave_bus_multiple_outputs : slave_bus_outputs_array(0 to number_of_ocp_nonburst);
  signal slave_bus_multiple_inputs : slave_bus_inputs_array(0 to number_of_ocp_nonburst);
  signal slave_bus_inputs : slave_bus_inputs_type;
  signal slave_bus_outputs : slave_bus_outputs_type;

  signal slave_bus_rom_out,slave_bus_reg_out : slave_bus_inputs_type;
  signal sbrom_addressed : std_logic;
  
  
  signal control_regs : register_array(0 to number_of_control_registers-1);
  signal status_regs : register_array(0 to number_of_status_registers-1);
  signal control_write : std_logic_vector(number_of_control_registers-1 downto 0);

  signal w0_status,w1_status,r0_status,r1_status : std_logic_vector(63 downto 0);

  signal rst_i : std_logic;

  component async_pulse
    port (
      clk : in std_logic;
      I   : in std_Logic;
      O   : out std_Logic);
  end component;


  signal write_start0,write_start1,read_start0,read_start1 : std_logic;

  signal irq_status : std_logic_vector(31 downto 0);
  signal irq_rising_edge : std_logic_vector(31 downto 0);
  signal irq_enable : std_Logic_vector(31 downto 0);
  signal irq_clear : std_Logic_vector(31 downto 0);
  
  component localbus_interface
    generic(
      la_top           :     natural := 23;
      family           :     integer := fifo_family_virtex4;
      bus_width        :     integer := 64);
    port(
      -- Clocking
      lclk             : in    std_logic;  -- Local Bus Clock
      rst              : out   std_logic;  -- Global Reset (LRESET)
      -- Local Bus Interface
      localbus_inputs  : in  localbus_inputs_type;
      localbus_outputs : out localbus_outputs_type;
      -- Slave Bus Interface - synchronous to LCLK
      slavebus_outputs : out slave_bus_outputs_type;
      slavebus_inputs  : in  slave_bus_inputs_type); 
  end component;
  
  
begin  -- rtl
  assert number_of_ocp_nonburst<8 report "Maximum of 7 OCP non-burst buses" severity error;
  
  
  rst <= rst_i;
  
  lbif0: localbus_interface
    generic map (
      la_top => la_top,
      family => family)
    port map (
      lclk             => lclk,
      rst              => rst_i,
      localbus_inputs  => localbus_inputs,
      localbus_outputs => localbus_outputs,
      slavebus_inputs  => slave_bus_inputs,
      slavebus_outputs => slave_bus_outputs);


  sbm0: slavebus_mux
    generic map (
      number_of_output_buses => number_of_ocp_nonburst+1)
    port map (
      clk                        => lclk,
      slave_bus_outputs          => slave_bus_outputs,
      slave_bus_inputs           => slave_bus_inputs,
      slave_bus_multiple_outputs => slave_bus_multiple_outputs,
      slave_bus_multiple_inputs  => slave_bus_multiple_inputs);
  
  
  sbrb0: slavebus_register_bank
    generic map (
      number_of_control_registers => number_of_control_registers,
      control_base_address        => 0,
      number_of_status_registers  => number_of_status_registers,
      status_base_address         => number_of_control_registers*4)
    port map (
      clk                     => lclk,
      slave_bus_in            => slave_bus_multiple_outputs(0),
      slave_bus_out           => slave_bus_reg_out,
      control_register_values => control_regs,
      control_register_writes => control_write,
      status_register_values  => status_regs);

  sbrom0: slavebus_rom
    port map (
      clk           => lclk,
      slave_bus_in  => slave_bus_multiple_outputs(0),
      slave_bus_out => slave_bus_rom_out,
      rom_addressed => sbrom_addressed);
  
  slave_bus_multiple_inputs(0) <= slave_bus_rom_out when sbrom_addressed = '1' else slave_bus_reg_out;
  

  
  gen_slave_ocp_if: for i in 0 to number_of_ocp_nonburst-1 generate
    sb2ocp_i: slavebus2ocp32
      port map (
        lclk              => lclk,
        ocp_clk           => ocp_clk_nb,
        rst               => rst_i,
        slave_bus_outputs => slave_bus_multiple_outputs(i+1),
        slave_bus_inputs  => slave_bus_multiple_inputs(i+1),
        ocp_bus_outputs   => ocp_nonburst_bus_outputs_array(i),
        ocp_bus_inputs    => ocp_nonburst_bus_inputs_array(i),
        page_register     => control_regs(i+1+ocp_nonburst_reg_base));
  end generate gen_slave_ocp_if;
  

  
  status_regs(0) <= system_status;
  status_regs(irq_status_addr) <= irq_status;

  irq_clear <= control_regs(irq_clr_addr);
  irq_enable <= control_regs(irq_enb_addr);

  
  edge_detect_irq: for i in 0 to 31 generate
    ap_i: async_pulse
      port map (
        clk => lclk,
        I   => system_irq(i),
        O   => irq_rising_edge(i));
  end generate edge_detect_irq;

  latch_irq: process (lclk, rst_i)
  begin  -- process latch_irq
    if rst_i = '1' then                   -- asynchronous reset
      irq_status <= (others => '0');
      irq <= '0';
    elsif lclk'event and lclk = '1' then  -- rising clock edge
      for i in 0 to 31 loop
        if control_write(irq_clr_addr) = '1' and irq_clear(i) = '1' then
          irq_status(i) <= '0';
        elsif irq_rising_edge(i) = '1' and irq_enable(i) = '1' then
          irq_status(i) <= '1';
        end if;
      end loop;  -- i

      -- Bring IRQ line low when clearing status bits
      -- Will generate a new interrupt if other source is active
      if control_write(irq_clr_addr) = '1' then
        irq <= '0';
      else
        irq <= OR_reduce(irq_status);
      end if;           
    end if;
  end process latch_irq;

  

  
end rtl;
