--
-- localbus_interface.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2007-2009 Alpha Data Ltd.
--
-- localbus_interface
--
-- Provides an interface between the Local bus and
-- a simple non-bursting slave bus
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.localbus.all;
use work.localbus_if_types.all;
use work.slave_bus_if_types.all;
use work.fifo_if_types.all;

entity localbus_interface is
  generic(
    la_top           :     natural := 23;
    bus_width        :     integer := 64);
  port(
    -- Clocking
    lclk             : in    std_logic;  -- Local Bus Clock
    rst              : out   std_logic;  -- Global Reset (LRESET)
    -- Local Bus Interface
    localbus_inputs  : in  localbus_inputs_type;
    localbus_outputs : out localbus_outputs_type;
    -- Slave Bus Interface - synchronous to LCLK
    slavebus_outputs : out slave_bus_outputs_type;
    slavebus_inputs  : in  slave_bus_inputs_type);      
end entity;

architecture mixed of localbus_interface is
    --
    -- This value is the number of words / free locations that the design attempts to aggregate
    -- in the FIFO before attempting a demand-mode DMA burst. This is intended to increase the
    -- efficiency of the local bus by increasing the average burst length for demand-mode DMA
    -- transfers.
    --

    constant threshold : natural := 32;

    signal rst_i : std_logic;
    signal lads_i : std_logic;
    signal lblast_i : std_logic;
    signal lbe_iq_l : std_logic_vector(7 downto 0);
    signal lwrite_i : std_logic;
    signal ldack_i : std_logic_vector(3 downto 0);
    signal ldack_q : std_logic_vector(3 downto 0);
    signal ld_iq : std_logic_vector(63 downto 0);
    signal ld_o : std_logic_vector(63 downto 0);
    signal la_i : std_logic_vector(la_top downto 2);
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal lbterm_i : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal ldreq_o_l : std_logic_vector(3 downto 0);
    signal ld_oe_l : std_logic;

    signal fholda : std_logic;
    
    signal la_q : std_logic_vector(la_top downto 2);
    
    signal ds_qlads : std_logic;
    signal ds_idle : std_logic;
    signal ds_decode : std_logic;
    signal ds_write : std_logic;
    signal ds_xfer : std_logic;
 
    signal ds_ready : std_logic;
    signal ds_stop : std_logic;

  

    signal logic0 : std_logic;
    signal logic1 : std_logic;

    signal ready_timeout : std_logic;
    signal ready_timeout_count : std_logic_vector(7 downto 0);

    signal ld_o_slavebus : std_logic_vector(63 downto 0);
    signal slavebus_rd  : std_logic;
    signal slavebus_wr : std_logic;

   
    attribute iob : string;
    attribute iob of lbe_iq_l : signal is "true";
    attribute iob of ld_iq : signal is "true";
    
begin

  assert bus_width = 64 or bus_width = 32 report "Bus Width must be 32 or 64 bits" severity error;
  
 
    logic0 <= '0';
    logic1 <= '1';

    rst <= rst_i;
    
    --
    -- Convert the inputs to active high.
    --

    rst_i      <= not localbus_inputs.lreset_l;
    lads_i     <= not localbus_inputs.lads_l;
    lblast_i   <= not localbus_inputs.lblast_l;
    lbterm_i   <= not localbus_inputs.lbterm_l_i;
    lwrite_i   <= localbus_inputs.lwrite;
    la_i       <= localbus_inputs.la_i(la_top downto 2);
    ldack_i    <= not localbus_inputs.ldack_l;
    fholda     <= localbus_inputs.fholda;    
  
    --
    -- Register LD and LBE# (should be mapped into IOBs)
    --

    gen_lbe_iq_l : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            lbe_iq_l <= (others => '0');
        elsif lclk'event and lclk = '1' then
            lbe_iq_l <= localbus_inputs.lbe_l;
        end if;
    end process;

    gen_ld_iq : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            ld_iq <= (others => '0');
        elsif lclk'event and lclk = '1' then
            ld_iq <= localbus_inputs.lad_i;
        end if;
    end process;

    --
    -- Output active low signals.
    --

    localbus_outputs.ldreq_l <= ldreq_o_l;

    --
    -- Generate qualified versions of 'lads_l'.
    --
    -- For direct slave access, this is when 'lads_l' is asserted AND the FPGA
    -- is being addressed AND the FPGA is not the local bus master AND the
    -- cycle is not demand-mode DMA.
    --
    -- For demand-mode DMA access, this is when 'lads_l' is asserted AND the
    -- FPGA is being addressed and the FPGA is not the local bus master. The
    -- demand-mode DMA acknowledge signals are checked within the 'plxddsm'
    -- module.
    --

    ds_qlads <= lads_i and not la_i(la_top) and not fholda;
    

    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    localbus_outputs.lbterm_l_o <= lbterm_o_l;
    localbus_outputs.lbterm_l_t <= lbterm_oe_l;

    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --

    localbus_outputs.lready_l_o <= lready_o_l;
    localbus_outputs.lready_l_t <= lready_oe_l;
   
    --
    -- Drive the local data bus on a read.
    --

    localbus_outputs.lad_o <= ld_o;
    localbus_outputs.lad_t <= (others => ld_oe_l);

    --
    -- Latch 'lad' on assertion of 'lads_l', in order to improve timing. There
    -- is no need to increment the address, because we don't allow bursting for
    -- register addresses, and DMA transfers that burst are expected to use
    -- constant local addressing.
    --

    latch_address : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            la_q <= (others => '0');
            ldack_q <= (others => '0');
        elsif lclk'event and lclk = '1' then
            if lads_i = '1' then
                la_q <= la_i;
                ldack_q <= ldack_i;
            end if;
        end if;
    end process;

    --
    -- Generate 'ld_o', data to be output on local bus.
    --
    -- Depending on whether this is a demand-mode DMA burst or a normal direct
    -- slave burst, and the local bus address, we output one of
    --
    --   o FIFO data 'fifo_q'
    --   o 'icount' and 'ocount' register values
    --

    gen_ld_o : process(rst_i, lclk)
    begin
        if rst_i = '1' then
            ld_o <= (others => '0');
        elsif lclk'event and lclk = '1' then
            ld_o <= ld_o_slavebus;
        end if;
    end process;

 
 

 


   
    ds_stop <= '1';          -- Prevent bursting if not demand-mode DMA
  
  
    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    plxdssm0: plxdssm
        port map(
            clk => lclk,
            rst => rst_i,
            sr => logic0,
            qlads => ds_qlads,
            lblast => lblast_i,
            lbterm => lbterm_i,
            lwrite => lwrite_i,
            eld_oe => open,
            ld_oe_l => ld_oe_l,
            lready_o_l => lready_o_l,
            lready_oe_l => lready_oe_l,
            lbterm_o_l => lbterm_o_l,
            lbterm_oe_l => lbterm_oe_l,
            idle => ds_idle,
            transfer => ds_xfer,
            decode => ds_decode,
            write => ds_write,
            ready => ds_ready, -- MUST be same as fed into 'plxddsm'
            stop => ds_stop);  -- MUST be same as fed into 'plxddsm'

    --
    -- Local Bus to slave bus conversion.
    -- 
    lb2slave0: process (lclk, rst_i)
    begin  -- process lb2slave0
      if rst_i = '1' then               -- asynchronous reset
        ds_ready <= '1';
        ready_timeout_count <= (others => '0');
        ld_o_slavebus <= (others => '0');
        slavebus_rd <= '0';
        slavebus_wr <= '0';
      elsif lclk'event and lclk = '1' then  -- rising clock edge
        -- Set ds_ready low on direct slave read
        if ds_qlads = '1' and ldack_i = "0000"  and lwrite_i = '0' then
          ds_ready <= '0';
        elsif slavebus_inputs.qv = '1' or ready_timeout = '1' then
          ds_ready <= '1';
        end if;

        -- Timeout of 256 cycles if slave does not respond
        if ds_ready = '0' then
          ready_timeout_count <= ready_timeout_count +1;
        else
          ready_timeout_count <= (others => '0');
        end if;
        ready_timeout <= AND_reduce(ready_timeout_count);

        -- Output single word
        if slavebus_inputs.qv = '1' then
          ld_o_slavebus <= slavebus_inputs.q;
        elsif ready_timeout = '1' then
          ld_o_slavebus <= X"55AA55AA55AADEAD";
        end if;

        -- Generate read request
        if ds_qlads = '1' and ldack_i = "0000"  and lwrite_i = '0' then
          slavebus_rd <= '1';
        else
          slavebus_rd <= '0';
        end if;    

        -- Generate write request
        if ds_xfer = '1' and ds_write = '1' and OR_reduce(ldack_q) = '0' then
          slavebus_wr <= '1';
        else
          slavebus_wr <= '0';
        end if;        
        slavebus_outputs.rd <= slavebus_rd;
        slavebus_outputs.wr <= slavebus_wr;      
        if slavebus_wr = '1' or slavebus_rd = '1' then
          slavebus_outputs.a <= la_q(21 downto 2) & "00";
          slavebus_outputs.d <= ld_iq;
          slavebus_outputs.be <= not lbe_iq_l; 
        end if; 
      end if;
    end process lb2slave0;


    

    
end mixed;
