--
-- ocp2memif32.vhd
-- Memory Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
--
-- OCP to memif bridge
-- 32 bit data type
-- Synchronous clocking

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.ocpad.all;

entity ocp2memif32 is
  generic (
    use_chipscope : boolean := false;
    addr_width    : integer := 28);
  port (
    clk     : in  std_logic;
    -- OCP Port
    ocp_m2s : in  ocpad_burst32_m2sT;
    ocp_s2m : out ocpad_burst32_s2mT;
    -- MemIF Port (64 bits wide)
    sr      : out std_logic;
    ce      : out std_logic;
    tag     : out std_logic_vector(1 downto 0);
    w       : out std_logic;
    a       : out std_logic_vector(addr_width-1 downto 0);
    d       : out std_logic_vector(31 downto 0);
    be      : out std_logic_vector(3 downto 0);
    q       : in  std_logic_vector(31 downto 0);
    qtag    : in  std_logic_vector(1 downto 0);
    valid   : in  std_logic;
    ready   : in  std_logic);

end ocp2memif32;

architecture rtl of ocp2memif32 is


  
  type state_type is (idle, reading, writing, end_read);

  signal state        : state_type;
  signal burst_count  : std_logic_vector(ocpad_burst_width-1 downto 0);
  signal rburst_count : std_logic_vector(ocpad_burst_width-1 downto 0);

  signal SData     : std_logic_vector(31 downto 0);
  signal SResp     : std_logic_vector(1 downto 0);
  signal SRespLast : std_logic;
  signal SError    : std_logic;

  signal rrq, wrq, lrq : std_logic;
  signal r_i, w_i      : std_logic;
  signal a_i           : std_logic_vector(addr_width-1 downto 0);

  signal wrq_latch : std_logic;
  signal rrq_latch : std_logic;
  signal lrq_latch : std_logic;

  signal daccept : std_logic;


  signal SDataAccept, SCmdAccept : std_logic;


  -------------------------------------------------------------------
  --
  --  ICON core component declaration
  --
  -------------------------------------------------------------------
  component icon
    port
      (
        control0 : out std_logic_vector(35 downto 0)
        );
  end component;


  -------------------------------------------------------------------
  --
  --  ILA core component declaration
  --
  -------------------------------------------------------------------
  component ila
    port
      (
        control : in std_logic_vector(35 downto 0);
        clk     : in std_logic;
        trig0   : in std_logic_vector(71 downto 0)
        );
  end component;


  -------------------------------------------------------------------
  --
  --  ILA core signal declarations
  --
  -------------------------------------------------------------------
  signal control : std_logic_vector(35 downto 0);
  signal data    : std_logic_vector(63 downto 0);
  signal trig0   : std_logic_vector(71 downto 0);

  signal debug_counter      : std_logic_vector(11 downto 0) := (others => '0');
  signal debug_reading_last : std_logic                     := '0';

  signal timeout_count : std_logic_vector(13 downto 0);
  signal timeout       : std_logic;
  signal clr_timeout : std_logic;

  
begin  -- rtl


  sr <= '0';

  tag <= "00";



  fsm : process (clk)
  begin  -- process fsm
    if clk'event and clk = '1' then     -- rising clock edge
      SError <= '0';                    --not ready;
      if ocp_m2s.reset_n = '0' then
        state         <= idle;
        wrq           <= '0';
        rrq           <= '0';
        a_i           <= (others => '0');
        SResp         <= "00";
        wrq_latch     <= '0';
        lrq_latch     <= '0';
        timeout_count <= (others => '0');
        timeout       <= '0';
        clr_timeout   <= '0';
      else
        --wrq_latch <= wrq and not ready;
        --lrq_latch <= lrq and not ready;
        clr_timeout <= '0';
        if clr_timeout = '1' then
          timeout_count <= (others => '0');
        else
          timeout_count <= timeout_count+1;
        end if;
        timeout <= AND_reduce(timeout_count);
        case state is
          when idle =>
            if ready = '1' then
              if ocp_m2s.Cmd = ocp_Cmd_WRITE then
                state <= writing;
                if addr_width > 30 then
                  a_i <= EXT(ocp_m2s.Addr(31 downto 2), addr_width);
                else
                  a_i <= ocp_m2s.Addr(addr_width+1 downto 2);
                end if;
                burst_count <= ocp_m2s.BurstLength;
                rrq         <= '0';
                wrq         <= '0';
                
              elsif ocp_m2s.Cmd = ocp_Cmd_READ then
                state <= reading;
                if addr_width > 30 then
                  a_i <= EXT(ocp_m2s.Addr(31 downto 2), addr_width);
                else
                  a_i <= ocp_m2s.Addr(addr_width+1 downto 2);
                end if;
                burst_count  <= ocp_m2s.BurstLength;
                rburst_count <= ocp_m2s.BurstLength;
                rrq          <= '1';
                wrq          <= '0';
              else
                rrq <= '0';
                wrq <= '0';
              end if;
            else
              rrq <= '0';
              wrq <= '0';
            end if;
            SRespLast     <= '0';
            SResp         <= "00";
            lrq           <= '0';
            clr_timeout <= '1';
          when reading =>
            wrq <= '0';
            lrq <= '0';
            if r_i = '1' then
              a_i         <= a_i+1;
              burst_count <= burst_count-1;
            end if;
            if OR_reduce(burst_count(ocpad_burst_width-1 downto 1)) = '1' then
              rrq <= '1';
            else
              if ready = '1' then
                rrq <= '0';
              end if;
            end if;
            if valid = '1' then
              rburst_count <= rburst_count-1;
              SData        <= q;
              SResp        <= "01";
              if OR_reduce(rburst_count(ocpad_burst_width-1 downto 1)) = '0' then
                state     <= end_read;
                SRespLast <= '1';
              else
                SRespLast <= '0';
              end if;
               clr_timeout <= '1';
            else
              SResp         <= "00";
              SRespLast     <= '0';             
            end if;
            if timeout = '1' then
              state <= idle;
              SError <= '1';  
            end if;
          when writing =>
            rrq <= '0';
            wrq <= (ocp_m2s.DataValid and ready)  or (wrq and not ready);
            if ocp_m2s.DataLast = '1' and ocp_m2s.DataValid = '1' and ready = '1' then
              lrq <= '1';
            end if;

            if ocp_m2s.DataValid = '1' and ready = '1' then
              d  <= ocp_m2s.Data;
              be <= ocp_m2s.DataByteEn;
            end if;
            if w_i = '1' then
              a_i         <= a_i+1;
              burst_count <= burst_count-1;
--               if OR_reduce(burst_count(ocpad_burst_width-1 downto 1)) = '0' then
--                 state <= idle;
--               end if;
              if lrq = '1' then
                state <= idle;
              end if;
              clr_timeout <= '1';      
            end if;
            SRespLast <= '0';
            SResp     <= "00";
            if timeout = '1' then
              state <= idle;
              SError <= '1';  
            end if;
          when end_read =>
            rrq           <= '0';
            wrq           <= '0';
            state         <= idle;
            SRespLast     <= '0';
            SResp         <= "00";
            lrq           <= '0';
            clr_timeout <= '1';
          when others =>
            state <= idle;
        end case;
        
      end if;
    end if;
  end process fsm;


  r_i <= rrq and ready;
  w_i <= wrq and ready;

  ce <= w_i or r_i;
  w  <= w_i;
  a  <= a_i;

  ocp_s2m.Data       <= SData;
  ocp_s2m.Resp       <= SResp;
  ocp_s2m.RespLast   <= SRespLast;
  ocp_s2m.error      <= SError;
  ocp_s2m.DataAccept <= ocp_m2s.DataValid and ready when state = writing              else '0';
  ocp_s2m.CmdAccept  <= '1'                         when state = idle and ready = '1' else '0';

  SDataAccept <= ocp_m2s.DataValid and ready when state = writing              else '0';
  SCmdAccept  <= '1'                         when state = idle and ready = '1' else '0';



  debug : process (clk)
  begin  -- process debug
    if clk'event and clk = '1' then     -- rising clock edge
      if state = idle and ready = '1' and (ocp_m2s.Cmd = ocp_Cmd_WRITE or ocp_m2s.Cmd = ocp_Cmd_READ) then
        debug_counter <= "000000000001";
      else
        if OR_reduce(debug_counter) = '1' then
          debug_counter <= debug_counter+1;
        end if;
      end if;
      if state = idle and ready = '1' and ocp_m2s.Cmd = ocp_Cmd_WRITE then
        debug_reading_last <= '0';
      elsif state = idle and ready = '1' and ocp_m2s.Cmd = ocp_Cmd_READ then
        debug_reading_last <= '1';
      end if;
    end if;
  end process debug;


  gen_chipscope : if use_chipscope generate
    -------------------------------------------------------------------
    --
    --  ICON core instance
    --
    -------------------------------------------------------------------
    i_icon : icon
      port map
      (
        control0 => control
        );

    -------------------------------------------------------------------
    --
    --  ILA core instance
    --
    -------------------------------------------------------------------
    i_ila : ila
      port map
      (
        control => control,
        clk     => clk,
        trig0   => trig0
        );

    trig0(2 downto 0) <= ocp_m2s.Cmd;
    trig0(3)          <= ocp_m2s.DataValid;
    trig0(4)          <= SCmdAccept;
    trig0(5)          <= SDataAccept;
    trig0(7 downto 6) <= SResp;

    data(0)            <= rrq;
    data(1)            <= wrq;
    data(2)            <= r_i;
    data(3)            <= w_i;
    data(4)            <= ready;
    data(5)            <= valid;
    data(6)            <= '1' when state = idle    else '0';
    data(7)            <= '1' when state = reading else '0';
    data(11 downto 8)  <= burst_count(3 downto 0);
    data(15 downto 12) <= rburst_count(3 downto 0);
    data(31 downto 16) <= ocp_m2s.Addr(15 downto 0);
    --data(39 downto 32) <= ocp_m2s.Data(7 downto 0);
    --data(47 downto 40) <= SData(7 downto 0);
    data(43 downto 32) <= debug_counter;
    data(44)           <= debug_reading_last;
    data(55 downto 48) <= ocp_m2s.BurstLength(7 downto 0);
    data(56)           <= ocp_m2s.BurstSingleReq;
    data(57)           <= ocp_m2s.RespAccept;
    data(58)           <= ocp_m2s.DataLast;
    data(59)           <= SRespLast;
    data(60)           <= SError;
    data(61)           <= ocp_m2s.Reset_n;
    data(62)           <= ocp_m2s.DataByteEn(0);
    data(63)           <= ocp_m2s.DataByteEn(2);

    trig0(71 downto 8) <= data;
  end generate gen_chipscope;
  
  
end rtl;
