--
-- ocpsimmem64.vhd
-- Memory Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
--
-- OCP Simulation Memory
--
-- Simple OCP Memory bank for simulations
--
-- Default Initialisation to all 0's
--
-- Set initialise dwidth to 8,16,32 or 64 to
-- initialise with incrementing values of width *,16,32 or 64 bits)
--
-- Set initialise file to true and set a valid input_file name
-- to initialise from a file
--
-- Set a valid output file name, and use save, save_start and save_end
-- To save the memory contents in a specified address range.


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.ocpad.all;

entity ocpsimmem64 is
  generic (
    initialise_dwidth : integer := 0;
    initialise_file   : boolean := false;
    input_file        : string  := "";    
    addr_width        : integer := 16;
    latency           : integer := 1);
  port (
    clk        : in  std_logic;
    -- OCP Port
    ocp_m2s    : in  ocpad_burst64_m2sT;
    ocp_s2m    : out ocpad_burst64_s2mT;
    -- Save Data
    save       : in  boolean;
    save_start : in  integer;
    save_end   : in  integer;
    output_file : in string  := "");

end ocpsimmem64;

architecture sim of ocpsimmem64 is

  constant data_width : integer := 64;
  constant be_width   : integer := 8;

  type integer_file is file of integer;
  type state_type is (idle, reading, writing);

  signal state        : state_type;
  signal burst_count  : std_logic_vector(7 downto 0);
  signal rburst_count : std_logic_vector(7 downto 0);

  signal SData     : std_logic_vector(data_width-1 downto 0);
  signal SResp     : std_logic_vector(1 downto 0);
  signal SRespLast : std_logic;
  signal SError    : std_logic;

  signal rrq, wrq : std_logic;
  signal r_i, w_i : std_logic;
  signal a_i      : std_logic_vector(addr_width-1 downto 0);
  signal d        : std_logic_vector(data_width-1 downto 0);
  signal q        : std_logic_vector(data_width-1 downto 0);
  signal be       : std_logic_vector(be_width-1 downto 0);
  signal valid    : std_logic;

  signal rrq_latch : std_logic;

  signal daccept : std_logic;


  signal SDataAccept, SCmdAccept : std_logic;

  type   memory_type is array (0 to 2**addr_width-1) of std_logic_vector(data_width-1 downto 0);
  signal mem : memory_type := (others => (others => '0'));

  signal initialise : boolean := false;
  
  type q_pipe_type is array (0 to latency-1) of std_logic_vector(data_width-1 downto 0);
  signal q_pipe : q_pipe_type  := (others => (others => '0'));
  signal qv_pipe : std_logic_vector(latency-1 downto 0);
  
begin  -- rtl


  


  
  fsm : process (clk)
  begin  -- process fsm
    if clk'event and clk = '1' then     -- rising clock edge
      SError <= '0';
      if ocp_m2s.reset_n = '0' then
        state <= idle;
        wrq   <= '0';
        rrq   <= '0';
        a_i   <= (others => '0');
        SResp <= "00";
        
      else
        
        case state is
          when idle =>
            if ocp_m2s.Cmd = ocp_Cmd_WRITE then
              state <= writing;
              if addr_width > 29 then
                a_i <= EXT(ocp_m2s.Addr(31 downto 3), addr_width);
              else
                a_i <= ocp_m2s.Addr(addr_width+2 downto 3);
              end if;
              burst_count <= ocp_m2s.BurstLength;
              rrq         <= '0';
              wrq         <= '0';
              
            elsif ocp_m2s.Cmd = ocp_Cmd_READ then
              state <= reading;
              if addr_width > 29 then
                a_i <= EXT(ocp_m2s.Addr(31 downto 3), addr_width);
              else
                a_i <= ocp_m2s.Addr(addr_width+2 downto 3);
              end if;
              burst_count  <= ocp_m2s.BurstLength;
              rburst_count <= ocp_m2s.BurstLength;
              rrq          <= '1';
              wrq          <= '0';
            else
              rrq <= '0';
              wrq <= '0';
            end if;

            SRespLast <= '0';
            SResp     <= "00";
          when reading =>
            wrq <= '0';
            if r_i = '1' then
              a_i         <= a_i+1;
              burst_count <= burst_count-1;
            end if;
            if burst_count(7 downto 1) /= "0000000" then
              rrq <= '1';
            else              
              rrq <= '0';            
            end if;
            if valid = '1' then
              rburst_count <= rburst_count-1;
              SData        <= q;
              SResp        <= "01";
              if rburst_count(7 downto 1) = "0000000" then
                state     <= idle;
                SRespLast <= '1';
              else
                SRespLast <= '0';
              end if;
            else
              SResp     <= "00";
              SRespLast <= '0';
            end if;
          when writing =>
            rrq <= '0';
            wrq <= ocp_m2s.DataValid;
            if ocp_m2s.DataValid = '1' then
              d  <= ocp_m2s.Data;
              be <= ocp_m2s.DataByteEn;
            end if;
            if w_i = '1' then
              a_i         <= a_i+1;
              burst_count <= burst_count-1;
              if burst_count(7 downto 1) = "0000000" then
                state <= idle;
              end if;
            end if;
            SRespLast <= '0';
            SResp     <= "00";
          when others =>
            state <= idle;
        end case;
        
      end if;
    end if;
  end process fsm;


  r_i <= rrq;
  w_i <= wrq;

  ocp_s2m.Data       <= SData;
  ocp_s2m.Resp       <= SResp;
  ocp_s2m.RespLast   <= SRespLast;
  ocp_s2m.error      <= SError;
  ocp_s2m.DataAccept <= ocp_m2s.DataValid when state = writing else '0';
  ocp_s2m.CmdAccept  <= '1'               when state = idle    else '0';

  SDataAccept <= ocp_m2s.DataValid when state = writing else '0';
  SCmdAccept  <= '1'               when state = idle    else '0';

  q <= q_pipe(latency-1);
  valid <= qv_pipe(latency-1);
  
  
  model_mem : process (clk)
    file read_file  : integer_file;
    file write_file : integer_file;
    variable ai     : integer;
    variable fdata  : integer;
  begin  -- process model_mem
    if clk'event and clk = '1' then     -- rising clock edge
      if initialise then
        if initialise_dwidth = 8 then
          for i in 0 to 2**addr_width-1 loop
            mem(i) <=  CONV_STD_LOGIC_VECTOR(8*i+7, 8) & CONV_STD_LOGIC_VECTOR(8*i+6, 8) & CONV_STD_LOGIC_VECTOR(8*i+5, 8) & CONV_STD_LOGIC_VECTOR(8*i+4, 8) & CONV_STD_LOGIC_VECTOR(8*i+3, 8) & CONV_STD_LOGIC_VECTOR(8*i+2, 8) & CONV_STD_LOGIC_VECTOR(8*i+1, 8) & CONV_STD_LOGIC_VECTOR(8*i, 8);
          end loop;  -- i          
        end if;
        if initialise_dwidth = 16 then
          for i in 0 to 2**addr_width-1 loop
            mem(i) <= CONV_STD_LOGIC_VECTOR(4*i+3, 16) & CONV_STD_LOGIC_VECTOR(4*i+2, 16) & CONV_STD_LOGIC_VECTOR(4*i+1, 16) & CONV_STD_LOGIC_VECTOR(4*i, 16);
          end loop;  -- i          
        end if;
        if initialise_dwidth = 32 then
          for i in 0 to 2**addr_width-1 loop
            mem(i) <=  CONV_STD_LOGIC_VECTOR(2*i+1, 32) & CONV_STD_LOGIC_VECTOR(2*i+0, 32);
          end loop;  -- i          
        end if;

        if initialise_dwidth = 64 then
          for i in 0 to 2**addr_width-1 loop
            mem(i) <= CONV_STD_LOGIC_VECTOR(i, 64);
          end loop;  -- i          
        end if;

        if initialise_file and input_file'length > 0 then
          file_open(read_file, input_file, read_mode);
          ai := 0;
          while not endfile(read_file) loop
            read(read_file, fdata);
            mem(ai)(31 downto 0) <= CONV_STD_LOGIC_VECTOR(fdata, 32);
            if not endfile(read_file) then
              read(read_file, fdata);
              mem(ai)(63 downto 32) <= CONV_STD_LOGIC_VECTOR(fdata, 32);
            end if;
            ai := ai +1;
          end loop;
          file_close(read_file);
        end if;
      end if;
      if w_i = '1' then
        for i in 0 to be_width-1 loop
          if be(i) = '1' then
            mem(CONV_INTEGER(a_i))(8*i+7 downto 8*i) <= d(8*i+7 downto 8*i);
          end if;       
        end loop;  -- i       
      end if;
      if r_i = '1' then
        q_pipe(0) <= mem(CONV_INTEGER(a_i));
      end if;
      qv_pipe(0) <= r_i;
      if latency>1 then
        for i in 1 to latency-1 loop
          q_pipe(i) <= q_pipe(i-1);
          qv_pipe(i) <= qv_pipe(i-1);
        end loop;
      end if;

      if save then
        if output_file'length > 0 then
          file_open(write_file, output_file, write_mode);
          ai := save_start/8;
          while ai < save_end/8 loop
            fdata := CONV_INTEGER(mem(ai)(31 downto 0));
            write(write_file, fdata);
            fdata := CONV_INTEGER(mem(ai)(63 downto 32));
            write(write_file, fdata);
            ai    := ai+1;
          end loop;
          file_close(write_file);
        end if;
      end if;
      
    end if;
  end process model_mem;


  process
  begin
    wait until clk = '0';
    wait until clk = '1';
    initialise <= true;
    wait until clk = '0';
    wait until clk = '1';
    initialise <= false;
    wait;
  end process;
    
end sim;
