--
-- profile_bridge_b128b32.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
-- Profile to Profile Asynchronous Bridge
--
-- B128 to B32
--
-- Connects a B128 master to a B32 Slave profile with different clock

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.ocpad.all;

entity profile_bridge_b128b32 is
  
  generic (
    family     : family_t := family_virtex4;
    fifo_order : integer := 4);
  port (
    master_clk : in  std_logic;
    slave_clk  : in  std_logic;
    master_m2s : in  ocpad_burst128_m2sT;
    master_s2m : out ocpad_burst128_s2mT;
    slave_m2s  : out ocpad_burst32_m2sT;
    slave_s2m  : in  ocpad_burst32_s2mT);

end profile_bridge_b128b32;

architecture mixed of profile_bridge_b128b32 is

  signal rst_i                                                                                 : std_logic;
  signal cmd_fifo_data_in, cmd_fifo_data_out                                                   : std_logic_vector(43 downto 0);
  signal cmd_fifo_wadv, cmd_fifo_radv, cmd_fifo_empty, cmd_fifo_nfull, cmd_fifo_full           : std_logic;
  signal mdata_fifo_data_in, mdata_fifo_data_out                                               : std_logic_vector(144 downto 0);
  signal mdata_fifo_wadv, mdata_fifo_radv, mdata_fifo_empty, mdata_fifo_nfull, mdata_fifo_full : std_logic;
  signal sdata_fifo_data_in, sdata_fifo_data_out                                               : std_logic_vector(131 downto 0);
  signal sdata_fifo_wadv, sdata_fifo_radv, sdata_fifo_empty, sdata_fifo_nfull, sdata_fifo_full : std_logic;

  signal slave_data_w0, slave_data_w1, slave_data_w2, slave_data_w3   : std_logic_vector(31 downto 0);
  signal master_word_count                : std_logic_vector(1 downto 0) := "00";
  signal slave_word_count : std_logic_vector(1 downto 0) := "00";
  
begin  -- mixed

  rst_i <= not master_m2s.Reset_n;


  -- Generate Asychronous FIFOs
  cmd_fifo : afifo
    generic map(
      width  => 44,
      order  => fifo_order,
      wpfl   => 3,
      wpfl2  => 0,
      rpel   => 2,
      family => family,
      safety => false,
      fflush => false)
    port map(
      rst    => rst_i,
      wk     => master_clk,
      wd     => cmd_fifo_data_in,
      wadv   => cmd_fifo_wadv,
      we     => open,
      wpf    => cmd_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => slave_clk,
      rq     => cmd_fifo_data_out,
      radv   => cmd_fifo_radv,
      rflush => '0',
      re     => cmd_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  mdata_fifo : afifo
    generic map(
      width  => 145,
      order  => fifo_order,
      wpfl   => 3,
      wpfl2  => 0,
      rpel   => 2,
      family => family,
      safety => false,
      fflush => false)
    port map(
      rst    => rst_i,
      wk     => master_clk,
      wd     => mdata_fifo_data_in,
      wadv   => mdata_fifo_wadv,
      we     => open,
      wpf    => mdata_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => slave_clk,
      rq     => mdata_fifo_data_out,
      radv   => mdata_fifo_radv,
      rflush => '0',
      re     => mdata_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  sdata_fifo : afifo
    generic map(
      width  => 132,
      order  => fifo_order,
      wpfl   => 3,
      wpfl2  => 0,
      rpel   => 2,
      family => family,
      safety => false,
      fflush => false)
    port map(
      rst    => rst_i,
      wk     => slave_clk,
      wd     => sdata_fifo_data_in,
      wadv   => sdata_fifo_wadv,
      we     => open,
      wpf    => sdata_fifo_nfull,
      wpf2   => open,
      wf     => open,
      wovf   => open,
      wlevel => open,
      rk     => master_clk,
      rq     => sdata_fifo_data_out,
      radv   => sdata_fifo_radv,
      rflush => '0',
      re     => sdata_fifo_empty,
      rpe    => open,
      rf     => open,
      rudf   => open,
      rlevel => open);

  -- Register Nearly Full Signal
  process (master_clk)
  begin  -- process
    if master_clk'event and master_clk = '1' then  -- rising clock edge
      cmd_fifo_full   <= cmd_fifo_nfull;
      mdata_fifo_full <= mdata_fifo_nfull;
    end if;
  end process;
  process (slave_clk)
  begin  -- process
    if slave_clk'event and slave_clk = '1' then    -- rising clock edge
      sdata_fifo_full <= sdata_fifo_nfull;
    end if;
  end process;


  -- Map Control Signals to FIFOs

  cmd_fifo_data_in         <= master_m2s.BurstSingleReq & master_m2s.BurstLength(5 downto 0) & "00" & master_m2s.Addr & master_m2s.Cmd;
  cmd_fifo_wadv            <= '1'                           when master_m2s.Cmd /= ocp_Cmd_IDLE and cmd_fifo_full = '0' else '0';
  master_s2m.CmdAccept     <= not cmd_fifo_full;
  slave_m2s.Cmd            <= cmd_fifo_data_out(2 downto 0) when cmd_fifo_empty = '0'                                   else ocp_Cmd_IDLE;
  slave_m2s.Addr           <= cmd_fifo_data_out(34 downto 3);
  slave_m2s.BurstLength    <= cmd_fifo_data_out(42 downto 35);
  slave_m2s.BurstSingleReq <= cmd_fifo_data_out(43);
  cmd_fifo_radv            <= slave_s2m.CmdAccept and not cmd_fifo_empty;

  mdata_fifo_data_in    <= master_m2s.DataLast & master_m2s.DataByteEn & master_m2s.Data;
  mdata_fifo_wadv       <= master_m2s.DataValid and not mdata_fifo_full;
  master_s2m.DataAccept <= not mdata_fifo_full;
  slave_m2s.Data        <= mdata_fifo_data_out(31 downto 0) when master_word_count = "00" else mdata_fifo_data_out(63 downto 32) when master_word_count = "01" else mdata_fifo_data_out(95 downto 64) when master_word_count = "10" else mdata_fifo_data_out(127 downto 96);
  slave_m2s.DataByteEn  <= mdata_fifo_data_out(131 downto 128) when master_word_count = "00" else mdata_fifo_data_out(135 downto 132) when master_word_count = "01" else mdata_fifo_data_out(139 downto 136) when master_word_count = "10" else mdata_fifo_data_out(143 downto 140);
  slave_m2s.DataLast    <= mdata_fifo_data_out(144) when master_word_count = "11" else '0';
  slave_m2s.DataValid   <= not mdata_fifo_empty;
  mdata_fifo_radv       <= slave_s2m.DataAccept and not mdata_fifo_empty when master_word_count = "11" else '0';

  sdata_fifo_data_in   <= slave_s2m.error & slave_s2m.RespLast & slave_s2m.Resp & slave_data_w3 & slave_data_w2 & slave_data_w1 & slave_data_w0;
  sdata_fifo_wadv <= '1' when slave_s2m.Resp /= ocp_Resp_NONE and sdata_fifo_full = '0' and slave_word_count = "11" else '0';
  slave_m2s.RespAccept <= not sdata_fifo_full;
  master_s2m.Data      <= sdata_fifo_data_out(127 downto 0);
  master_s2m.Resp      <= sdata_fifo_data_out(129 downto 128) when sdata_fifo_empty = '0'                                    else ocp_Resp_NONE;
  master_s2m.RespLast  <= sdata_fifo_data_out(130);
  master_s2m.error     <= sdata_fifo_data_out(131);
  sdata_fifo_radv      <= master_m2s.RespAccept and not sdata_fifo_empty;
  
  slave_data_w3    <= slave_s2m.Data;
  
  width_convert : process (slave_clk, rst_i)
  begin  -- process width_convert
    if rst_i = '1' then                               -- asynchronous reset
      slave_data_w0    <= (others => '0');
      slave_data_w1    <= (others => '0');
      slave_data_w2    <= (others => '0');
      master_word_count <= "00";
      slave_word_count  <= "00";
    elsif slave_clk'event and slave_clk = '1' then  -- rising clock edge
      if slave_s2m.Resp /= ocp_Resp_NONE and sdata_fifo_full = '0' then
        slave_data_w2 <= slave_data_w3;
        slave_data_w1 <= slave_data_w2;
        slave_data_w0 <= slave_data_w1;
        slave_word_count <= slave_word_count+1;     
      end if;
      if slave_s2m.DataAccept = '1' and mdata_fifo_empty = '0' then
        master_word_count <= master_word_count+1;
      end if;      
    end if;
  end process width_convert;
  
  process (slave_clk)
  begin  -- process
    if slave_clk'event and slave_clk = '1' then  -- rising clock edge
      slave_m2s.Reset_n <= master_m2s.Reset_n;
    end if;
  end process;
  
end mixed;
