--
-- profile_bridge_r64b64.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
-- Profile to Profile Synchronous Bridge
--
-- R64 to B64
--
-- Connects an R64 master to a B64 slave profile
-- Synchronous Design
-- Generates single word transfer
--
-- Can generate multi-word burst of same data for limited access to
-- B128 slaves

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.ocpad.all;

entity profile_bridge_r64b64 is
  
  generic (
    burst_length : integer := 1);
  port (
    clk : in  std_logic;
    master_m2s : in  ocpad_register64_m2sT;
    master_s2m : out ocpad_register64_s2mT;
    slave_m2s  : out ocpad_burst64_m2sT;
    slave_s2m  : in  ocpad_burst64_s2mT);

end profile_bridge_r64b64;

architecture mixed of profile_bridge_r64b64 is

  signal rst_i : std_logic;
 
  signal data_latched : std_logic;
  signal data : std_logic_vector(63 downto 0);
  signal be : std_logic_vector(7 downto 0);

  signal burst_count : std_logic_vector(7 downto 0);
  
begin  -- mixed

  rst_i <= not master_m2s.Reset_n;

  
  
  convert_profile: process (clk, rst_i)
  begin  -- process convert_profile
    if rst_i = '1' then                 -- asynchronous reset
      data <= (others => '0');
      be <= (others => '0');
      data_latched <= '0';
    elsif clk'event and clk = '1' then  -- rising clock edge
      if master_m2s.Cmd = ocp_Cmd_WRITE then
        data <= master_m2s.Data;
        be <=  master_m2s.DataByteEn;
        data_latched <= '1';
        burst_count <= CONV_STD_LOGIC_VECTOR(burst_length-1,8);
      elsif slave_s2m.DataAccept = '1' then
        if OR_reduce(burst_count) = '0'  then
          data_latched <= '0';
        else
          burst_count <= burst_count-1;
        end if;
      end if; 
    end if;
  end process convert_profile;
  
  slave_m2s.Cmd <= master_m2s.Cmd;
  slave_m2s.Addr <= master_m2s.Addr;
  slave_m2s.BurstLength <= CONV_STD_LOGIC_VECTOR(burst_length,8);
  slave_m2s.BurstSingleReq <= '1';

  slave_m2s.Data <= data;
  slave_m2s.DataValid <= data_latched;
  slave_m2s.DataLast <= '1';
  slave_m2s.DataByteEn <= be;

  slave_m2s.RespAccept <= '1';
  slave_m2s.Reset_n <= master_m2s.Reset_n;
  
  master_s2m.CmdAccept <= slave_s2m.CmdAccept;
  master_s2m.Resp <= slave_s2m.Resp;
  master_s2m.Data <= slave_s2m.Data;
  
end mixed;
