--
-- profile_bridge_r64r32.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
-- Profile to Profile Synchronous Bridge
--
-- R64 to R32
--
-- Connects an R64 master to a R32 slave profile
-- Synchronous Design
-- Generates 2 single word transfers
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;
use work.ocpad.all;

entity profile_bridge_r64r32 is
  
  generic (
    burst_length : integer := 1);
  port (
    clk : in  std_logic;
    master_m2s : in  ocpad_register64_m2sT;
    master_s2m : out ocpad_register64_s2mT;
    slave_m2s  : out ocpad_register32_m2sT;
    slave_s2m  : in  ocpad_register32_s2mT);

end profile_bridge_r64r32;

architecture mixed of profile_bridge_r64r32 is

  signal rst_i : std_logic;
  

  signal cmd0,cmd1 : ocp_CmdT;
  signal addr0,addr1 : ocp_Data32T;
  signal data0,data1 : ocp_Data32T;
  signal be0,be1 : ocp_Data4T;
  
  type state_type is (idle, command0, gap, command1);
  
  signal state : state_type;
  
  signal sdata : ocp_Data64T;
  signal sresp : ocp_RespT;
  
  
begin  -- mixed

  rst_i <= not master_m2s.Reset_n;

  slave_m2s.Reset_n <= master_m2s.Reset_n;
  
  slave_m2s.Cmd <= cmd0 when state = command0 else cmd1 when state = command1 else (others => '0');
  slave_m2s.Addr <= addr0 when state = command0 else addr1 when state = command1 else (others => '0');
  slave_m2s.Data <= data0 when state = command0 else data1 when state = command1 else (others => '0');
  slave_m2s.DataByteEn <= be0 when state = command0 else be1 when state = command1 else (others => '0');
  
  master_s2m.CmdAccept <= '1' when master_m2s.Cmd /= ocp_Cmd_IDLE and state = idle else '0';
  master_s2m.Resp <= sresp;
  master_s2m.Data <= sdata;
  
  convert_profile: process (clk, rst_i)
  begin  -- process convert_profile
    if rst_i = '1' then                 -- asynchronous reset
      state <= idle;
      cmd0 <= ocp_Cmd_IDLE;
      cmd1 <= ocp_Cmd_IDLE;  
      sresp <= ocp_Resp_NONE;  
    elsif clk'event and clk = '1' then  -- rising clock edge
      sresp <= ocp_Resp_NONE;
      if state = idle then
        if master_m2s.Cmd /= ocp_Cmd_IDLE then
          cmd0 <= master_m2s.Cmd;
          cmd1 <= master_m2s.Cmd;
          data0 <= master_m2s.Data(31 downto 0);
          data1 <= master_m2s.Data(63 downto 32);
          be0 <=  master_m2s.DataByteEn(3 downto 0);
          be1 <=  master_m2s.DataByteEn(7 downto 4); 
          addr0 <= master_m2s.Addr;
          addr1 <= master_m2s.Addr(31 downto 3) & "100";
          state <= command0;
        end if;
      elsif state = command0 then
        if slave_s2m.CmdAccept = '1' then
          cmd0 <= ocp_Cmd_IDLE;
          if cmd0 = ocp_Cmd_WRITE then
            state <= gap;
          end if;
        end if;
        if slave_s2m.Resp = ocp_Resp_VALID then
          state <= gap;
          sdata(31 downto 0) <= slave_s2m.Data;
        end if;    
      elsif state = gap then
        state <= command1;     
      else -- state = command1
        if slave_s2m.CmdAccept = '1' then
          cmd1 <= ocp_Cmd_IDLE;
          if cmd1 = ocp_Cmd_WRITE then
            state <= idle;
          end if;
        end if;
        if slave_s2m.Resp = ocp_Resp_VALID then
          state <= idle;
          sdata(63 downto 32) <= slave_s2m.Data;
          sresp <= ocp_Resp_VALID;
        end if;   
      end if;  
    end if;
  end process convert_profile;
  
end mixed;
